// ############################################################################################
//  VLGame-AppClass-CM.cc
//  A demo application for VortexGE (implementation of VLGame's application class)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// Common headers for demo application
#include "DemoCommon.h"

// VLGame's headers
#include "VLGame-AppClass.h"


//
// VLGApplication implementation
//
const char *VLGApplication::MediaPackageMagicNumbers  = "VLGame Package File";
const char *VLGApplication::MediaPackageFileName      = "../.mediaCaches/VLGame.pac";
const char *VLGApplication::SaveGameFileName          = "../.mediaCaches/VLGame.sav";
const char *VLGApplication::VortexLuaWorkingDirectory = "../.mediaCaches/xVLGame";

#ifdef __HIGH_RESOLUTION__
const ESDM appScreenMode = ESDM_640X480_NORMAL;
#else
const ESDM appScreenMode = ESDM_640X480_LOWRES;
#endif


VLGApplication::VLGApplication()
: VLGApplicationBase(appScreenMode, "Demo : VortexLua Game"),
  audioThread (*this),
  luaManager  (VLGApplication::VortexLuaWorkingDirectory),
  mediaPool   (&iAudioEnv()),
  player      (TInit_BasicCharSpr2D("player", NULL, NULL, NULL, &luaManager)),
  uscStartRoom(NULL)
{
  // Tell the user to be patient
  ShowInitSystemPleaseWaitMessage(iVideoEnv());

  // Load the saved state
  if(VFileIsReadable(VLGApplication::SaveGameFileName))
    luaManager.LoadWorkingDirectory(VLGApplication::SaveGameFileName);

  // Initialize system
  _initMediaPackage();
  _initScript();
  _loadAudios();
  _loadImages();
  _createScene();
  mediaPool.Compact(0);

  // Play the MIDI file
  if(!MidiIsDisabled()) {
    iMidiPlayer().SetVolume(35);
    iMidiPlayer().Play(true);
  }

  // Start the audio thread
  if(!AudioIsDisabled()) audioThread.Start();

  // Set the wanted FPS
  SetWantedFPS(50.0f);
}

VLGApplication::~VLGApplication() throw()
{
  try        { audioThread.Stop(); }
  catch(...) { audioThread.Kill(); }

  _destroyScene();
}

CDataSource &VLGApplication::_GPF(const string &fileName)
{
  CDataSource *ds;
  try {
    ds = &mediaPackage.GetFile(fileName);
  }
  catch(...) {
    VError(VERR_ReadStream, _NFL_, "The media package \"file://%s\" seems to be corrupted,\nplease delete it and rerun VLGame", VLGApplication::MediaPackageFileName);
  }
  return(*ds);
}

void VLGApplication::_initMediaPackage()
{
  // Create the media package file if not exist
  // NOTE: In real application, the package should be pre-created
  //       with another application.
  if(!VFileIsReadable(VLGApplication::MediaPackageFileName)) {
    try {
      CArchiver arc("../media");
      arc.SetUserDataStr(VLGApplication::MediaPackageMagicNumbers);
      arc.AddInputFile("raw/puck-S16Ls48.raw");
      arc.AddInputFile("wav/jump.wav");
      arc.AddInputFile("wav/cling.wav");
      arc.AddInputFile("wav/fire.wav");
      arc.AddInputFile("midi/bach.mid");
      arc.AddInputFile("image/t-sprite/big-worm-red.bmp");
      arc.AddInputFile("image/t-sprite/big-worm-green.bmp");
      arc.AddInputFile("image/t-sprite/small-worm.bmp");
      arc.AddInputFile("image/t-sprite/food.bmp");
      arc.AddInputFile("image/t-sprite/3-hearths.bmp");
      arc.AddInputFile("image/t-sprite/2-stars.bmp");
      arc.AddInputFile("image/t-sprite/2-clouds.bmp");
      arc.AddInputFile("image/t-sprite/14-blocks.bmp");
      arc.AddInputFile("image/fire.bmp");
      arc.AddInputFile("image/tex1.jpg");
      arc.AddInputFile("image/tex2.jpg");
      arc.ProcessFiles(VLGApplication::MediaPackageFileName);
    }
    catch(...) {
      VRemoveFile(VLGApplication::MediaPackageFileName, true);
      VError(VERR_ReadStream, _NFL_, "Couldn't create the VLGame media package \"file://%s\"", VLGApplication::MediaPackageFileName);
    }
  }

  // Load the media package 
  try {
    mediaPackage.Open(VLGApplication::MediaPackageFileName);
    if(mediaPackage.GetUserDataStr() != VLGApplication::MediaPackageMagicNumbers)
      VError(VERR_InvalidHeaderData, _NFL_, "\"file://%s\" is an invalid VLGame media package file", VLGApplication::MediaPackageFileName);
  }
  catch(...) {
    VError(VERR_ReadStream, _NFL_, "The media package \"file://%s\" seems to be corrupted;\nplease delete it and rerun VLGame", VLGApplication::MediaPackageFileName);
  }
}

void VLGApplication::_initScript()
{
  _testScript();
}

void VLGApplication::_loadAudios()
{
  if(!AudioIsDisabled()) {
    sbWalk = mediaPool.GetNewSoundBuffer("Walking Sound", _GPF("raw/puck-S16Ls48.raw"), CMediaPool::ESADF_RAW, ERF_S16LEStereo, 48000);
    sbDoor = mediaPool.GetNewSoundBuffer("Door Sound",    _GPF("wav/jump.wav"));
    sbTalk = mediaPool.GetNewSoundBuffer("Talking Sound", _GPF("wav/cling.wav"));
    sbFire = mediaPool.GetNewSoundBuffer("Fire Sound",    _GPF("wav/fire.wav"));
    sbWalk->AdjustAmplitude(0.15f, 0.15f);
    sbDoor->AdjustAmplitude(0.50f, 0.50f);
    sbTalk->AdjustAmplitude(0.15f, 0.15f);
    sbFire->AdjustAmplitude(0.75f, 0.75f);
  }

  if(!MidiIsDisabled()) iMidiPlayer().LoadMID(_GPF("midi/bach.mid"));
}

void VLGApplication::_loadImages()
{
  // Load sprite images
  CImageBuilder imageBuilder;

  imageBuilder.LoadBMP(_GPF("image/t-sprite/big-worm-red.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("Big Worm Red", imageBuilder, 2, 2);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/big-worm-green.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("Big Worm Green", imageBuilder, 2, 2);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/small-worm.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("Small Worm", imageBuilder, 2, 2);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/food.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("Food", imageBuilder, 1, 2);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/3-hearths.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("3 Hearths", imageBuilder, 1, 3);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/2-stars.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("2 Stars", imageBuilder, 2, 4);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/2-clouds.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("2 Clouds", imageBuilder, 2, 2);

  imageBuilder.LoadBMP(_GPF("image/t-sprite/14-blocks.bmp"), true);
  imageBuilder.RoughResize(imageBuilder.GetWidth() * 4, imageBuilder.GetHeight() * 4);
  mediaPool.NewAnimatedImage("14 Blocks", imageBuilder, 14, 13);

  imageBuilder.LoadBMP(_GPF("image/fire.bmp"), true);
  imageBuilder.Resize(640*5, 96*4);
  mediaPool.NewAnimatedImage("Animated Fire", imageBuilder, 1, 10);

  // Load texture images
  CDataSource &ds1 = _GPF("image/tex1.jpg");
  try {
    imageBuilder.LoadJPG(ds1, true);
    //imageBuilder.LoadPNG("/home/aloy/Images/konsole-vge.png", true);
    imageBuilder.Resize(128, 128);
    //imageBuilder.Resize(256, 256);
    //imageBuilder.Resize(512, 512);
    imageBuilder.SmearEffect(2);
  }
  catch(...) {
    imageBuilder.CreateNoiseImage(128, 128, 4, 22, 4, 2, 8, 2);
  }
  mediaPool.NewTexture("Texture Room", imageBuilder);

  CDataSource &ds2 = _GPF("image/tex2.jpg");
  try {
    imageBuilder.LoadJPG(ds2, true);
    imageBuilder.Resize(128, 128);
  }
  catch(...) {
    imageBuilder.CreateNoiseImage(128, 128, 18, 18, 20, 2, 2, 2);
    imageBuilder.OilPaintingEffect(1);
  }
  mediaPool.NewTexture("Texture Door", imageBuilder);
}

void VLGApplication::_createScene()
{
  // Set environment states
  i3DEnv().SetFogState(14, 16, 14, 13);
  i3DEnv().SetViewport(0, 0, i3DEnv().RenderBufferWidth(), i3DEnv().RenderBufferHeight());
  i3DEnv().SetZClipping(200, 1000);
  i3DEnv().SetDefaultOmniLightPosition(-200, 200, 400);
  i3DEnv().SetDefaultOmniLightColor(24, 31, 24);

  // Initialize the start room
  uscStartRoom = new VLGUnitScene_StartRoom(luaManager, mediaPackage, mediaPool, i3DEnv());

  // Initialize player
  player.Initialize(CAnimatedImage_MP(mediaPool, "Big Worm Green"), TFloatXYZ(-150, 25, 250));
}

void VLGApplication::_destroyScene()
{
  delete uscStartRoom;

  //! \todo Find a better way to do this automatically !!!
  mediaPool.ReleaseSoundBuffer("Walking Sound");
  mediaPool.ReleaseSoundBuffer("Door Sound");
  mediaPool.ReleaseSoundBuffer("Talking Sound");
  mediaPool.ReleaseSoundBuffer("Fire Sound");
}


//
// NOTE: This still just a test for VortexLua
//
void VLGApplication::_testScript()
{
  CLuaManager        mgr(VLGApplication::VortexLuaWorkingDirectory);
  CLuaExecutable     exe   ("E1");
  CLuaExecutionSpace xspace("X1");

  mgr.LERegister(exe);
  mgr.LXRegister(xspace);

  struct TestAction {
    static CLuaActionList::EActionCallbackResult testAct1(CLuaExecutionSpace &executingNode, const CLuaActionCallbackParams &actionInParams, CLuaActionCallbackParams &actionOutParams)
    { 
      if(actionInParams.Count()) VInfo(_NFL_, "Executing node  : %p\nParameters given :", &executingNode);
      for(uint32_t idx = 0; idx < actionInParams.Count(); idx++) {
             if(actionInParams.IsBoolean(idx)) VPrint("    Boolean -> %s\n", actionInParams.ToBoolean(idx) ? "true" : "false");
        else if(actionInParams.IsNumber (idx)) VPrint("    Number  -> %f\n", actionInParams.ToNumber(idx));
        else if(actionInParams.IsString (idx)) VPrint("    String  -> %s\n", actionInParams.ToString(idx).c_str());
      }
      actionOutParams.PushString("yo!");
      actionOutParams.PushNumber(VLUA_TO_NUMBER(5.3));
      actionOutParams.PushBoolean(true);
      return(CLuaActionList::EACR_OK); 
    }
  };

  xspace.ActionList().Add("testAct1", TestAction::testAct1);

  const char code[] =                                                                \
  "                                                                              \n" \
  "print('>>> Testing vge.getVar()')                                             \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "r,e = vge.getVar('[SELF]','var1',vge.typeBoolean)                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "r,e = vge.getVar('[SELF]','var1',vge.typeNumber)                              \n" \
  "                                                                              \n" \
  "                                                                              \n" \
  "print('>>> Testing helper functions')                                         \n" \
  "print(vge.errorString(vge.errorInvalidParameter))                              \n" \
  "print(vge.errorString(vge.errorNotExist))                                     \n" \
  "print(vge.errorString(0))                                                     \n" \
  "print(vge.errorString(9999))                                                  \n" \
  "print(vge.getNodeName('[SELF]'))                                              \n" \
  "print(vge.getGroupName('[SELF]'))                                             \n" \
  "print(vge.getNodeName('[EXE]'))                                               \n" \
  "print(vge.getGroupName('[EXE]'))                                              \n" \
  "                                                                              \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.extensionVersion()')                                   \n" \
  "mj,mn,rv = vge.extensionVersion()                                             \n" \
  "print('VortexLua version',mj,mn,rv)                                           \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.extensionVersionMinimum()')                            \n" \
  "print('0.0.9',vge.extensionVersionMinimum(0,0,9))                             \n" \
  "print('0.2.9',vge.extensionVersionMinimum(0,2,9))                             \n" \
  "print('0.3.2',vge.extensionVersionMinimum(0,3,2))                             \n" \
  "print('0.3.3',vge.extensionVersionMinimum(0,3,3))                             \n" \
  "print('0.3.4',vge.extensionVersionMinimum(0,3,4))                             \n" \
  "print('0.4.0',vge.extensionVersionMinimum(0,4,0))                             \n" \
  "print('0.9.9',vge.extensionVersionMinimum(0,9,9))                             \n" \
  "print('1.0.0',vge.extensionVersionMinimum(1,0,0))                             \n" \
  "                                                                              \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.act()')                                                \n" \
  "r1 = vge.act('testAct1')                                                      \n" \
  "r2 = vge.act('testAct2')                                                      \n" \
  "print('testAct1',vge.errorString(r1))                                         \n" \
  "print('testAct2',vge.errorString(r2))                                         \n" \
  "                                                                              \n" \
  "r1,n,a,b,c,d,e = vge.act('testAct1',true,2.7,'hello')                         \n" \
  "print('number of returned parameter = ',n)                                    \n" \
  "print('arg #1 = ',a)                                                          \n" \
  "print('arg #2 = ',b)                                                          \n" \
  "print('arg #3 = ',c)                                                          \n" \
  "print('arg #4 = ',d)                                                          \n" \
  "print('arg #5 = ',e)                                                          \n" \
  "                                                                              \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.addVar()')                                             \n" \
  "print(vge.errorString(vge.addVar('[SELF]','var1',vge.typeBoolean)))           \n" \
  "print(vge.errorString(vge.addVar('[SELF]','var2',vge.typeNumber)))            \n" \
  "print(vge.errorString(vge.addVar('[SELF]','var3',vge.typeString)))            \n" \
  "print(vge.errorString(vge.addVar('[SELF]','var3',vge.typeString)))            \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.hasVar()')                                             \n" \
  "print(vge.hasVar('[SELF]','var1',vge.typeAny))                                \n" \
  "print(vge.hasVar('[SELF]','var1',vge.typeBoolean))                            \n" \
  "print(vge.hasVar('[SELF]','var1',vge.typeNumber))                             \n" \
  "print(vge.hasVar('[SELF]','var1',vge.typeString))                             \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.getVar()')                                             \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "r,e = vge.getVar('[SELF]','var1',vge.typeBoolean)                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "r,e = vge.getVar('[SELF]','var1',vge.typeNumber)                              \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.setVar()')                                             \n" \
  "print(vge.errorString(vge.setVar('[SELF]','var1',true)))                      \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "print(vge.errorString(vge.setVar('[SELF]','var1',false)))                     \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "print(vge.errorString(vge.setVar('[SELF]','var1','XXX')))                     \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "print(vge.errorString(vge.setVar('[SELF]','var1',10000)))                     \n" \
  "r,e = vge.getVar('[SELF]','var1')                                             \n" \
  "print(r, ' ', vge.errorString(e))                                             \n" \
  "                                                                              \n" \
/*"print('>>> Testing vge.delVar()')                                             \n" \
  "print(vge.errorString(vge.delVar('[SELF]','var1')))                           \n" \
  "print(vge.errorString(vge.delVar('[SELF]','var2')))                           \n" \
  "print(vge.errorString(vge.delVar('[SELF]','var3')))                           \n" \
  "print(vge.errorString(vge.delVar('[SELF]','var3')))                           \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.hasVar()')                                             \n" \
  "print(vge.hasVar('[SELF]','var1',vge.typeAny))                                \n" \
  "print(vge.hasVar('[SELF]','var2',vge.typeAny))                                \n" \
  "print(vge.hasVar('[SELF]','var3',vge.typeAny))                                \n" \*/
  "                                                                              \n" \
  "                                                                              \n" \
  "print('>>> Testing vge.queryNodeInGroup()')                                   \n" \
  "t = vge.queryNodeInGroup('[FOO]');  print(t)                                  \n" \
  "t = vge.queryNodeInGroup('[SELF]'); print(t); print(0, t[0])                  \n" \
  "t = vge.queryNodeInGroup('[ALL]');  print(t)                                  \n" \
  "if t ~= nil then                                                              \n" \
  "  i = 0                                                                       \n" \
  "  repeat                                                                      \n" \
  "    print(i, t[i])                                                            \n" \
  "    i = i + 1                                                                 \n" \
  "  until t[i] == nil                                                           \n" \
  "end                                                                           \n" \
  "                                                                              \n" \
  "                                                                              \n";
  mgr.Compile(exe, code);
  //xspace.Execute(exe);
}
