/** \file AnimatedImageController.h */
// ############################################################################################
//  AnimatedImageController.h
//  Definition & implementation of CAImgControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___ANIMATED_IMAGE_CONTROLLER_H_
#define _2D_MANIPULATION___ANIMATED_IMAGE_CONTROLLER_H_

// VortexGE headers
#include "../System/DataSinkSource.h"
#include "ImageClasses.h"


namespace vge {

  //!
  //! \internal
  //! \brief Animation set structure
  //!
  //! \api pretty stable
  struct _TAnimSet {
    uint32_t *pun32FrameList;  // Frame references to CAnimatedImage
    uint32_t  un32TotalFrames; // Total frames for this animation set
  };


  //!
  //! \brief The advanced frame control for animated image class
  //!
  //! \api pretty stable
  class VGE_EXPORT CAImgControl : VCLASS_UNCOPYABLE {
    public:
      CAImgControl();
      ~CAImgControl() throw();

      // Main initialization functions
      //!
      //! Use this for selecting the CAnimatedImage instance to be used.
      void SetImageSource(CAnimatedImage *pcsAnimatedImage);
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeAll();

      // Animation set management functions
      //!
      //! Create a new animation set.
      //! \param[in] un32TotalASFrames How many frame should be preallocated for this new animation set?
      //! \return Index to the new animation set.
      uint32_t AddAnimationSet(uint32_t un32TotalASFrames = 0);
      //!
      //! Create new animation sets.
      //! \param[in] un32ASCount        How many animation sets should be created?
      //! \param[in] aun32TotalASFrames How many frame should be preallocated for these new animation sets?
      //! \return Start index to the new animation sets.
      uint32_t AddAnimationSets(uint32_t un32ASCount, uint32_t aun32TotalASFrames[] = NULL);
      //!
      //! Use this to check how many animation sets have been allocated.
      uint32_t GetAnimationSetCount() const;

      // Frame management functions
      //!
      //! Add a new frame to the animation set.
      //! \param[in] un32AnimationSetIndex Index to the animation set.
      //! \param[in] un32FrameReference    Reference to a frame in the used CAnimatedImage instance.
      void AddASFrame(uint32_t un32AnimationSetIndex, uint32_t un32FrameReference);
      //!
      //! Add new frames to the animation set.
      //! \param[in] un32AnimationSetIndex Index to the animation set.
      //! \param[in] un32FrameCount        How many frames will be addded?
      //! \param[in] aun32FrameReferences  References to frames in the used CAnimatedImage instance.
      void AddASFrames(uint32_t un32AnimationSetIndex, uint32_t un32FrameCount, uint32_t aun32FrameReferences[]);
      //!
      //! Use this to check how many frames have been allocated in the given animation set.
      uint32_t GetASFrameCount(uint32_t un32AnimationSetIndex) const;
      //!
      //! Changing a frame reference in given the animation set.
      //! \param[in] un32AnimationSetIndex Index to the animation set.
      //! \param[in] un32ASFrameIndex      Index to the frame in the animation set.
      //! \param[in] un32FrameReference    Reference to a frame in the used CAnimatedImage instance.
      void SetASFrame(uint32_t un32AnimationSetIndex, uint32_t un32ASFrameIndex, uint32_t un32FrameReference);
      //!
      //! Changing frame references in given the animation set.
      //! \param[in] un32AnimationSetIndex Index to the animation set.
      //! \param[in] aun32FrameReferences  Reference to frames in the used CAnimatedImage instance.
      //!
      //! \note The total elements in the aun32FrameReferences array must be equal to the value returned by GetASFrameCount().
      void SetASFrames(uint32_t un32AnimationSetIndex, uint32_t aun32FrameReferences[]);

      // Animation control functions
      //!
      //! Activate the given animation set.
      inline void SelectAnimationSet(uint32_t un32AnimationSetIndex);
      //!
      //! Order the used CAnimatedImage to select the next frame according to the registration in the animation set.
      //! \param[in] bWarp If "true", the frame will be warped back to the 1st frame if past the last frame.
      inline void SelectNextASFrame(bool bWarp = false);
      //!
      //! Order the used CAnimatedImage to select the previous frame according to the registration in the animation set.
      //! \param[in] bWarp If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void SelectPrevASFrame(bool bWarp = false);
      //!
      //! Order the used CAnimatedImage to select the next frame according to the registration in the animation set.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the next frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the 1st frame if past the last frame.
      inline void SelectNextASFrameDly(uint32_t un32Delay, bool bWarp = false);
      //!
      //! Order the used CAnimatedImage to select the previous frame according to the registration in the animation set.
      //! \param[in] un32Delay The number of delay (in frames) to be taken before stepping to the previous frame.
      //! \param[in] bWarp     If "true", the frame will be warped back to the last frame if past the 1st frame.
      inline void SelectPrevASFrameDly(uint32_t un32Delay, bool bWarp = false);

      // I/O functions
      //!
      //! Save to VortexGE's native VAC format.
      //! \param[in] sink The name of the file (.vac) to be saved to or an instance of ostream derived class.
      void SaveVAC(const CDataSink &sink) const;
      //!
      //! Load from VortexGE's native VAC format.
      //! \param[in] source The name of the file (.vac) to be loaded or an instance of istream derived class.
      void LoadVAC(const CDataSource &source);

    protected:
      // Initialization functions
      //!
      //! \internal Used for allocating frame list for the corresponding animation set.
      //!           \sntc
      void _AllocASFrame_(uint32_t un32AnimationSetIndex, uint32_t un32TotalASFrames);

      // Updating functions
      //!
      //! \internal Used for selecting the next frame in the CAnimatedImage.
      //!           \sntc
      inline void _UpdateImgSrcFrame_();

      // Data
      CAnimatedImage *_pcsImgSrc;     // Reference to CAnimatedImage
      uint8_t       **_ppun8ImgBuff;  // Reference to the image source buffer
      _TAnimSet      *_ptASList;      // Animaton set list
      uint32_t        _un32TotalAS;   // Total animation set
      uint32_t        _un32SelAS;     // Selected animation set
      uint32_t        _un32SelFrame;  // Selected frame of the selected animation set
      uint32_t        _un32StepDelay; // The step delay
  };

  void CAImgControl::SelectAnimationSet(uint32_t un32AnimationSetIndex)
  {
    _un32SelAS     = min(un32AnimationSetIndex, _un32TotalAS - 1);
    _un32SelFrame  = 0;
    _un32StepDelay = 0;
    _UpdateImgSrcFrame_();
  }

  void CAImgControl::SelectNextASFrame(bool bWarp)
  {
    if(_un32SelFrame >= (_ptASList[_un32SelAS].un32TotalFrames - 1)) {
      if(bWarp) {
        _un32SelFrame = 0;
        _UpdateImgSrcFrame_();
      }
      return;
    }
    _un32SelFrame++;
    _UpdateImgSrcFrame_();
  }

  void CAImgControl::SelectPrevASFrame(bool bWarp)
  {
    if(!_un32SelFrame) {
      if(bWarp) {
        _un32SelFrame = _ptASList[_un32SelAS].un32TotalFrames - 1;
        _UpdateImgSrcFrame_();
      }
      return;
    }
    _un32SelFrame--;
    _UpdateImgSrcFrame_();
  }

  void CAImgControl::SelectNextASFrameDly(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    SelectNextASFrame(bWarp);
  }

  void CAImgControl::SelectPrevASFrameDly(uint32_t un32Delay, bool bWarp)
  {
    // Do delay
    if(++_un32StepDelay < un32Delay) return;
    _un32StepDelay = 0;
    // Do step
    SelectNextASFrame(bWarp);
  }

  void CAImgControl::_UpdateImgSrcFrame_()
  {
    _pcsImgSrc->_pun8Buffer = _ppun8ImgBuff[_ptASList[_un32SelAS].pun32FrameList[_un32SelFrame]];
  }

} // namespace vge
#endif
