/** \file BitmapFont.h */
// ############################################################################################
//  BitmapFont.h
//  Definition & implementation of CBitmapFont class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___BITMAP_FONT_H_
#define _2D_MANIPULATION___BITMAP_FONT_H_

// VortexGE headers
#include "../VideoEnvironment/Color.h"
#include "../System/DataSinkSource.h"


namespace vge {

  // Size of the font matrix
  const uint32_t _FONT_ROW_ = uint32_t(24);                          //!< \internal Modifying this means modifying also the _XLOOPS_ value in "util/fontDisplay.cc"
  const uint32_t _FONT_COL_ = uint32_t(48);                          //!< \internal Modifying this means modifying also the _YLOOPS_ value in "util/fontDisplay.cc"
  const uint32_t _FONT_MAP_ = uint32_t(_FONT_ROW_ * _FONT_COL_ / 8); //!< \internal A map containing markers that will indicate whether the corresponding font cells are occupied or empty

  // Delta between the real font size with its cell (used when loading BMP files)
  const uint32_t _FONT_DX_ = uint32_t(2); //!< \internal Adjustment factor
  const uint32_t _FONT_DY_ = uint32_t(2); //!< \internal Adjustment factor

  //!
  //! \brief Font styles
  //!
  //! \api pretty stable
  // Modifying this means modifying also the drawing sequences in the DrawFont() function in "util/fontDisplay.cc".
  enum EFontStyle {
    EFS_Regular = 0, EFS_Italic = 288, EFS_Bold = 576, EFS_BoldItalic = 864
  };


  //!
  //! \brief The bitmap font class
  //!
  //! A class used for drawing text using bitmap fonts.
  //!
  //! \api pretty stable
  class VGE_EXPORT CBitmapFont : VCLASS_UNCOPYABLE {
    public:
      CBitmapFont();
      ~CBitmapFont() throw();

      // Font functions
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeFont();

      // Drawing functions
      //!
      //! Draw a text using bitmap font on the active 3D environment render buffer.
      //! \param[in] n32X,n32Y  Location of the text.
      //! \param[in] csText     The text.
      //! \param[in] xColorVRGB Color of the text, see the function V_RGB() and V_CLR().
      //! \param[in] eStyle     Style of the text.
      void PutText(int32_t n32X, int32_t n32Y, const string& csText, XColor15 xColorVRGB, EFontStyle eStyle = EFS_Regular) const;
      //!
      //! Draw a text using bitmap font on the active 3D environment render buffer.
      //! \param[in] n32X,n32Y      Location of the text.
      //! \param[in] csText         The text.
      //! \param[in] xTextColorVRGB Foreground color of the text, see the function V_RGB() and V_CLR().
      //! \param[in] xBackColorVRGB Background color of the text, see the function V_RGB() and V_CLR().
      //! \param[in] eStyle         Style of the text.
      void PutTextSolid(int32_t n32X, int32_t n32Y, const string &csText, XColor15 xTextColorVRGB, XColor15 xBackColorVRGB, EFontStyle eStyle = EFS_Regular) const;

      // I/O functions
      //!
      //! Load and generate bitmap font from Windows(TM) bitmap file BMP format.
      //! \param[in] source The name of the file (.bmp) to be loaded or an instance of istream derived class.
      void LoadBMP(const CDataSource &source);
      //!
      //! Load from VortexGE's native VBF format.
      //! \param[in] source The name of the file (.vbf) to be loaded or an instance of istream derived class.
      void LoadVBF(const CDataSource &source);
      //!
      //! Save to VortexGE's native VBF format.
      //! \param[in] sink The name of the file (.vbf) to be saved to or an instance of ostream derived class.
      void SaveVBF(const CDataSink &sink) const;

    protected:
      char    *_apcFontArray[_FONT_ROW_ * _FONT_COL_]; // The Font array
      uint32_t _un32FW;                                // Width of the font
      uint32_t _un32FH;                                // Height of the font
      uint32_t _un32FW1;                               // Width of the font - 1
      uint32_t _un32FH1;                               // Height of the font - 1
  };

} // namespace vge
#endif
