// ############################################################################################
//  ImageBuilder3.cc
//  Implementation of CImageBuilder class (3rd part)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../VideoEnvironment/VideoEnvironment.h"
#include "ImageBuilder.h"
#include "Fractal/BasicFractals.h"


namespace vge {

//
// Builder functions
//
void CImageBuilder::ClearImage(uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  // Check
  if(!_ptBuffer)
    VError(VERR_NotInitialized, _NFL_, "Uninitialized CImageBuilder can't be processed using\nClearImage(); use CreateEmptyImage() instead");

  // Fill with color
  u_char *pucTmp = reinterpret_cast<u_char*>(_ptBuffer);
  for(uint32_t un32Y = 0; un32Y < _un32Height; un32Y++) {
    for(uint32_t un32X = 0; un32X < _un32Width; un32X++) {
      *pucTmp++ = un8B;
      *pucTmp++ = un8G;
      *pucTmp++ = un8R;
    }
  }

  // Reset the image to a standard RGB 555 image
  memset(&_tImgInfo, 0, sizeof(_tImgInfo));
}

void CImageBuilder::CloneImage(CImageBuilder **pcsNewImage)
{
  // Temporary variables
  CImageBuilder *pcsNI;
  uint32_t            un32Size;
  u_char              *pucData;

  // Check
  if(!_ptBuffer)
    VError(VERR_NotInitialized, _NFL_, "No data to be cloned (this CImageBuilder is not initialized)");

  // Create a new image builder
  pcsNI = _nothrow_new CImageBuilder;
  if(pcsNI == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for creating a new CImageBuilder instance");

  // Allocate memory for image buffer
  un32Size = _un32Width * _un32Height * 3;
  pucData  = reinterpret_cast<u_char*>(malloc(un32Size));
  if(pucData == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  // Copy datas
  pcsNI->_un32Width  = _un32Width;
  pcsNI->_un32Height = _un32Height;
  pcsNI->_ptBuffer   = reinterpret_cast<TUN8RGB*>(pucData);
  memcpy(pcsNI->_ptBuffer, _ptBuffer, _un32Width * _un32Height * 3);
  memcpy(&pcsNI->_tImgInfo, &_tImgInfo, sizeof(_tImgInfo));

  // OK !
  *pcsNewImage = pcsNI;
}

void CImageBuilder::CreateEmptyImage(const CImageBuilder &csImageSizeRef)
{
  CreateEmptyImage(csImageSizeRef._un32Width, csImageSizeRef._un32Height);
}

void CImageBuilder::CreateEmptyImage(uint32_t un32ImageWidth, uint32_t un32ImageHeight)
{
  // Temporary variables
  uint32_t un32Size;
  u_char   *pucData;

  // No need to allocate a new buffer if the previous image
  //       is exist and the size is the same, just clear it
  if((_ptBuffer != NULL) && (_un32Width == un32ImageWidth) && (_un32Height == un32ImageHeight))
    ClearImage(0, 0, 0);

  // Allocate memory for image buffer
  un32Size = un32ImageWidth * un32ImageHeight * 3;
  pucData  = reinterpret_cast<u_char*>(malloc(un32Size));
  if(pucData == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
  memset(pucData, 0, un32Size);

  // Destroy previous image, if any
  FreeImage();

  // Copy variables
  _un32Width  = un32ImageWidth;
  _un32Height = un32ImageHeight;
  _ptBuffer   = reinterpret_cast<TUN8RGB*>(pucData);
}

void CImageBuilder::CreateNoiseImage(CImageBuilder &csImageSizeRef,
                                     uint8_t un8RBase, uint8_t un8GBase, uint8_t un8BBase,
                                     uint8_t un8RNoise, uint8_t un8GNoise, uint8_t un8BNoise)
{
  CreateNoiseImage(csImageSizeRef._un32Width, csImageSizeRef._un32Height,
                   un8RBase, un8GBase, un8BBase, un8RNoise, un8GNoise, un8BNoise);
}

void CImageBuilder::CreateNoiseImage(uint32_t un32ImageWidth, uint32_t un32ImageHeight,
                                     uint8_t un8RBase, uint8_t un8GBase, uint8_t un8BBase,
                                     uint8_t un8RNoise, uint8_t un8GNoise, uint8_t un8BNoise)
{
  // No need to allocate a new buffer if the previous image
  //       is exist and the size is the same, just use it
  u_char *pucData;
  bool    bNB = false;
  if((_ptBuffer != NULL) && (_un32Width == un32ImageWidth) && (_un32Height == un32ImageHeight))
    pucData  = reinterpret_cast<u_char*>(_ptBuffer);
  else {
    pucData  = NULL;
    bNB      = true;
  }
  const uint32_t un32Size = un32ImageWidth * un32ImageHeight * 3;

  // Allocate memory for temporary buffer
  if(bNB) {
    pucData  = reinterpret_cast<u_char*>(malloc(un32Size));
    if(pucData == NULL)
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");
  }


#define PUT_NOISE_PIXEL(B, N, F)                                                 \
      if(!N)                                                                     \
        *pucTmp++ = B;                                                           \
      else {                                                                     \
        int8_t C = B + int8_t(rngf.GenerateBipolar(-F, F));                      \
        *pucTmp++ = (C > MAX_COLOR_VALUE) ? MAX_COLOR_VALUE : max<int8_t>(C, 0); \
      }

  // Create image
  CRandomFloat rngf(_rnGen);
  const float  nr     = un8RNoise;
  const float  ng     = un8GNoise;
  const float  nb     = un8BNoise;
  u_char      *pucTmp = pucData;
  for(uint32_t un32Y = 0; un32Y < un32ImageHeight; un32Y++) {
    for(uint32_t un32X = 0; un32X < un32ImageWidth; un32X++) {
      PUT_NOISE_PIXEL(un8BBase, un8BNoise, nb)
      PUT_NOISE_PIXEL(un8GBase, un8GNoise, ng)
      PUT_NOISE_PIXEL(un8RBase, un8RNoise, nr)
    }
  }

#undef PUT_NOISE_PIXEL

  // Update the image data (if necessary)
  if(bNB) {
    // Destroy previous image, if any
    FreeImage();
    // Copy variables
    _un32Width  = un32ImageWidth;
    _un32Height = un32ImageHeight;
    _ptBuffer   = reinterpret_cast<TUN8RGB*>(pucData);
  }
  // Reset the image to a standard RGB 555 image
  else
    memset(&_tImgInfo, 0, sizeof(_tImgInfo));
}


//
// Image cropping functions
//
void CImageBuilder::Crop(uint32_t un32X1, uint32_t un32Y1, uint32_t un32X2, uint32_t un32Y2)
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag == _IMG_TYPE_UDCT_RGB5_) ||
     (_tImgInfo.un8Flag == _IMG_TYPE_YV12_))
    VError(VERR_InvalidCall, _NFL_, "Can't process an _IMG_TYPE_UDCT_RGB5_ nor _IMG_TYPE_YV12_ image");

  // Preparation
  if(un32X2 >= _un32Width)  un32X2 = _un32Width - 1;
  if(un32Y2 >= _un32Height) un32Y2 = _un32Height - 1;
  uint32_t un32W     = un32X2 - un32X1 + 1;
  uint32_t un32H     = un32Y2 - un32Y1 + 1;
  uint32_t un32Size  = un32W * un32H * 3;
  u_char   *pucData  = reinterpret_cast<u_char*>(malloc(un32Size));
  if(pucData == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for temporary image buffer");

  // Perform cropping
  u_char *pucSrc = reinterpret_cast<u_char*>(_ptBuffer) + ((un32Y1 * _un32Width) + un32X1) * 3;
  u_char *pucDst = pucData;
  for(uint32_t un32Y = 0; un32Y < un32H; un32Y++) {
    memcpy(pucDst, pucSrc, un32W * 3);
    pucSrc += (_un32Width * 3);
    pucDst += (un32W * 3);
  }

  // Destroy previous image
  FreeImage();

  // Copy variables
  _un32Width  = un32W;
  _un32Height = un32H;
  _ptBuffer   = reinterpret_cast<TUN8RGB*>(pucData);

}

void CImageBuilder::AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB)
{
  AutoCrop(un8BckColR, un8BckColG, un8BckColB, 0, 0, 0);
}

void CImageBuilder::AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB,
                                   int8_t n8Tolerance)
{
  AutoCrop(un8BckColR, un8BckColG, un8BckColB, n8Tolerance, n8Tolerance, n8Tolerance);
}

void CImageBuilder::AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB,
                                        int8_t n8ToleranceR, int8_t n8ToleranceG, int8_t n8ToleranceB)
{
  // Check
  if(_ptBuffer == NULL)
    VError(VERR_NotInitialized, _NFL_, "No data to be transferred (this CImageBuilder is not initialized)");
  if((_tImgInfo.un8Flag == _IMG_TYPE_UDCT_RGB5_) ||
     (_tImgInfo.un8Flag == _IMG_TYPE_YV12_))
    VError(VERR_InvalidCall, _NFL_, "Can't process an _IMG_TYPE_UDCT_RGB5_ nor _IMG_TYPE_YV12_ image");

  // Grab the source image
  TUN8RGB *ptSrc = _ptBuffer;

  // Search for X1
  bool     bGotX1 = false;
  uint32_t un32X1 = 0;
  for(uint32_t un32X = 0; un32X < _un32Width; un32X++) {
    TUN8RGB *ptSrcI = ptSrc + un32X;
    for(uint32_t un32Y = 0; un32Y < _un32Height; un32Y++) {
      uint8_t un8Rd = (un8BckColR > ptSrcI->un8R) ? (un8BckColR - ptSrcI->un8R) : (ptSrcI->un8R - un8BckColR);
      uint8_t un8Gd = (un8BckColG > ptSrcI->un8G) ? (un8BckColG - ptSrcI->un8G) : (ptSrcI->un8G - un8BckColG);
      uint8_t un8Bd = (un8BckColB > ptSrcI->un8B) ? (un8BckColB - ptSrcI->un8B) : (ptSrcI->un8B - un8BckColB);
      uint8_t un8Dd = ((un8Rd <= n8ToleranceR) ? 0 : 1) |
                      ((un8Gd <= n8ToleranceG) ? 0 : 1) |
                      ((un8Bd <= n8ToleranceB) ? 0 : 1);
      if(un8Dd) {
        bGotX1 = true;
        un32X1 = un32X;
        un32X  = _un32Width;
        break;
      }
      ptSrcI += _un32Width;
    }
  }
  if(!bGotX1)
    VError(VERR_InvalidParameter, _NFL_, "All pixels in this CImageBuilder fall in the\ntolerance range of the given background color;\nauto-cropping is not possible");

  // Search for X2
  uint32_t un32X2 = _un32Width - 1;
  for(uint32_t un32X = un32X1; un32X < _un32Width; un32X++) {
    TUN8RGB *ptSrcI = ptSrc + un32X;
    for(uint32_t un32Y = 0; un32Y < _un32Height; un32Y++) {
      uint8_t un8Rd = (un8BckColR > ptSrcI->un8R) ? (un8BckColR - ptSrcI->un8R) : (ptSrcI->un8R - un8BckColR);
      uint8_t un8Gd = (un8BckColG > ptSrcI->un8G) ? (un8BckColG - ptSrcI->un8G) : (ptSrcI->un8G - un8BckColG);
      uint8_t un8Bd = (un8BckColB > ptSrcI->un8B) ? (un8BckColB - ptSrcI->un8B) : (ptSrcI->un8B - un8BckColB);
      uint8_t un8Dd = ((un8Rd <= n8ToleranceR) ? 0 : 1) |
                      ((un8Gd <= n8ToleranceG) ? 0 : 1) |
                      ((un8Bd <= n8ToleranceB) ? 0 : 1);
      if(un8Dd) {
        un32X2 = (un32X - 1);
        break;
      }
      ptSrcI += _un32Width;
    }
  }

  // Search for Y1
  uint32_t un32Y1 = 0;
  for(uint32_t un32Y = 0; un32Y < _un32Height; un32Y++) {
    TUN8RGB *ptSrcI = ptSrc + (un32Y * _un32Width);
    for(uint32_t un32X = 0; un32X < _un32Width; un32X++) {
      uint8_t un8Rd = (un8BckColR > ptSrcI->un8R) ? (un8BckColR - ptSrcI->un8R) : (ptSrcI->un8R - un8BckColR);
      uint8_t un8Gd = (un8BckColG > ptSrcI->un8G) ? (un8BckColG - ptSrcI->un8G) : (ptSrcI->un8G - un8BckColG);
      uint8_t un8Bd = (un8BckColB > ptSrcI->un8B) ? (un8BckColB - ptSrcI->un8B) : (ptSrcI->un8B - un8BckColB);
      uint8_t un8Dd = ((un8Rd <= n8ToleranceR) ? 0 : 1) |
                      ((un8Gd <= n8ToleranceG) ? 0 : 1) |
                      ((un8Bd <= n8ToleranceB) ? 0 : 1);
      if(un8Dd) {
        un32Y1 = un32Y;
        un32Y  = _un32Height;
        break;
      }
      ptSrcI++;
    }
  }

  // Search for Y2
  uint32_t un32Y2 = _un32Height - 1;
  for(uint32_t un32Y = un32Y1; un32Y < _un32Height; un32Y++) {
    TUN8RGB *ptSrcI = ptSrc + (un32Y * _un32Width);
    for(uint32_t un32X = 0; un32X < _un32Width; un32X++) {
      uint8_t un8Rd = (un8BckColR > ptSrcI->un8R) ? (un8BckColR - ptSrcI->un8R) : (ptSrcI->un8R - un8BckColR);
      uint8_t un8Gd = (un8BckColG > ptSrcI->un8G) ? (un8BckColG - ptSrcI->un8G) : (ptSrcI->un8G - un8BckColG);
      uint8_t un8Bd = (un8BckColB > ptSrcI->un8B) ? (un8BckColB - ptSrcI->un8B) : (ptSrcI->un8B - un8BckColB);
      uint8_t un8Dd = ((un8Rd <= n8ToleranceR) ? 0 : 1) |
                      ((un8Gd <= n8ToleranceG) ? 0 : 1) |
                      ((un8Bd <= n8ToleranceB) ? 0 : 1);
      if(un8Dd) {
        un32Y2 = un32Y - 1;
        break;
      }
      ptSrcI++;
    }
  }

  // Crop it
  Crop(un32X1, un32Y1, un32X2, un32Y2);
}

} // namespace vge
