// ############################################################################################
//  HierarchicalNode.cc
//  Implementation of CHNode class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "HierarchicalNode.h"

// C++ headers
#include <fstream>


namespace vge {

// VHN header
#define _VHN_ID1_ 'V'
#define _VHN_ID2_ 'H'
#define _VHN_ID3_ 'N'
#define _VHN_VER_ 100

struct _TVHNHeader
{
  char     acID[3];       // File ID, must be "VHN"
  uint8_t  un8Version;    // File version
  uint32_t un32Reserved1; // Reserved, must be zero
  uint32_t un32Reserved2; // Reserved, must be zero
} _attr_packed;


//
// The hierarchical node class
//
CHNode::CHNode()
: _pcsParent(NULL), _un32UserID(0)
{
  memset(_apcsChild, 0, sizeof(_apcsChild[0]) * _MAX_CHILD_);

  StcIdentity();
  DynSetIdentity();
  _tM.csTot.SetToIdentityMatrix();

  _tC.bStc = true;
  _tC.bDyn = true;
  _tC.bPsc = false;
  _tC.cs1.SetToIdentityMatrix();
  _tC.cs2.SetToIdentityMatrix();
}

CHNode::~CHNode() throw()
{
  DelAllChild();
}

void CHNode::SetParent(CHNode *pcsNewParent)
{
  // Temporary variables
  uint8_t un8Cnt;

  // If the new parent is not NULL
  if(pcsNewParent) {
    // Search for an empty node in the new parent
    for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
      if(pcsNewParent->_apcsChild[un8Cnt] == NULL) break;
    if(un8Cnt >= _MAX_CHILD_)
      VError(VERR_TooMany, _NFL_, "Parent node %p can't have any more child", pcsNewParent);

    // Add myself to the new parent
    pcsNewParent->_apcsChild[un8Cnt] = this;
  }

  // Remove myself from the old parent (if any)
  if(_pcsParent) {
    for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
    if(_pcsParent->_apcsChild[un8Cnt] == this) {
      _pcsParent->_apcsChild[un8Cnt] = NULL;
      break;
    }
  }

  // OK !
  _pcsParent = pcsNewParent;
}

CHNode *CHNode::AddChild(uint32_t un32UserID)
{
  // Temporary variables
  CHNode *pcsNew;
  uint8_t       un8Cnt;

  // Search for an empty node
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
    if(_apcsChild[un8Cnt] == NULL) break;
  if(un8Cnt >= _MAX_CHILD_)
    VError(VERR_TooMany, _NFL_, "A hierarchical node can't have more than %d children", _MAX_CHILD_);

  // Allocate memory for new node
  pcsNew = _nothrow_new CHNode;
  if(pcsNew == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new child node");

  // Set properties
  pcsNew->_pcsParent  = this;
  pcsNew->_un32UserID = un32UserID;
  _apcsChild[un8Cnt]  = pcsNew;

  // OK
  return(pcsNew);
}

void CHNode::DelChild(CHNode *pcsChild)
{
  // Temporary variables
  uint8_t un8Cnt;

  // Search and delete a specific child
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
  if(_apcsChild[un8Cnt] == pcsChild) {
    delete pcsChild;
    _apcsChild[un8Cnt] = NULL;
    break;
  }
}

void CHNode::DelAllChild()
{
  // Temporary variables
  uint8_t un8Cnt;

  // Search and delete all childdren
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
  if(_apcsChild[un8Cnt]) {
    delete _apcsChild[un8Cnt];
    _apcsChild[un8Cnt] = NULL;
  }
}

CHNode *CHNode::FindChild(uint32_t un32UserID, bool bRecursiveSubChild) const
{
  // Temporary variables
  uint8_t       un8Cnt;
  CHNode *pcsTmp;

  // Check
  if(un32UserID == 0) return(NULL);

  // Search recursively
  for(un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++) {
    if(_apcsChild[un8Cnt]) {
      if(_apcsChild[un8Cnt]->_un32UserID == un32UserID) return(_apcsChild[un8Cnt]);
      if(bRecursiveSubChild) {
        pcsTmp = _apcsChild[un8Cnt]->FindChild(un32UserID, true);
        if(pcsTmp) return(pcsTmp);
      }
    }
  }
  return(NULL);
}

void CHNode::UpdateMatrix(bool bUpdateChildren)
{
  // Update myself
  //       cs1 = Stc X Dyn
  //       cs2 = Dyn
  if(!(_tC.bStc || _tC.bDyn || _tC.bPsc)) {
    _tM.csTot.SetFromMatrix(_tC.cs1);
    if(_pcsParent) _tM.csTot.MultMatrixCCN(_pcsParent->_tM.csTot);
  }
  else if(_tC.bStc && (!(_tC.bDyn || _tC.bPsc))) {
    _tC.bStc = false;

    _tC.cs1.SetFromMatrix(_tM.csStc);
    _tC.cs1.MultMatrixCCN(_tC.cs2);

    _tM.csTot.SetFromMatrix(_tC.cs1);
    if(_pcsParent) _tM.csTot.MultMatrixCCN(_pcsParent->_tM.csTot);
  }
  else if((!_tC.bDyn) && _tC.bPsc) {
    _tC.bStc = _tC.bPsc = false;

    _tC.cs2.Translate(_tC.afPsc[0], _tC.afPsc[1], _tC.afPsc[2]);

    _tC.cs1.SetFromMatrix(_tM.csStc);
    _tC.cs1.MultMatrixCCN(_tC.cs2);

    _tM.csTot.SetFromMatrix(_tC.cs1);
    if(_pcsParent) _tM.csTot.MultMatrixCCN(_pcsParent->_tM.csTot);
  }
  else {
    _tC.bStc = _tC.bDyn = _tC.bPsc = false;

    _tC.cs2.SetToScallingMatrix(_tM.afScl[0], _tM.afScl[1], _tM.afScl[2]);
    _tC.cs2.Rotate(_tM.an32Rot[0], _tM.an32Rot[1], _tM.an32Rot[2]);
    _tC.cs2.Translate(_tM.afPos[0], _tM.afPos[1], _tM.afPos[2]);

    _tC.cs1.SetFromMatrix(_tM.csStc);
    _tC.cs1.MultMatrixCCN(_tC.cs2);

    _tM.csTot.SetFromMatrix(_tC.cs1);
    if(_pcsParent) _tM.csTot.MultMatrixCCN(_pcsParent->_tM.csTot);
  }
  // This is the old (un-cached) version :
  //   _tM.csTot.SetFromMatrix(_tM.csStc);
  //   _tM.csTot.Scale(_tM.afScl[0], _tM.afScl[1], _tM.afScl[2]);
  //   _tM.csTot.Rotate(_tM.an32Rot[0], _tM.an32Rot[1], _tM.an32Rot[2]);
  //   _tM.csTot.Translate(_tM.afPos[0], _tM.afPos[1], _tM.afPos[2]);
  //   if(_pcsParent) _tM.csTot.MultMatrixCCN(_pcsParent->_tM.csTot);

  // Update children
  if(bUpdateChildren) {
    for(uint8_t un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++)
      if(_apcsChild[un8Cnt]) _apcsChild[un8Cnt]->UpdateMatrix(true);
  }
}

void CHNode::SaveVHN(const CDataSink &sink) const
{
  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Make header
  _TVHNHeader tHeader;
  memset(&tHeader, 0, sizeof(tHeader));
  tHeader.acID[0]    = _VHN_ID1_;
  tHeader.acID[1]    = _VHN_ID2_;
  tHeader.acID[2]    = _VHN_ID3_;
  tHeader.un8Version = _VHN_VER_;

  // Write the header
  os.write(reinterpret_cast<const char*>(&tHeader), sizeof(tHeader));
  if(os.fail())
    VError(VERR_WriteStream, _NFL_, "Couldn't write VHN header data to stream \"%s\"", strmName);

  // Write all the data
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    _WriteNodeData_(os, this);
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write node data to stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

void CHNode::LoadVHN(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream    &is      = source.PStream();

  // Read the header
  _TVHNHeader tHeader;
  is.read(reinterpret_cast<char*>(&tHeader), sizeof(tHeader));
  if(is.fail())
    VError(VERR_ReadStream, _NFL_, "Couldn't read VHN header data from stream \"%s\"", strmName);

  // Check header
  if((tHeader.acID[0] != _VHN_ID1_) ||
     (tHeader.acID[1] != _VHN_ID2_) ||
     (tHeader.acID[2] != _VHN_ID3_))
    VError(VERR_InvalidStream, _NFL_, "\"%s\" is an invalid VHN stream", strmName);
  if(tHeader.un8Version != _VHN_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "\"%s\" has an incorrect version", strmName);
  if(tHeader.un32Reserved1 || tHeader.un32Reserved2)
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header data found in stream \"%s\"", strmName);

  // Read all the data to a temporary parent node
  CHNode csNode;
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    _ReadNodeData_(is, &csNode);
  }
  catch(ios_base::failure&) {
    csNode.DelAllChild();
    VError(VERR_ReadStream, _NFL_, "Couldn't read node data from stream \"%s\"", strmName);
  }
  catch(...) {
    csNode.DelAllChild();
    throw;
  }

  // Delete all my old childdren, if any
  DelAllChild();

  // Transfer hierarchy data
  _un32UserID = csNode._un32UserID;
  memcpy(_tM.afScl,              csNode._tM.afScl,              sizeof(float) * 3);
  memcpy(_tM.an32Rot,            csNode._tM.an32Rot,            sizeof(int32_t) * 3);
  memcpy(_tM.afPos,              csNode._tM.afPos,              sizeof(float) * 3);
  memcpy(_tM.csStc._aafMatrix, csNode._tM.csStc._aafMatrix, sizeof(float) * 16);
  memcpy(_apcsChild, csNode._apcsChild, sizeof(_apcsChild[0]) * _MAX_CHILD_);

  // Change parent
  for(uint8_t un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++) {
    if(csNode._apcsChild[un8Cnt]) csNode._apcsChild[un8Cnt]->_pcsParent = this;
  }

  // Make sure that the temporary parent node doesn't delete
  // its children when it goes out of this function scope
  memset(csNode._apcsChild, 0, sizeof(_apcsChild[0]) * _MAX_CHILD_);

  // Update matrix
  UpdateMatrix(true);
}

void CHNode::_WriteNodeData_(ostream &os, const CHNode *pcsHNode) const
{
  // Count available children
  uint8_t un8Tot = 0;
  for(uint8_t un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++) if(pcsHNode->_apcsChild[un8Cnt]) un8Tot++;

  // Write my own data
  os.write(reinterpret_cast<const char*>(&pcsHNode->_un32UserID),           sizeof(_un32UserID));
  os.write(reinterpret_cast<const char*>(pcsHNode->_tM.afScl),              sizeof(float) *  3);
  os.write(reinterpret_cast<const char*>(pcsHNode->_tM.an32Rot),            sizeof(int32_t) * 3);
  os.write(reinterpret_cast<const char*>(pcsHNode->_tM.afPos),              sizeof(float) *  3);
  os.write(reinterpret_cast<const char*>(pcsHNode->_tM.csStc._aafMatrix), sizeof(float) * 16);
  os.write(reinterpret_cast<const char*>(&un8Tot),                            sizeof(un8Tot));

  // Check
  if(!un8Tot) return;

  // Write all my children recursively
  for(uint8_t un8Cnt = 0; un8Cnt < _MAX_CHILD_; un8Cnt++) {
    if(pcsHNode->_apcsChild[un8Cnt])
      _WriteNodeData_(os, pcsHNode->_apcsChild[un8Cnt]);
  }
}

void CHNode::_ReadNodeData_(pistream &is, CHNode *pcsHNode)
{
  // Read my own data
  uint8_t un8Tot = 0;
  is.read(reinterpret_cast<char*>(&pcsHNode->_un32UserID),           sizeof(_un32UserID));
  is.read(reinterpret_cast<char*>(pcsHNode->_tM.afScl),              sizeof(float) * 3);
  is.read(reinterpret_cast<char*>(pcsHNode->_tM.an32Rot),            sizeof(int32_t) * 3);
  is.read(reinterpret_cast<char*>(pcsHNode->_tM.afPos),              sizeof(float) * 3);
  is.read(reinterpret_cast<char*>(pcsHNode->_tM.csStc._aafMatrix), sizeof(float) * 16);
  is.read(reinterpret_cast<char*>(&un8Tot),                            sizeof(un8Tot));

  // Check
  if(!un8Tot) return;
  if(un8Tot > _MAX_CHILD_)
    VError(VERR_NotSupported, _NFL_, "To many child nodes to be loaded");

  // Read and create all my children recursively
  for(uint8_t un8Cnt = 0; un8Cnt < un8Tot; un8Cnt++) {
    CHNode *pcsTNode = pcsHNode->AddChild();
    _ReadNodeData_(is, pcsTNode);
  }
}

} // namespace vge
