/** \file Matrix3D.h */
// ############################################################################################
//  Matrix3D.h
//  Definition & implementation of CMatrix3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___MATRIX_3D_H_
#define _3D_ENVIRONMENT___MATRIX_3D_H_

// VortexGE headers
#include "Miscellaneous3DStructs.h"
#include "Point3D.h"


namespace vge {

  //!
  //! \brief The 3D matrix class
  //!
  //! \api pretty stable
  class VGE_EXPORT CMatrix3D {
    public:
      // Display functions
      //!
      //! Dump the matrix contents to console.
      void DumpMatrix();

      // Transformation functions
      //!
      //! Transform the given vector (transform from local coordinate
      //! to world coordinate).
      inline _TVector3D &Transform(_TVector3D &tVVector3D) const ;
      //!
      //! Pre-dynamic transform the given vector (transform from local
      //! coordinate to world coordinate).
      inline _TVector3D &PreDynTransform(_TVector3D &tVVector3D) const
      { return(Transform(tVVector3D)); }
      //!
      //! Post-dynamic transform the given vector (transform from world
      //! coordinate to world coordinate).
      inline _TVector3D &PstDynTransform(_TVector3D &tVVector3D) const;
      //!
      //! Transform the given point (transform from local coordinate
      //! to world coordinate).
      inline CPoint3D &Transform(CPoint3D &csVPoint3D) const ;
      //!
      //! Pre-dynamic transform the given point (transform from local
      //! coordinate to world coordinate).
      inline CPoint3D &PreDynTransform(CPoint3D &csVPoint3D) const
      { return(Transform(csVPoint3D)); }
      //!
      //! Post-dynamic transform the given point (transform from world 
      //! coordinate to world coordinate).
      inline CPoint3D &PstDynTransform(CPoint3D &csVPoint3D) const ;

      // Set operations
      //!
      //! Make this matrix to a "look-at" matrix from the given camera
      //! position and target.
      void SetToLookAtMatrix(const float afCamPosXYZ[3], const float afCamTargetXYZ[3]);
      //!
      //! Make this matrix to a "look-at" matrix from the given camera
      //! position, target and the "up" direction.
      void SetToLookAtMatrix(const float afCamPosXYZ[3], const float afCamTargetXYZ[3], const float afUpVectXYZ[3]);
      //!
      //! Make this matrix to a shearing matrix from the given parameters.
      void SetToShearingMatrix(float fX, float fY);
      //!
      //! Make this matrix to a scalling matrix from the given parameters.
      void SetToScallingMatrix(float fX, float fY, float fZ);
      //!
      //! Make this matrix to a rotation matrix from the given parameters.
      void SetToRotatationMatrix(int32_t n32X, int32_t n32Y, int32_t n32Z);
      //!
      //! Make this matrix to a translation matrix from the given parameters.
      void SetToTranslationMatrix(float fX, float fY, float fZ);
      //!
      //! Make this matrix to an identity matrix.
      inline void SetToIdentityMatrix()
      { MakeIdentityMatrix(_aafMatrix); }
      //!
      //! Make the given array to an identity matrix.
      inline void MakeIdentityMatrix(float aafDstMatrix[4][4]) const;
      //!
      //! Make the given matrix to an identity matrix.
      inline void MakeIdentityMatrix(CMatrix3D &csVDstMatrix) const
      { MakeIdentityMatrix(csVDstMatrix._aafMatrix); }
      //!
      //! Copy the given array to this matrix.
      inline void SetFromMatrix(float const aafNewMatrix[4][4])
      { CopyMatrices(_aafMatrix, aafNewMatrix); }
      //!
      //! Copy the given matrix to this matrix.
      inline void SetFromMatrix(CMatrix3D &csVNewMatrix)
      { SetFromMatrix(csVNewMatrix._aafMatrix); }

      // Matrix inversion
      //!
      //! Make inverse matrix of the given source array in the given 
      //! destination array.
      void InvertMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const;
      //!
      //! Make inverse matrix of the given array.
      void InvertMatrix(float aafMatrix[4][4]) const;
      //!
      //! Make this matrix become its inverse matrix.
      inline void Invert()
      { InvertMatrix(_aafMatrix); }
    
      // Merged transformation operations
      //!
      //! Merge shear the current matrix using the given parameter.s
      void Shear(float fX, float fY);
      //!
      //! Merge scale the current matrix using the given parameters.
      void Scale(float fX, float fY, float fZ);
      //!
      //! Merge rotate the current matrix using the given parameters.
      void Rotate(int32_t n32X, int32_t n32Y, int32_t n32Z);
      //!
      //! Merge rotate (X rotation only) the current matrix using
      //! the given parameter.
      void RotateX(int32_t n32X);
      //!
      //! Merge rotate (Y rotation only) the current matrix using
      //! the given parameter.
      void RotateY(int32_t n32Y);
      //!
      //! Merge rotate (Z rotation only) the current matrix using
      //! the given parameter.
      void RotateZ(int32_t n32Z);
      //!
      //! Merge translate the current matrix using the given parameters.
      void Translate(float fX, float fY, float fZ);

      // Transpose operations
      //!
      //! Make transpose matrix of the given source array in the given 
      //! destination array.
      inline void TransposeMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const;
      //!
      //! Make transpose matrix of the given matrix in the given
      //! destination matrix.
      inline void TransposeMatrices(CMatrix3D &csVDstMatrix, CMatrix3D &csVSrcMatrix) const
      { TransposeMatrices(csVDstMatrix._aafMatrix, csVSrcMatrix._aafMatrix); }

      // Direct access to the matrix's array
      //!
      //! Direct access to the array holding the matrix data.
      inline float *GetMatrixArray() { return(&_aafMatrix[0][0]); }
      //!
      //! Direct access to the array holding the matrix data.
      inline const float *GetMatrixArray() const { return(&_aafMatrix[0][0]); }
    
      // Copy operations
      //!
      //! Copy the given source array to the given destination array.
      inline void CopyMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const;
      //!
      //! Copy the given source matrix to the given destination matrix.
      inline void CopyMatrices(CMatrix3D &csVDstMatrix, CMatrix3D &csVSrcMatrix) const { CopyMatrices(csVDstMatrix._aafMatrix, csVSrcMatrix._aafMatrix); }

      // Multiplication operations
      //!
      //! Multiply this matrix with the given array, in the form of:\n
      //! <span style="margin-left:2em;"><b>|This| = |This| x |New|</b></span>
      inline void MultMatrixCCN(const float aafNewMatrix[4][4])
      { MultMatricesDDS(_aafMatrix, aafNewMatrix); }
      //!
      //! Multiply this matrix with the given matrix, in the form of:\n
      //! <span style="margin-left:2em;"><b>|This| = |This| x |New|</b></span>
      inline void MultMatrixCCN(const CMatrix3D &csVNewMatrix)
      { MultMatrixCCN(csVNewMatrix._aafMatrix); }
      //!
      //! Multiply this matrix with the given array, in the form of:\n
      //! <span style="margin-left:2em;"><b>|This| = |New| x |This|</b></span>
      inline void MultMatrixCNC(const float aafNewMatrix[4][4])
      { MultMatricesDSD(_aafMatrix, aafNewMatrix); }
      //!
      //! Multiply this matrix with the given matrix, in the form of:\n
      //! <span style="margin-left:2em;"><b>|This| = |New| x |This|</b></span>
      inline void MultMatrixCNC(const CMatrix3D &csVNewMatrix)
      { MultMatrixCNC(csVNewMatrix._aafMatrix); }
      //!
      //! Multiply the arrays, in the form of:\n
      //! <span style="margin-left:2em;"><b>|Dst| = |Dst| x |Src|</b></span>
      inline void MultMatricesDDS(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const;
      //!
      //! Multiply the matrices, in the form of:\n
      //! <span style="margin-left:2em;"><b>|Dst| = |Dst| x |Src|</b></span>
      inline void MultMatricesDDS(CMatrix3D &csVDstMatrix, const CMatrix3D &csVSrcMatrix) const
      { MultMatricesDDS(csVDstMatrix._aafMatrix, csVSrcMatrix._aafMatrix); }
      //!
      //! Multiply the arrays, in the form of:\n
      //! <span style="margin-left:2em;"><b>|Dst| = |Src| x |Dst|</b></span>
      inline void MultMatricesDSD(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const;
      //!
      //! Multiply the matrices, in the form of:\n
      //! <span style="margin-left:2em;"><b>|Dst| = |Src| x |Dst|</b></span>
      inline void MultMatricesDSD(CMatrix3D &csVDstMatrix, const CMatrix3D &csVSrcMatrix) const
      { MultMatricesDSD(csVDstMatrix._aafMatrix, csVSrcMatrix._aafMatrix); }

      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

      // Friend classes
      friend class CHNode;
      friend class CWalkingCameraControl;

    protected:
      // Merged transformation operations
      inline void _MergeMatrix_(const float aafNewMatrix[4][4]) { MultMatrixCCN(aafNewMatrix); }
      inline void _MergeMatrices_(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const { MultMatricesDSD(aafDstMatrix, aafSrcMatrix); }

      // Data
      float _sse _aafMatrix[4][4]; // The combined matrix data
  };

  _TVector3D &CMatrix3D::Transform(_TVector3D &tVVector3D) const
  {
    tVVector3D.fXt = (tVVector3D.fX * _aafMatrix[0][0]) +
                     (tVVector3D.fY * _aafMatrix[1][0]) +
                     (tVVector3D.fZ * _aafMatrix[2][0]) +
                                      _aafMatrix[3][0];
    tVVector3D.fYt = (tVVector3D.fX * _aafMatrix[0][1]) +
                     (tVVector3D.fY * _aafMatrix[1][1]) +
                     (tVVector3D.fZ * _aafMatrix[2][1]) +
                                      _aafMatrix[3][1];
    tVVector3D.fZt = (tVVector3D.fX * _aafMatrix[0][2]) +
                     (tVVector3D.fY * _aafMatrix[1][2]) +
                     (tVVector3D.fZ * _aafMatrix[2][2]) +
                                      _aafMatrix[3][2];
    return(tVVector3D);
  }

  _TVector3D &CMatrix3D::PstDynTransform(_TVector3D &tVVector3D) const
  {
    float fXt = tVVector3D.fXt;
    float fYt = tVVector3D.fYt;
    float fZt = tVVector3D.fZt;

    tVVector3D.fXt = (fXt * _aafMatrix[0][0]) +
                     (fYt * _aafMatrix[1][0]) +
                     (fZt * _aafMatrix[2][0]) +
                            _aafMatrix[3][0];
    tVVector3D.fYt = (fXt * _aafMatrix[0][1]) +
                     (fYt * _aafMatrix[1][1]) +
                     (fZt * _aafMatrix[2][1]) +
                            _aafMatrix[3][1];
    tVVector3D.fZt = (fXt * _aafMatrix[0][2]) +
                     (fYt * _aafMatrix[1][2]) +
                     (fZt * _aafMatrix[2][2]) +
                            _aafMatrix[3][2];
    return(tVVector3D);
  }

  CPoint3D &CMatrix3D::Transform(CPoint3D &csVPoint3D) const
  {
    csVPoint3D._tWorld.fX = (csVPoint3D._tLocal.fX * _aafMatrix[0][0]) +
                            (csVPoint3D._tLocal.fY * _aafMatrix[1][0]) +
                            (csVPoint3D._tLocal.fZ * _aafMatrix[2][0]) +
                                                     _aafMatrix[3][0];
    csVPoint3D._tWorld.fY = (csVPoint3D._tLocal.fX * _aafMatrix[0][1]) +
                            (csVPoint3D._tLocal.fY * _aafMatrix[1][1]) +
                            (csVPoint3D._tLocal.fZ * _aafMatrix[2][1]) +
                                                     _aafMatrix[3][1];
    csVPoint3D._tWorld.fZ = (csVPoint3D._tLocal.fX * _aafMatrix[0][2]) +
                            (csVPoint3D._tLocal.fY * _aafMatrix[1][2]) +
                            (csVPoint3D._tLocal.fZ * _aafMatrix[2][2]) +
                                                     _aafMatrix[3][2];
    return(csVPoint3D);
  }

  CPoint3D &CMatrix3D::PstDynTransform(CPoint3D &csVPoint3D) const
  {
    float fXw = csVPoint3D._tWorld.fX;
    float fYw = csVPoint3D._tWorld.fY;
    float fZw = csVPoint3D._tWorld.fZ;

    csVPoint3D._tWorld.fX = (fXw * _aafMatrix[0][0]) +
                            (fYw * _aafMatrix[1][0]) +
                            (fZw * _aafMatrix[2][0]) +
                                   _aafMatrix[3][0];
    csVPoint3D._tWorld.fY = (fXw * _aafMatrix[0][1]) +
                            (fYw * _aafMatrix[1][1]) +
                            (fZw * _aafMatrix[2][1]) +
                                   _aafMatrix[3][1];
    csVPoint3D._tWorld.fZ = (fXw * _aafMatrix[0][2]) +
                            (fYw * _aafMatrix[1][2]) +
                            (fZw * _aafMatrix[2][2]) +
                                   _aafMatrix[3][2];
    return(csVPoint3D);
  }

  void CMatrix3D::MakeIdentityMatrix(float aafDstMatrix[4][4]) const
  {
    aafDstMatrix[0][0] = 1; aafDstMatrix[0][1] = 0; aafDstMatrix[0][2] = 0; aafDstMatrix[0][3] = 0;
    aafDstMatrix[1][0] = 0; aafDstMatrix[1][1] = 1; aafDstMatrix[1][2] = 0; aafDstMatrix[1][3] = 0;
    aafDstMatrix[2][0] = 0; aafDstMatrix[2][1] = 0; aafDstMatrix[2][2] = 1; aafDstMatrix[2][3] = 0;
    aafDstMatrix[3][0] = 0; aafDstMatrix[3][1] = 0; aafDstMatrix[3][2] = 0; aafDstMatrix[3][3] = 1;
  }

  void CMatrix3D::TransposeMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const
  {
    aafDstMatrix[0][0] = aafSrcMatrix[0][0]; aafDstMatrix[0][1] = aafSrcMatrix[1][0]; aafDstMatrix[0][2] = aafSrcMatrix[2][0]; aafDstMatrix[0][3] = aafSrcMatrix[3][0];
    aafDstMatrix[1][0] = aafSrcMatrix[0][1]; aafDstMatrix[1][1] = aafSrcMatrix[1][1]; aafDstMatrix[1][2] = aafSrcMatrix[2][1]; aafDstMatrix[1][3] = aafSrcMatrix[3][1];
    aafDstMatrix[2][0] = aafSrcMatrix[0][2]; aafDstMatrix[2][1] = aafSrcMatrix[1][2]; aafDstMatrix[2][2] = aafSrcMatrix[2][2]; aafDstMatrix[2][3] = aafSrcMatrix[3][2];
    aafDstMatrix[3][0] = aafSrcMatrix[0][3]; aafDstMatrix[3][1] = aafSrcMatrix[1][3]; aafDstMatrix[3][2] = aafSrcMatrix[2][3]; aafDstMatrix[3][3] = aafSrcMatrix[3][3];
  }

  void CMatrix3D::CopyMatrices(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const
  {
    aafDstMatrix[0][0] = aafSrcMatrix[0][0]; aafDstMatrix[0][1] = aafSrcMatrix[0][1]; aafDstMatrix[0][2] = aafSrcMatrix[0][2]; aafDstMatrix[0][3] = aafSrcMatrix[0][3];
    aafDstMatrix[1][0] = aafSrcMatrix[1][0]; aafDstMatrix[1][1] = aafSrcMatrix[1][1]; aafDstMatrix[1][2] = aafSrcMatrix[1][2]; aafDstMatrix[1][3] = aafSrcMatrix[1][3];
    aafDstMatrix[2][0] = aafSrcMatrix[2][0]; aafDstMatrix[2][1] = aafSrcMatrix[2][1]; aafDstMatrix[2][2] = aafSrcMatrix[2][2]; aafDstMatrix[2][3] = aafSrcMatrix[2][3];
    aafDstMatrix[3][0] = aafSrcMatrix[3][0]; aafDstMatrix[3][1] = aafSrcMatrix[3][1]; aafDstMatrix[3][2] = aafSrcMatrix[3][2]; aafDstMatrix[3][3] = aafSrcMatrix[3][3];
  }

  void CMatrix3D::MultMatricesDDS(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const
  {
    float aafTmp[4][4];

    for(size_t i = 0; i < 4; ++i)
      for(size_t j = 0; j < 4; ++j)
        aafTmp[i][j] = (aafDstMatrix[i][0] * aafSrcMatrix[0][j]) +
                       (aafDstMatrix[i][1] * aafSrcMatrix[1][j]) +
                       (aafDstMatrix[i][2] * aafSrcMatrix[2][j]) +
                       (aafDstMatrix[i][3] * aafSrcMatrix[3][j]);
    CopyMatrices(aafDstMatrix, aafTmp);
  }

  void CMatrix3D::MultMatricesDSD(float aafDstMatrix[4][4], const float aafSrcMatrix[4][4]) const
  {
    float aafTmp[4][4];

    for(size_t i = 0; i < 4; ++i)
      for(size_t j = 0; j < 4; ++j)
        aafTmp[i][j] = (aafSrcMatrix[i][0] * aafDstMatrix[0][j]) +
                       (aafSrcMatrix[i][1] * aafDstMatrix[1][j]) +
                       (aafSrcMatrix[i][2] * aafDstMatrix[2][j]) +
                       (aafSrcMatrix[i][3] * aafDstMatrix[3][j]);
    CopyMatrices(aafDstMatrix, aafTmp);
  }

} // namespace vge
#endif
