/** \file Point3D.h */
// ############################################################################################
//  Point3D.h
//  Definition & implementation of CPoint3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___POINT_3D_H_
#define _3D_ENVIRONMENT___POINT_3D_H_

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/MemoryOperations.h"


namespace vge {

  //!
  //! \brief Triplet float XYZ sturcture
  //!
  //! \api pretty stable
  struct TFloatXYZ {
    float x, y, z;
  
    inline TFloatXYZ(float x_ = 0, float y_ = 0, float z_ = 0)
    : x(x_), y(y_), z(z_)
    {}
    
    inline TFloatXYZ &operator+=(const TFloatXYZ &val)
    { x += val.x; y += val.y; z += val.z; return(*this); }
  
    inline TFloatXYZ &operator-=(const TFloatXYZ &val)
    { x -= val.x; y -= val.y; z -= val.z; return(*this); }

    inline TFloatXYZ operator+(const TFloatXYZ &val)
    { TFloatXYZ tmp; tmp.x = x + val.x; tmp.y = y + val.y; tmp.z = z + val.z; return(tmp); }

    inline TFloatXYZ operator-(const TFloatXYZ &val)
    { TFloatXYZ tmp; tmp.x = x - val.x; tmp.y = y - val.y; tmp.z = z - val.z; return(tmp); }
  };


  //!
  //! \internal
  //! \brief Triplet float XYZ structure (padded and alligned to
  //!        128-bit if SSE is used)
  //!
  //! \api pretty stable
  struct _TFXYZ {
    float fX, fY, fZ;
#ifdef _ALIGN_128_BITS_
    float _fdummy_to_make_this_structure_128_bits_long_;
#endif
  };

  //! \internal
  //! \brief Container for the Z-clipped world coordinates
  //!
  //! \api pretty stable
  struct _TZClipXYZ {
    float afX[5];
    float afY[5];
    float afZ[5];
  };


  //!
  //! \brief The 3D point class
  //!
  //! \api pretty stable
  class VGE_EXPORT CPoint3D {
    public:
      //!
      //! Initialize all the local XYZ components with zero.
      inline CPoint3D();
      //!
      //! Initialize all the local XYZ components with the given value.
      inline explicit CPoint3D(float fLocalXYZ);
      //!
      //! Initialize all the local XYZ components with the given XYZ values.
      inline CPoint3D(float fLocalX, float fLocalY, float fLocalZ);
      //!
      //! Initialize all the local XYZ components with the given _TFXYZ instance.
      inline explicit CPoint3D(const _TFXYZ &tLocal);
      //!
      //! Initialize all the local XYZ components with the given TFloatXYZ instance.
      inline explicit CPoint3D(const TFloatXYZ &tLocal);
      //!
      //! Initialize all the local XYZ components with the given CPoint3D instance.
      inline CPoint3D(const CPoint3D &csVPoint3D);
      //!
      //! Initialize all the local XYZ components with the given CPoint3D instance.
      inline CPoint3D(const CPoint3D *pcsVPoint3D);

      // Data functions
      //!
      //! Set the X component of the local coordinate.
      inline void SetLocalX(float fNewLocalX) { _tLocal.fX = fNewLocalX; }
      //!
      //! Set the Y component of the local coordinate.
      inline void SetLocalY(float fNewLocalY) { _tLocal.fY = fNewLocalY; }
      //!
      //! Set the Z component of the local coordinate.
      inline void SetLocalZ(float fNewLocalZ) { _tLocal.fZ = fNewLocalZ; }
      //!
      //! Returns the X component of the local coordinate.
      inline float GetLocalX() const { return(_tLocal.fX); }
      //!
      //! Returns the Y component of the local coordinate.
      inline float GetLocalY() const { return(_tLocal.fY); }
      //!
      //! Returns the Z component of the local coordinate.
      inline float GetLocalZ() const { return(_tLocal.fZ); }
      //!
      //! Returns the X component of the world (transformed) coordinate.
      inline float GetWorldX() const { return(_tWorld.fX); }
      //!
      //! Returns the Y component of the world (transformed) coordinate.
      inline float GetWorldY() const { return(_tWorld.fY); }
      //!
      //! Returns the Z component of the world (transformed) coordinate.
      inline float GetWorldZ() const { return(_tWorld.fZ); }
      //!
      //! Returns the X component of the average normal (normal value
      //! which was calculated by taking the average of all panels'
      //! normals which share the same vertices).
      inline float GetAvgNX() const { return(_tAvgN.fX); }
      //!
      //! Returns the Y component of the average normal (normal value
      //! which was calculated by taking the average of all panels'
      //! normals which share the same vertices).
      inline float GetAvgNY() const { return(_tAvgN.fY); }
      //!
      //! Returns the Z component of the average normal (normal value
      //! which was calculated by taking the average of all panels'
      //! normals which share the same vertices).
      inline float GetAvgNZ() const { return(_tAvgN.fZ); }

      // Math functions
      //!
      //! Calculate the magnitude of this point.
      inline float Mag() const;
      //!
      //! Calculate dot product between this point and the give point.
      //! Both the points must be unit points (their magnitudes are 1).
      inline float DotUnit(const CPoint3D &csVPoint3D) const;
      //!
      //! Calculate dot product between this point and the give point.
      //! The points can have any magnitude.
      inline float DotNonUnit(const CPoint3D &csVPoint3D) const;

      // Overloaded operators
      inline bool      operator==(const CPoint3D &csVPoint3D) const;
      inline bool      operator!=(const CPoint3D &csVPoint3D) const;
      inline CPoint3D  operator+ (const CPoint3D &csVPoint3D);
      inline CPoint3D  operator- (const CPoint3D &csVPoint3D);
      inline CPoint3D  operator* (const CPoint3D &csVPoint3D);
      inline CPoint3D  operator/ (const CPoint3D &csVPoint3D);
      inline CPoint3D &operator+=(const CPoint3D &csVPoint3D);
      inline CPoint3D &operator-=(const CPoint3D &csVPoint3D);
      inline CPoint3D &operator*=(const CPoint3D &csVPoint3D);
      inline CPoint3D &operator/=(const CPoint3D &csVPoint3D);
      inline CPoint3D  operator+ (float fValue);
      inline CPoint3D  operator- (float fValue);
      inline CPoint3D  operator* (float fValue);
      inline CPoint3D  operator/ (float fValue);
      inline CPoint3D &operator+=(float fValue);
      inline CPoint3D &operator-=(float fValue);
      inline CPoint3D &operator*=(float fValue);
      inline CPoint3D &operator/=(float fValue);
      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

      // Friend classes
      friend class CPanel3D;
      friend class CPanel3DBuffer;
      friend class CMatrix3D;
      friend class CSprite2D;

    protected:
      _TFXYZ _sse _tLocal; // The local (untransformed) XYZ
      _TFXYZ _sse _tWorld; // The world (transformed) XYZ
      _TFXYZ _sse _tAvgN;  // The local (untransformed) vertex normal
  };

  CPoint3D::CPoint3D()
  {
    _tLocal.fX = _tLocal.fY = _tLocal.fZ = 0;
  }

  CPoint3D::CPoint3D(float fLocalXYZ)
  {
    _tLocal.fX = _tLocal.fY = _tLocal.fZ = fLocalXYZ;
  }

  CPoint3D::CPoint3D(float fLocalX, float fLocalY, float fLocalZ)
  {
    _tLocal.fX = fLocalX;
    _tLocal.fY = fLocalY;
    _tLocal.fZ = fLocalZ;
  }

  CPoint3D::CPoint3D(const _TFXYZ &tLocal)
  {
    _tLocal.fX = tLocal.fX;
    _tLocal.fY = tLocal.fY;
    _tLocal.fZ = tLocal.fZ;
  }

  CPoint3D::CPoint3D(const TFloatXYZ &tLocal)
  {
    _tLocal.fX = tLocal.x;
    _tLocal.fY = tLocal.y;
    _tLocal.fZ = tLocal.z;
  }

  CPoint3D::CPoint3D(const CPoint3D &csVPoint3D)
  {
    _tLocal.fX = csVPoint3D._tLocal.fX;
    _tLocal.fY = csVPoint3D._tLocal.fY;
    _tLocal.fZ = csVPoint3D._tLocal.fZ;
  }

  CPoint3D::CPoint3D(const CPoint3D *pcsVPoint3D)
  {
    _tLocal.fX = pcsVPoint3D->_tLocal.fX;
    _tLocal.fY = pcsVPoint3D->_tLocal.fY;
    _tLocal.fZ = pcsVPoint3D->_tLocal.fZ;
  }

  float CPoint3D::Mag() const
  {
    return(sqrt(_tLocal.fX * _tLocal.fX +
                _tLocal.fY * _tLocal.fY +
                _tLocal.fZ * _tLocal.fZ));
  }

  float CPoint3D::DotUnit(const CPoint3D &csVPoint3D) const
  {
    return(_tLocal.fX * csVPoint3D._tLocal.fX +
           _tLocal.fY * csVPoint3D._tLocal.fY +
           _tLocal.fZ * csVPoint3D._tLocal.fZ);
  }

  float CPoint3D::DotNonUnit(const CPoint3D &csVPoint3D) const
  {
    return((_tLocal.fX * csVPoint3D._tLocal.fX +
            _tLocal.fY * csVPoint3D._tLocal.fY +
            _tLocal.fZ * csVPoint3D._tLocal.fZ) / (Mag() * csVPoint3D.Mag()));
  }

  bool CPoint3D::operator==(const CPoint3D &csVPoint3D) const
  {
    return((_tLocal.fX == csVPoint3D._tLocal.fX) &&
           (_tLocal.fY == csVPoint3D._tLocal.fY) &&
           (_tLocal.fZ == csVPoint3D._tLocal.fZ));
  }

  bool CPoint3D::operator!=(const CPoint3D &csVPoint3D) const
  {
    return((_tLocal.fX != csVPoint3D._tLocal.fX) ||
           (_tLocal.fY != csVPoint3D._tLocal.fY) ||
           (_tLocal.fZ != csVPoint3D._tLocal.fZ));
  }

  CPoint3D CPoint3D::operator+(const CPoint3D &csVPoint3D)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX + csVPoint3D._tLocal.fX;
    clsTmp._tLocal.fY = _tLocal.fY + csVPoint3D._tLocal.fY;
    clsTmp._tLocal.fZ = _tLocal.fZ + csVPoint3D._tLocal.fZ;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator-(const CPoint3D &csVPoint3D)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX - csVPoint3D._tLocal.fX;
    clsTmp._tLocal.fY = _tLocal.fY - csVPoint3D._tLocal.fY;
    clsTmp._tLocal.fZ = _tLocal.fZ - csVPoint3D._tLocal.fZ;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator*(const CPoint3D &csVPoint3D)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX * csVPoint3D._tLocal.fX;
    clsTmp._tLocal.fY = _tLocal.fY * csVPoint3D._tLocal.fY;
    clsTmp._tLocal.fZ = _tLocal.fZ * csVPoint3D._tLocal.fZ;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator/(const CPoint3D &csVPoint3D)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX / csVPoint3D._tLocal.fX;
    clsTmp._tLocal.fY = _tLocal.fY / csVPoint3D._tLocal.fY;
    clsTmp._tLocal.fZ = _tLocal.fZ / csVPoint3D._tLocal.fZ;
    return(clsTmp);
  }

  CPoint3D &CPoint3D::operator+=(const CPoint3D &csVPoint3D)
  {
    _tLocal.fX += csVPoint3D._tLocal.fX;
    _tLocal.fY += csVPoint3D._tLocal.fY;
    _tLocal.fZ += csVPoint3D._tLocal.fZ;
    return(*this);
  }

  CPoint3D &CPoint3D::operator-=(const CPoint3D &csVPoint3D)
  {
    _tLocal.fX -= csVPoint3D._tLocal.fX;
    _tLocal.fY -= csVPoint3D._tLocal.fY;
    _tLocal.fZ -= csVPoint3D._tLocal.fZ;
    return(*this);
  }

  CPoint3D &CPoint3D::operator*=(const CPoint3D &csVPoint3D)
  {
    _tLocal.fX *= csVPoint3D._tLocal.fX;
    _tLocal.fY *= csVPoint3D._tLocal.fY;
    _tLocal.fZ *= csVPoint3D._tLocal.fZ;
    return(*this);
  }

  CPoint3D &CPoint3D::operator/=(const CPoint3D &csVPoint3D)
  {
    _tLocal.fX /= csVPoint3D._tLocal.fX;
    _tLocal.fY /= csVPoint3D._tLocal.fY;
    _tLocal.fZ /= csVPoint3D._tLocal.fZ;
    return(*this);
  }

  CPoint3D CPoint3D::operator+(float fValue)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX + fValue;
    clsTmp._tLocal.fY = _tLocal.fY + fValue;
    clsTmp._tLocal.fZ = _tLocal.fZ + fValue;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator-(float fValue)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX - fValue;
    clsTmp._tLocal.fY = _tLocal.fY - fValue;
    clsTmp._tLocal.fZ = _tLocal.fZ - fValue;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator*(float fValue)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX * fValue;
    clsTmp._tLocal.fY = _tLocal.fY * fValue;
    clsTmp._tLocal.fZ = _tLocal.fZ * fValue;
    return(clsTmp);
  }

  CPoint3D CPoint3D::operator/(float fValue)
  {
    CPoint3D clsTmp;
    clsTmp._tLocal.fX = _tLocal.fX / fValue;
    clsTmp._tLocal.fY = _tLocal.fY / fValue;
    clsTmp._tLocal.fZ = _tLocal.fZ / fValue;
    return(clsTmp);
  }

  CPoint3D &CPoint3D::operator+=(float fValue)
  {
    _tLocal.fX += fValue;
    _tLocal.fY += fValue;
    _tLocal.fZ += fValue;
    return(*this);
  }

  CPoint3D &CPoint3D::operator-=(float fValue)
  {
    _tLocal.fX -= fValue;
    _tLocal.fY -= fValue;
    _tLocal.fZ -= fValue;
    return(*this);
  }

  CPoint3D &CPoint3D::operator*=(float fValue)
  {
    _tLocal.fX *= fValue;
    _tLocal.fY *= fValue;
    _tLocal.fZ *= fValue;
    return(*this);
  }

  CPoint3D &CPoint3D::operator/=(float fValue)
  {
    _tLocal.fX /= fValue;
    _tLocal.fY /= fValue;
    _tLocal.fZ /= fValue;
    return(*this);
  }


  //
  // Vertex helper functions
  //
  //!
  //! Search pclsList to determine uniquness of vertex csVPoint3D
  //!
  //! \api pretty stable
  extern VGE_EXPORT bool UniqueVert(const CPoint3D &csVPoint3D, const CPoint3D *pcsList, uint32_t un32Range);
  //!
  //! Search pclsList to locate the index of a vertex that matches csVPoint3D
  //!
  //! \api pretty stable
  extern VGE_EXPORT uint32_t GetVertIndex(const CPoint3D &csVPoint3D, const CPoint3D *pcsList, uint32_t un32Range);


  // Internal inline functions implementation related to
  // Sutherland-Hodgman algorithm
  namespace xprv {

    inline void _SHCopy(_TZClipXYZ &tC, uint32_t un32C, CPoint3D **ppcsV, uint32_t un32V)
    {
      tC.afX[un32C] = ppcsV[un32V]->GetWorldX();
      tC.afY[un32C] = ppcsV[un32V]->GetWorldY();
      tC.afZ[un32C] = ppcsV[un32V]->GetWorldZ();
    }

    inline void _SHClip(_TZClipXYZ &tC, uint32_t un32C, CPoint3D **ppcsV, uint32_t un32S, uint32_t un32E, float fT, uint32_t un32Z)
    {
      tC.afX[un32C] = ppcsV[un32S]->GetWorldX() + (ppcsV[un32E]->GetWorldX() - ppcsV[un32S]->GetWorldX()) * fT;
      tC.afY[un32C] = ppcsV[un32S]->GetWorldY() + (ppcsV[un32E]->GetWorldY() - ppcsV[un32S]->GetWorldY()) * fT;
      tC.afZ[un32C] = float(un32Z);
    }

  } // namespace xprv

} // namespace vge
#endif
