/** \file AudioEnvironment.h */
// ############################################################################################
//  AudioEnvironment.h
//  Definition of CAudioEnv interface class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _AUDIO_ENVIRONMENT___AUDIO_ENVIRONMENT_H_
#define _AUDIO_ENVIRONMENT___AUDIO_ENVIRONMENT_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"
#include "LowLevelSoundBufferFunctions.h"


namespace vge {

  // Foward reference from "lib/AudioEnvironment/SoundBuffer.h"
  class CSoundBuffer;

  // Foward reference from "lib/AudioEnvironment/MidiPlayer.h"
  class CMidiPlayer;


  //!
  //! \brief Audio environment interface class
  //!
  //! Affecting environment variables :
  //!   - VGE_AUDIO_SYSTEM
  //!   - VGE_AUDIO_DEVICE
  //!
  //! See the section "Environment Variables" form the main doumentation (not the
  //! API documentation) for more details.
  //!
  //! \api pretty stable
  class VGE_EXPORT CAudioEnv : VCLASS_UNCOPYABLE {
    public:
      CAudioEnv();
      virtual ~CAudioEnv() throw();

      // Auto instantiation functions
      //!
      //! Automatically search for usable audio system and return the
      //! corresponding CAudioEnv instance.
      static CAudioEnv& AutoDetect();

      // Quick initialization functions
      //!
      //! Quick initialize the audio environment instance, the same as calling
      //! QueryDeviceName(), OpenDevice(), InitializeDevice() and ActivateDevice()
      //! alltogether.
      //! \param[in] pcDeviceName Audio device to be opened, better just use the default value so that it will autodetect.
      virtual void DoQuickInit(const char *pcDeviceName = NULL) = 0;

      // Initialization functions
      //!
      //! Query for usable defice name.
      //! \param[out] pcResult   Name of the found audio device. If set to NULL,
      //!                        this function will just check if the audio
      //!                        system is supported in this system or not (if
      //!                        not supported it will throw an exception).
      //! \param[in]  bufferSize Size of the buffer pointed by the 'pcResult'
      //!                        parameter.
      //!
      //! By default, this function expect the 'pcResult' parameter to point to
      //! a buffer of which size is at least <b>PATH_MAX</b> bytes.
      //!
      //! \return The string length of the found audio device name (including the
      //!         terminating NULL character). If it is larger than the value in
      //!         the 'bufferSize' parameter, the user will need to enlarge the
      //!         buffer (buffer of <b>PATH_MAX</b> bytes long should be enough).
      virtual size_t QueryDeviceName(char *pcResult, size_t bufferSize = PATH_MAX) = 0;
      //!
      //! Open the audio device.
      //! \param[in] pcDeviceName Audio device to be opened, better just use the default value.
      virtual void OpenDevice(const char *pcDeviceName) = 0;
      //!
      //! Initialize the audio device using the given settings.
      //! \param[in]  ptPrefered Wanted audio settings.
      //! \param[out] ptRealized Realized audio settings.
      virtual void InitializeDevice(const TAudioSettings *ptPrefered = NULL, TAudioSettings *ptRealized = NULL) = 0;
      //!
      //! Activate the audio device.
      virtual void ActivateDevice()   = 0;
      //!
      //! Deactivate the audio device.
      virtual void DeactivateDevice() = 0;
      //!
      //! Get the audio device settings.
      virtual void GetAudioSpec(uint32_t *un32Channels, uint32_t *un32Bits, uint32_t *un32Rates, ESoundBufferFormat *eFormat, uint32_t *un32FragSize) const = 0;
      //!
      //! Attach the given MIDI player to this audio device.
      virtual void SetMidiPlayer(CMidiPlayer *pcsPlayer) = 0;

      // Sound functions
      //!
      //! Check if this audio device is usable.
      virtual bool IsAvailable() const = 0;
      //!
      //! Suspend the audio device.
      virtual void SuspendDevice() = 0;
      //!
      //! Resume the audio device.
      virtual void ResumeDevice() = 0;
      //!
      //! Must be called periodically to mix the sound buffers and output the sound.
      virtual void UpdateBuffer() = 0;

      // Sound buffer functions
      //!
      //! Create a new sound buffer to be palyed with this audio device.
      virtual CSoundBuffer *NewSoundBuffer() = 0;
  };

} // namespace vge
#endif
