// ############################################################################################
//  MessageCatalog.cc
//  Implementation of CMessageCatalog class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/Random.h"
#include "../HighLevelClass/ApplicationObject.h"
#include "../System/SystemUtilities.h"
#include "MessageCatalog.h"

// C++ headers
#include <fstream>
#include <sstream>


namespace vge {

//
// Constants used in all message catalog (*.cat) files
//
#define _MCAT_FILE_VER_    100
#define _MCAT_INDEX_MAGIC_ "VGE-MCAT-INDEX"
#define _MCAT_TRANS_MAGIC_ "VGE-MCAT-TRANS"


//
// Prototype of local functions for parsing and converting strings
//
static const char *_parseNumber(string &dst, const char *i, const char *e, char prefix, u_char base);
static string _convertString(const string &src);


//
// Message catalog class
//
void CMessageCatalog::LoadCatalog(const string &pathToMessageCatalog, const char *messagesLocale, bool appendMode)
{
  _TKeys                 keys;
  vector<_TTranslations> translations;

  const string pth = (pathToMessageCatalog[pathToMessageCatalog.size() - 1] == '/') ?
                      pathToMessageCatalog : (pathToMessageCatalog + '/');

  VLoadMessageCatalog(pth, keys, translations);
  VOptimizeMessageCatalog(_theCatalog, keys, translations, appendMode);
  SelectLocale(messagesLocale);
}


//
// _TKeys structure
//
uint32_t _TKeys ::GenerateNewNumericID()
{
  CRandomCoreUInt32 rnGen(time(NULL));
  for(;;) {
    uint32_t nid = rnGen.Generate();
    bool     unq = true;
    for(size_t i = 0; i < msg.size(); ++i)
      if(msg[i].id == nid) { unq = false; break; }
    if(unq) return(nid);
  }
  return(0);
}


//
// _CMsgCat class
//
bool _CMsgCat::_SelectLocale(const string &messagesLocale)
{
  _CMTable::const_iterator it = _tt.find(messagesLocale);
  if(it != _tt.end()) {
    _ma = &it->second;
    _cl = messagesLocale;
    return(true);
  }
  _ma = NULL;
  _cl = "";
  return(false);
}

void _CMsgCat::_GenerateLocaleNames(vector<string> &lcNames, const string &lcname)
{
  struct _SH {

    inline static string _toLower(const string &txt)
    {
      string tmp;
      for(string::size_type i = 0; i < txt.length(); ++i)
        tmp += tolower(txt[i]);
      return tmp;
    }

    inline static string _toUpper(const string &txt)
    {
      string tmp;
      for(string::size_type i = 0; i < txt.length(); ++i)
        tmp += toupper(txt[i]);
      return tmp;
    }

    inline static string _removeDash(const string &txt)
    {
      string tmp;
      for(string::size_type i = 0; i < txt.length(); ++i)
        if(txt[i] != '-') tmp += txt[i];
      return tmp;
    }

    static void _genEA(vector<string> &vec, const string &txt)
    {
      const string txr = _SH::_removeDash(txt);
      _addUnique(vec, txt);
      _addUnique(vec, txr);
      _addUnique(vec, _SH::_toLower(txt));
      _addUnique(vec, _SH::_toLower(txr));
      _addUnique(vec, _SH::_toUpper(txt));
      _addUnique(vec, _SH::_toUpper(txr));
    }

    static void _addUnique(vector<string> &vec, const string &txt)
    {
      for(vector<string>::size_type i = 0; i < vec.size(); ++i)
         if(vec[i] == txt) return;
      vec.push_back(txt);
    }

  };

  // Find the tokens and generate (almost) all possible locale names
  const string::size_type np = string::npos;
  const string::size_type ps = lcname.find_first_of('_');
  const string::size_type pd = lcname.find_first_of('.');
  const string::size_type pa = lcname.find_first_of('@');
  // Is it in the form of something like "en_IE.iso885915@euro" ?
  if(ps!=np && pd!=np && pa!=np && ps<pd && pd<pa) {
    vector<string> en, an;
    const string ln = lcname.substr(0, ps);
    const string cn = lcname.substr(ps+1, pd-ps-1);
    _SH::_genEA(en, lcname.substr(pd+1, pa-pd-1));
    _SH::_genEA(an, lcname.substr(pa+1, np));
    // Generate (almost) all possible locale names
    lcNames.push_back(lcname);
    for(vector<string>::size_type i = 0; i < en.size(); ++i) {
      for(vector<string>::size_type j = 0; j < an.size(); ++j) {
        _SH::_addUnique(lcNames, ln + "_" + cn + "." + en[i] + "@" + an[i]);
      }
    }
    for(vector<string>::size_type i = 0; i < en.size(); ++i) {
      _SH::_addUnique(lcNames, ln + "_" + cn + "." + en[i]);
    }
    _SH::_addUnique(lcNames, ln + "_" + cn);
    _SH::_addUnique(lcNames, ln);
  }
  // Is it in the form of something like "en_IE.iso885915" ?
  else if(ps!=np && pd!=np && pa==np && ps<pd) {
    vector<string> en;
    const string ln = lcname.substr(0, ps);
    const string cn = lcname.substr(ps+1, pd-ps-1);
    _SH::_genEA(en, lcname.substr(pd+1, pa-pd-1));
    // Generate (almost) all possible locale names
    lcNames.push_back(lcname);
    for(vector<string>::size_type i = 0; i < en.size(); ++i) {
      _SH::_addUnique(lcNames, ln + "_" + cn + "." + en[i]);
    }
    _SH::_addUnique(lcNames, ln + "_" + cn);
    _SH::_addUnique(lcNames, ln);    }
  // Is it in the form of something like "en_IE@euro" ?
  else if(ps!=np && pd==np && pa!=np && ps<pa) {
    vector<string> an;
    const string ln = lcname.substr(0, ps);
    const string cn = lcname.substr(ps+1, pa-ps-1);
    _SH::_genEA(an, lcname.substr(pa+1, np));
    // Generate (almost) all possible locale names
    lcNames.push_back(lcname);
    for(vector<string>::size_type i = 0; i < an.size(); ++i) {
      _SH::_addUnique(lcNames, ln + "_" + cn + "@" + an[i]);
    }
    _SH::_addUnique(lcNames, ln + "_" + cn);
    _SH::_addUnique(lcNames, ln);
  }
  // Is it in the form of something like "en_IE" ?
  else if(ps!=np && pd==np && pa==np) {
    const string ln = lcname.substr(0, ps);
    const string cn = lcname.substr(ps+1, np);
    // Generate (almost) all possible locale names
    lcNames.push_back(ln + "_" + cn);
    lcNames.push_back(ln);
  }
  // It should be in the form of something like "en" (or it is an invalid locale name)
  else {
    const string ln = lcname;
    const string ll = _SH::_toLower(lcname);
    const string lu = _SH::_toUpper(lcname);
    // Generate (almost) all possible locale names
    lcNames.push_back(ln);
    _SH::_addUnique(lcNames, ll);
    _SH::_addUnique(lcNames, ln + "_" + lu);
    _SH::_addUnique(lcNames, ll + "_" + lu);
  }
}

bool _CMsgCat::SelectLocale(const char *messagesLocale)
{
  const string lcname = messagesLocale ? messagesLocale : CBasicAppObjects::SystemLocale();

  // If it is "C" or "POSIX" locale then we won't need translations
  if((lcname == "C") || (lcname == "POSIX")) {
    _ma = NULL;
    _cl = lcname;
    return(true);
  }

  // Generate (almost) all possible locale names
  vector<string> lcNames;
  _GenerateLocaleNames(lcNames, lcname);

  // Try to load all possible locale
  for(vector<string>::size_type i = 0; i < lcNames.size(); ++i) {
    if(_SelectLocale(lcNames[i])) return(true);
  }
  return(false);
}

void _CMsgCat::Clear()
{
  _ka.clear();
  _tt.clear();
  _cl.clear();
  _ma = NULL;
}

const char *_CMsgCat::LookUp(const char *message) const
{
  if(!message) return(NULL);
  if(_ma) {
    _CKArray::const_iterator kit = _ka.find(message);
    if(kit != _ka.end()) {
      _CMArray::const_iterator mit = _ma->find(kit->second);
      if(mit != _ma->end()) return(mit->second.c_str());
    }
  }
  return(message);
}


//
// Exported functions implementation
//
void VSaveMessageCatalog(const string &dstPath, const _TKeys &keys, const vector<_TTranslations> &translations)
{
  // List the directory content and search for *.cat files
  vector<string> list0, list1;
  VListDirectory(list0, dstPath.c_str());
  for(size_t i = 0; i < list0.size(); ++i) {
    string &e = list0[i];
    if(e[0] == '.')
      continue;
    else if(e == "INDEX.cat")
      continue;
    else if(e.rfind(".cat") == (e.length() - 4))
       list1.push_back(e);
  }

  // Delete all *.cat files in the directory which are not part
  // of the catalog to be saved
  for(size_t i = 0; i < list1.size(); ++i ) {
    // Check if part of the catalog
    bool bPart = false;
    for(size_t j = 0; j < translations.size(); ++j) {
      const string onm = translations[j].locale + ".cat";
      if(onm == list1[i]) {
        bPart = true;
        break;
      }
    }
    // Delete if it is not part of the catalog
    if(!bPart) VRemoveFile(dstPath + list1[i]);
  }

  // Used for saving
  string   fnm, onm;
  ofstream os;

  // 1st save the index
  {
    // Open the file
    onm = dstPath + "INDEX.cat";
    fnm = dstPath + "$INDEX.cat";
    os.open(fnm.c_str());
    if(os.fail())
      VError(VERR_OpenFile, _NFL_, "Couldn't open message catalog file \"%s\" for writing", fnm.c_str());
    os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
    // Write the magic numbers
    os.write(_MCAT_INDEX_MAGIC_, strlen(_MCAT_INDEX_MAGIC_));
    // Write the version
    uint32_t ver = _MCAT_FILE_VER_;
    os.write(reinterpret_cast<char*>(&ver), sizeof(ver));
    // Write the optional information
    uint32_t inz = keys.info.size();
    os.write(reinterpret_cast<char*>(&inz), sizeof(inz));
    if(inz) os.write(keys.info.data(), inz);
    // Write the optional extra data
    uint32_t exz = keys.extra.size();
    os.write(reinterpret_cast<char*>(&exz), sizeof(exz));
    if(exz) os.write(&keys.extra[0], exz);
    // Find the number of invalid items
    uint32_t invc = 0;
    for(size_t f = 0; f < keys.msg.size(); ++f) {
      if(!keys.msg[f].valid) invc++;
    }
    // Write the number of message items
    uint32_t msz = keys.msg.size() - invc;
    os.write(reinterpret_cast<char*>(&msz), sizeof(msz));
    // Write the messages
    for(size_t i = 0; i < keys.msg.size(); ++i) {
      if(keys.msg[i].valid) {
        // Write the numeric ID
        os.write(reinterpret_cast<const char*>(&keys.msg[i].id), sizeof(keys.msg[i].id));
        // Write the text
        uint32_t txz = keys.msg[i].text.length();
        os.write(reinterpret_cast<char*>(&txz), sizeof(txz));
        if(txz) os.write(keys.msg[i].text.data(), txz);
      }
    }
    // Close the file
    os.close();
    // Remove the old file and rename the temporary file
    VRemoveFile(onm, true);
    VRenameFile(fnm, onm);
  }

  // 2nd save all the translations
  for(size_t i = 0; i < translations.size(); ++i) {
    // For convenience
    const _TTranslations &t = translations[i];
    // Check if it is empty
    if(translations[i].locale.empty()) continue;
    // Open the file
    onm = dstPath + t.locale + ".cat";
    fnm = dstPath + "$" + t.locale + ".cat";
    os.open(fnm.c_str());
    if(os.fail())
      VError(VERR_OpenFile, _NFL_, "Couldn't open message catalog file \"%s\" for writing", fnm.c_str());
    os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
    // Write the magic numbers
    os.write(_MCAT_TRANS_MAGIC_, strlen(_MCAT_TRANS_MAGIC_));
    // Write the version
    uint32_t ver = _MCAT_FILE_VER_;
    os.write(reinterpret_cast<char*>(&ver), sizeof(ver));
    // Write the optional information
    uint32_t inz = t.info.size();
    os.write(reinterpret_cast<char*>(&inz), sizeof(inz));
    if(inz) os.write(t.info.data(), inz);
    // Write the optional extra data
    uint32_t exz = t.extra.size();
    os.write(reinterpret_cast<char*>(&exz), sizeof(exz));
    if(exz) os.write(&t.extra[0], exz);
    // Find invalid items
    vector<bool> invItem(t.msg.size());
    uint32_t     invc = 0;
    for(size_t f = 0; f < t.msg.size(); ++f) {
      if(!t.msg[f].valid) {
        invItem[f] = false;
        invc++;
      }
      else {
        uint32_t mid = t.msg[f].id;
        bool     vlm = false;
        for(size_t m = 0; m < keys.msg.size(); ++m) {
          if(mid == keys.msg[m].id) {
            vlm = true;
            break;
          }
        }
        if(!vlm) {
          invItem[f] = false;
          invc++;
        }
        else
          invItem[f] = true;
      }
    }
    // Write the number of message items
    uint32_t msz = t.msg.size() - invc;
    os.write(reinterpret_cast<char*>(&msz), sizeof(msz));
    // Write the messages
    for(size_t j = 0; j < t.msg.size(); ++j) {
      if(invItem[j]) {
        // Write the numeric ID
        os.write(reinterpret_cast<const char*>(&t.msg[j].id), sizeof(t.msg[j].id));
        // Write the text
        uint32_t txz = t.msg[j].text.length();
        os.write(reinterpret_cast<char*>(&txz), sizeof(txz));
        if(txz) os.write(t.msg[j].text.data(), txz);
      }
    }
    // Close the file
    os.close();
    // Remove the old file and rename the temporary file
    VRemoveFile(onm, true);
    VRenameFile(fnm, onm);
  }
}

void VLoadMessageCatalog(const string &srcPath, _TKeys &keys, vector<_TTranslations> &translations)
{
  // List the directory content and search for *.cat files
  vector<string> list0, list1;
  VListDirectory(list0, srcPath.c_str());
  for(size_t i = 0; i < list0.size(); ++i) {
    string &e = list0[i];
    if(e[0] == '.')
      continue;
    else if(e == "INDEX.cat")
      continue;
    else if(e.rfind(".cat") == (e.length() - 4))
       list1.push_back(e);
  }

  // Used for loading
  string   fnm;
  ifstream is;

  // 1st load the index
  {
    // Open the file
    fnm = srcPath + "INDEX.cat";
    is.open(fnm.c_str());
    if(is.fail())
      VError(VERR_OpenFile, _NFL_, "Couldn't open message catalog file \"%s\" for reading", fnm.c_str());
    is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
    // Read the magic numbers
    const size_t lomn = strlen(_MCAT_INDEX_MAGIC_);
    vector<char> mnb(lomn);
    is.read(&mnb[0], lomn);
    if(memcmp(&mnb[0], _MCAT_INDEX_MAGIC_, lomn))
      VError(VERR_InvalidHeaderData, _NFL_, "Invalid header in INDEX.cat in \"%s\"", srcPath.c_str());
    // Read the version
    uint32_t ver;
    is.read(reinterpret_cast<char*>(&ver), sizeof(ver));
    if(ver != _MCAT_FILE_VER_)
      VError(VERR_IncorrectVersion, _NFL_, "Invalid version number in the header of INDEX.cat in \"%s\"", srcPath.c_str());
    // Read the optional information
    uint32_t inz;
    is.read(reinterpret_cast<char*>(&inz), sizeof(inz));
    if(inz) {
      keys.info.resize(inz);
      is.read(&keys.info[0], inz);
    }
    else
      keys.info.clear();
    // Read the optional extra data
    uint32_t exz;
    is.read(reinterpret_cast<char*>(&exz), sizeof(exz));
    if(exz) {
      keys.extra.resize(exz);
      is.read(&keys.extra[0], exz);
    }
    else
      keys.extra.clear();
    // Read the number of message items
    uint32_t msz;
    is.read(reinterpret_cast<char*>(&msz), sizeof(msz));
    if(msz)
      keys.msg.resize(msz);
    else
      keys.msg.clear();
    // Read the messages
    for(size_t i = 0; i < msz; ++i) {
      // Read the numeric ID
      keys.msg[i].valid = true;
      is.read(reinterpret_cast<char*>(&keys.msg[i].id), sizeof(keys.msg[i].id));
      // Read the text
      uint32_t txz;
      is.read(reinterpret_cast<char*>(&txz), sizeof(txz));
      if(txz) {
        keys.msg[i].text.resize(txz);
        is.read(&keys.msg[i].text[0], txz);
      }
      else
        keys.msg[i].text.clear();
    }
    // Close the file
    is.close();
  }

  // 2nd load all the translations
  translations.resize(list1.size());
  for(size_t i = 0; i < list1.size(); ++i ) {
    // For convenience
    _TTranslations &t = translations[i];
    // Open the file
    fnm = srcPath + list1[i];
    is.open(fnm.c_str());
    if(is.fail())
      VError(VERR_OpenFile, _NFL_, "Couldn't open message catalog file \"%s\" for reading", fnm.c_str());
    // Read the magic numbers
    char mnb[64];
    memset(mnb, 0, sizeof(mnb));
    is.read(mnb, strlen(_MCAT_TRANS_MAGIC_));
    if(is.fail() || strcmp(mnb, _MCAT_TRANS_MAGIC_)) {
      is.close();
      t.Clear();
      VWarning(_NFL_, "Invalid header in catalog file \"%s\"\nFile skipped", fnm.c_str());
      continue;
    }
    // Read the version
    uint32_t ver;
    is.read(reinterpret_cast<char*>(&ver), sizeof(ver));
    if(is.fail() || (ver != _MCAT_FILE_VER_)) {
      is.close();
      t.Clear();
      VWarning(_NFL_, "Invalid version number in the header of the catalog file \"%s\"\nFile skipped", fnm.c_str());
      continue;
    }
    // Read the rest of the data
    is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
    try {
      // Read the optional information
      uint32_t inz;
      is.read(reinterpret_cast<char*>(&inz), sizeof(inz));
      if(inz) {
        t.info.resize(inz);
        is.read(&t.info[0], inz);
      }
      else
        t.info.clear();
      // Read the optional extra data
      uint32_t exz;
      is.read(reinterpret_cast<char*>(&exz), sizeof(exz));
      if(exz) {
        t.extra.resize(exz);
        is.read(&t.extra[0], exz);
      }
      else
        t.extra.clear();
      // Read the number of message items
      uint32_t msz;
      is.read(reinterpret_cast<char*>(&msz), sizeof(msz));
      if(msz)
        t.msg.resize(msz);
      else
        t.msg.clear();
      //VPrint("%s %d %d %d\n", fnm.c_str(), inz, exz, msz);
      // Read the messages
      for(size_t j = 0; j < msz; ++j) {
        // Read the numeric ID
        t.msg[j].valid = true;
        is.read(reinterpret_cast<char*>(&t.msg[j].id), sizeof(t.msg[j].id));
        // Read the text
        uint32_t txz;
        is.read(reinterpret_cast<char*>(&txz), sizeof(txz));
        if(txz) {
          t.msg[j].text.resize(txz);
          is.read(&t.msg[j].text[0], txz);
        }
        else
          t.msg[j].text.clear();
        //VPrint("\t%d %s %d\n", txz, t.msg[j].text.c_str(), t.msg[j].id);
      }
      // Store the locale name
      t.locale = list1[i].substr(0, list1[i].length() - 4);
    }
    catch(...) {
      t.Clear();
      VWarning(_NFL_, "Error when reading catalog file \"%s\"\nFile skipped", fnm.c_str());
    }
    // Close the file
    is.exceptions(ios::goodbit);
    is.close();
  }
}

void VOptimizeMessageCatalog(_CMsgCat &dstMCat, const _TKeys &keys, const vector<_TTranslations> &translations, bool appendMode)
{
  CRandomCoreUInt32 rnGen(time(NULL));

  // Store the new catalog in append mode
  if(appendMode) {
    // Numeric ID translation table
    map<uint32_t, uint32_t> idc;
    // Store the keys
    for(size_t i = 0; i < keys.msg.size(); ++i) {
      const string &key = _convertString(keys.msg[i].text);
      if(!key.empty()) {
        // Check if a key with exactly the same text already exist
        _CMsgCat::_CKArray::const_iterator kf = dstMCat._ka.find(key);
        if(kf != dstMCat._ka.end()) {
          //VPrint("%s\n", kf->first.c_str());
          idc[keys.msg[i].id] = kf->second;
          continue;
        }
        // Check if a key with the same numeric ID already exist,
        // if it is, generate new ID
        uint32_t id  = keys.msg[i].id;
        for(;;) {
          bool exist = false;
          for(_CMsgCat::_CKArray::const_iterator f = dstMCat._ka.begin(); f != dstMCat._ka.end(); ++f) {
            if(f->second == id) {
              //VPrint("%s %s\n", f->first.c_str(), keys.msg[i].text.c_str());
              exist = true;
              break;
            }
          }
          if(!exist) break;
          id = rnGen.Generate();
        }
        // Store the key
        dstMCat._ka[key] = id;
        // Store the numeric ID translation value
        idc[keys.msg[i].id] = id;
      }
    }
    // Store the message translations
    for(size_t i = 0; i < translations.size(); ++i) {
      // Store the locale
     _CMsgCat::_CMArray &ma = dstMCat._tt[translations[i].locale];
      // Store the translations for this locale
      for(size_t j = 0; j < translations[i].msg.size(); ++j) {
        // Check if a key with the same numeric ID is exist
        if(idc.find(translations[i].msg[j].id) == idc.end()) continue;
        // Put the translated message
        const string &tms = _convertString(translations[i].msg[j].text);;
        if(!tms.empty()) ma[idc[translations[i].msg[j].id]] = tms;
      }
    }
  }
  // Store the new catalog in replace mode
  else {
    // Store the keys
    dstMCat._ka.clear();
    for(size_t i = 0; i < keys.msg.size(); ++i) {
      const string &key = _convertString(keys.msg[i].text);
      if(!key.empty()) dstMCat._ka[key] = keys.msg[i].id;
    }
    // Store the message translations
    dstMCat._tt.clear();
    for(size_t i = 0; i < translations.size(); ++i) {
      // Store the locale
     _CMsgCat::_CMArray &ma = dstMCat._tt[translations[i].locale];
      // Store the translations for this locale
      for(size_t j = 0; j < translations[i].msg.size(); ++j) {
        // Check if a key with the same numeric ID is exist
        bool found = false;
        for(_CMsgCat::_CKArray::const_iterator f = dstMCat._ka.begin(); f != dstMCat._ka.end(); ++f) {
          if(f->second == translations[i].msg[j].id) {
            found = true;
            break;
          }
        }
        if(!found) continue;
        // Put the translated message
        const string &tms = _convertString(translations[i].msg[j].text);;
        if(!tms.empty()) ma[translations[i].msg[j].id] = tms;
      }
    }
  }
}


//
// Implementation of local functions for parsing and converting strings
//
static const char *_parseNumber(string &dst, const char *i, const char *e, char prefix, u_char base)
{
  // For checking if a character is a digit
  static const char octDigits[] = "01234567";
  static const char decDigits[] = "0123456789";
  static const char hexDigits[] = "0123456789ABCDEFabcdef";

  // Prepare the digit set
  const char *numNameID = (base == 8) ? "Octal" : ((base == 10) ? "Decimal" : "Hexadecimal");
  const char *numNamePX = (base == 8) ? "0" : ((base == 10) ? "" : "0x");
  const char *numDigits = (base == 8) ? octDigits : ((base == 10) ? decDigits : hexDigits);

  // Working iterator for parsing numbers
  const char *j = i;

  // Check for premature end of string
  if((j + 1) == e) {
    dst += '\\';
    if(prefix) dst += prefix;
    j = e;
  }
  // Parse the number
  else {
    string tmp;
    while(j++ < e) {
      if(strchr(numDigits, *j))
        tmp += *j;
      else {
        j--;
        break;
      }
    }
    if(tmp.empty()) {
      dst += '\\';
      if(prefix) dst += prefix;
    }
    else {
      const long v = strtol(tmp.c_str(), NULL, base);
      dst += v;
      if(v > 255)
        VWarning(_NFL_, "%s escape sequence \"%s%s\" is out of range", numNameID, numNamePX, tmp.c_str());
    }
  }

  // Return the working iterator
  return(j);
}

static string _convertString(const string &src)
{
  // For parsing simple escape sequences
  static const char c_EscFrom[] = "\\\?\'\"nrtvbfa";
  static const char c_EscTo[]   = "\\\?\'\"\n\r\t\v\b\f\a";

  // Prepare the source pointers
  const char *i = src.data();
  const char *e = i + src.length();

  // For storing the parsed & converted string
  string dst;

  // Parse and convert the string
  for(; i < e; ++i) {
    // Skip new line characters
    if((*i == '\n') || (*i == '\r')) continue;
    // Just put normal characters
    if(*i != '\\') {
      dst += *i;
      continue;
    }
    // Check for premature end of string
    if((i + 1) == e) {
      dst += '\\';
      break;
    }
    // Parse escape sequences
    switch(*++i) {
      // Parse escape sequences of octal numbers (\0oooo...)
      case '0': i = _parseNumber(dst, i, e, '0',  8); break;
      // Parse escape sequences of decimal numbers (\dddd...)
      case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8':
      case '9': i = _parseNumber(dst, i-1, e, 0, 10); break;
      // Parse escape sequences of hexadecimal numbers (\xhhhh...)
      case 'x': i = _parseNumber(dst, i, e, 'x', 16); break;
      // Parse simple espace sequences
      default : const char* pEsc = strchr(c_EscFrom, *i);
                if(!pEsc) {
                  dst += '\\';
                  dst += *i;
                }
                else
                  dst += c_EscTo[pEsc - c_EscFrom];
    }
  }

  // Return the parsed & converted string
  return(dst);
}

} // namespace vge
