/** \file MessageCatalog.h */
// ############################################################################################
//  MessageCatalog.h
//  Definition of CMessageCatalog class and its helper classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___MESSAGE_CATALOG_H_
#define _BASE___MESSAGE_CATALOG_H_

// VortexGE headers
#include "../Base/UString.h"
#include "../System/DataSinkSource.h"


namespace vge {

  //!
  //! \internal
  //! \brief Structure used when building, saving and loading message catalog
  //!
  //! \note All message texts must be encoded in UTF-8.
  //!
  //! \api beta
  struct VGE_EXPORT _TTextEntry {
    // These data are mean to be accessed directly by whatever using this structure
    bool     valid; //!< To indicate if this entry is valid (or invalid/deleted)
    uint32_t id;    //!< Numeric ID of this text entry
    string   text;  //!< The text (message)

    // For convenience
    //!
    //! Construct an empty (invalid) text entry
    inline _TTextEntry()
    : valid(false), id(0) {}
    //!
    //! Construct a text entry with the given numeric ID but with empty text
    inline _TTextEntry(uint32_t id_)
    : valid(true), id(id_) {}
    //!
    //! Construct a text entry with the given numeric ID and text
    inline _TTextEntry(uint32_t id_, const string &text_)
    : valid(true), id(id_), text(text_) {}
  };


  //!
  //! \brief Structure used when building, saving and loading message catalog
  //!
  //! One shouldn't need to use this structure directly unless wants to implement
  //! ones own message catalog editor.
  //!
  //! \note All message texts must be encoded in UTF-8.
  //!
  //! \api beta
  struct VGE_EXPORT _TKeys {
    // These data are mean to be accessed directly by whatever using this structure
    string              info;  //!< Optional information text
    vector<char>        extra; //!< Optional extra data (not used yet in this version)
    vector<_TTextEntry> msg;   //!< Array of keys (the original messages)

    // Information functions
    //!
    //! Check if this key structure is empty
    inline bool IsEmpty()
    { return(info.empty() && extra.empty() && msg.empty()); }

    // Management functions
    //!
    //! Clear this key structure
    inline void Clear()
    { info.clear(); extra.clear(); msg.clear(); }
    //!
    //! Generate unique random numeric ID
    uint32_t GenerateNewNumericID();
  };


  //!
  //! \brief Structure used when building, saving and loading message catalog
  //!
  //! One shouldn't need to use this structure directly unless wants to implement
  //! ones own message catalog editor.
  //!
  //! \note All message texts must be encoded in UTF-8.
  //!
  //! \api beta
  struct VGE_EXPORT _TTranslations {
    // These data are mean to be accessed directly by whatever using this structure
    string              locale; //!< Locale for this translations
    string              info;   //!< Optional information text
    vector<char>        extra;  //!< Optional extra data (not used yet in this version)
    vector<_TTextEntry> msg;    //!< Array of translated messages
  
    // For convenience
    //!
    //! Construct an empty (invalid) translation structure
    inline _TTranslations()
    : locale("") {}
    //!
    //! Construct a translation structure with the given locale
    //! but with empty messages array
    inline _TTranslations(const string &locale_)
    : locale(locale_) {}

    // Information functions
    //!
    //! Check if this translation structure is empty
    inline bool IsEmpty()
    { return(locale.empty() && info.empty() && extra.empty() && msg.empty()); }

    // Management functions
    //!
    //! Clear this translation structure
    inline void Clear()
    { locale.clear(); info.clear(); extra.clear(); msg.clear(); }
  };


  //!
  //! \internal
  //! \brief Used for storing message catalog in the compact form suitable
  //!        for fast message translation
  //!
  //! \note All message texts must be encoded in UTF-8.
  //!
  //! \api beta
  class VGE_EXPORT _CMsgCat : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct an empty message catalog.
      inline _CMsgCat()
      : _ma(NULL) {}

      // Initialization functions
      //!
      //! Select locale
      //! \param[in] messagesLocale  Locale name for determining which catalog to be used,
      //!                            giving NULL means using the current system locale
      //!                            returned by CBasicAppObjects::SystemLocale().
      //! \return True if the wanted locale is exist, false otherwise.
      bool SelectLocale(const char *messagesLocale = NULL);
      //!
      //! Clear the message catalog.
      void Clear();

      // Information functions
      //!
      //! Check if this message catalog is empty or not.
      inline bool IsEmpty() const
      { return(_ka.empty() && _tt.empty()); }
      //!
      //! Get the locale currently selecyed in this message catalog.
      inline const string &SelectedLocale() const
      { return(_cl); }

      // Messages catalog functions
      //!
      //! Translate the given message. If no translation is found, the original
      //! pointer (const char *message) will be returned back.
      const char *LookUp(const char *message) const;

    private:
      // Helper functions
      bool _SelectLocale(const string &messagesLocale);
      void _GenerateLocaleNames(vector<string> &lcNames, const string &lcname);
    
      typedef map<string, uint32_t> _CKArray; // Array of original messages (the keys/IDs)
      typedef map<uint32_t, string> _CMArray; // Array of translated messages for a particular locale
      typedef map<string, _CMArray> _CMTable; // Translation table for all loaded locales
  
      // Data
      _CKArray _ka; // Array of original messages
      _CMTable _tt; // Translation table
    
             string   _cl; // Selected locale
      const _CMArray *_ma; // Selected translated-messages array

      // Friend functions
      friend void VOptimizeMessageCatalog(_CMsgCat &dstMCat, const _TKeys &keys, const vector<_TTranslations> &translations, bool appendMode);
  };


  // Helper functions for saving, loading and optimizing message catalog
  //!
  //! \brief Used for saving message catalog
  //!
  //! \api beta
  extern VGE_EXPORT void VSaveMessageCatalog(const string &dstPath, const _TKeys &keys, const vector<_TTranslations> &translations);
  //!
  //! \brief Used for loading message catalog
  //!
  //! \api beta
  extern VGE_EXPORT void VLoadMessageCatalog(const string &srcPath, _TKeys &keys, vector<_TTranslations> &translations);
  //!
  //! \brief Used for optimizing and converting the loaded catalog into
  //!         a form suitable for fast message translation
  //!
  //! \api beta
  extern VGE_EXPORT void VOptimizeMessageCatalog(_CMsgCat &dstMCat, const _TKeys &keys, const vector<_TTranslations> &translations, bool appendMode);


  //!
  //! \brief A class used for managinga a message catalog
  //!
  //! \note The message texts must be encoded in UTF-8.
  //!
  //! See the page \ref MCat_DirContent_FFormat "Message Catalog Directory Contents and File Format" for more informations.
  //!
  //! \api beta
  class VGE_EXPORT CMessageCatalog : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct an empty message catalog.
      inline CMessageCatalog()
      {}
      //!
      //! Construct a message catalog using the given path and locale.
      //! \param[in] pathToMessageCatalog Path to the directory containing the catalog files.
      //! \param[in] messagesLocale       Locale name for determining which catalog to be used,
      //!                                 giving NULL means using the current system locale
      //!                                 returned by CBasicAppObjects::SystemLocale().
      inline CMessageCatalog(const string &pathToMessageCatalog, const char *messagesLocale = NULL)
      { LoadCatalog(pathToMessageCatalog, messagesLocale); }

      // Initialization functions
      //!
      //! (Re)initialize this message catalog using the given path and locale.
      //! \param[in] pathToMessageCatalog Path to the directory containing the catalog files.
      //! \param[in] messagesLocale       Locale name for determining which catalog to be used,
      //!                                 giving NULL means using the current system locale
      //!                                 returned by CBasicAppObjects::SystemLocale().
      //! \param[in] appendMode           If "true", the new catalog to be loaded will be
      //!                                 appended to the current one.
      void LoadCatalog(const string &pathToMessageCatalog, const char *messagesLocale = NULL, bool appendMode = false);
      //!
      //! Select which locale to be used when translating message
      //! \param[in] messagesLocale Locale name for determining which catalog to be used,
      //!                           giving NULL means using the current system locale
      //!                           returned by CBasicAppObjects::SystemLocale().
      //! \return True if the wanted locale is exist, false otherwise.
      inline bool SelectLocale(const char *messagesLocale = NULL)
      { return(_theCatalog.SelectLocale(messagesLocale)); }
      //!
      //! Clear the message catalog.
      inline void Clear()
      { _theCatalog.Clear(); }

      // Information functions
      //!
      //! Check if this message catalog is initialized.
      inline bool IsEmpty() const
      { return(_theCatalog.IsEmpty()); }
      //!
      //! Get the locale currently selected in this message catalog.
      inline const string &SelectedLocale() const
      { return(_theCatalog.SelectedLocale()); }

      // Messages catalog functions
      //!
      //! Lookup the given message in the catalog. If no translation is found,
      //! the original pointer (const char *message) will be returned back.
      inline const char *LookUp(const char *message) const
      { return(_theCatalog.LookUp(message)); }

    protected:
      class _CMsgCat _theCatalog;
  };

} // namespace vge
#endif


/**
\page MCat_DirContent_FFormat Message Catalog Directory Contents and File Format

A Message catalog directory basically contains :
  -# One index file (INDEX.cat)
  -# Some message translation files (all other *.cat files)

Example:
\code
bash$ ls /path/to/my/msgcat/dir
.   ..   INDEX.cat de_DE.cat   en_UK.cat   id_ID.cat
bash$
\endcode
\n
\n
All catalog files (INDEX.cat and all other *.cat files) are binary files.
User should not care about the format and to work with the files, just use
the functions :
  - vge::VSaveMessageCatalog()
  - vge::VLoadMessageCatalog()
  - vge::VOptimizeMessageCatalog()

\n
Unless one wants to implement one's own message catalog editor or any other
functionality which need to use VortexGE message catalog internal functions
(and their related structures), it is reccomended to just use the high level
class (vge::CMessageCatalog) and to use the provided utility application,
"mcatedit" to edit a message catalog directory.
\n
<hr>
Back to vge::CMessageCatalog.
*/
