// ############################################################################################
//  UString.cc
//  Implementation of CUString class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "MessageDisplayer.h"
#include "UString.h"


namespace vge {

//
// Helper functions
//
static bool _IsLegalUtf8_(const CUString::Utf8 *src, size_t slen)
{
  const CUString::Utf8 *ptr = src + slen;

  CUString::Utf8 c;
  switch(slen) {
    default: return(false);

    // Everything else falls through when "true"
    case 4 : c = *--ptr; if((c < 0x80) || (c > 0xBF)) return(false);
    case 3 : c = *--ptr; if((c < 0x80) || (c > 0xBF)) return(false);
    case 2 : c = *--ptr; if( c > 0xBF               ) return(false);
         switch (*src) {
          // No fall-through in this inner switch
          case 0xE0: if(c < 0xA0) return(false); break;
          case 0xED: if(c > 0x9F) return(false); break;
          case 0xF0: if(c < 0x90) return(false); break;
          case 0xF4: if(c > 0x8F) return(false); break;
          default  : if(c < 0x80) return(false);
         }
    case 1 : if((*src >= 0x80) && (*src < 0xC2)) return(false);
  }
  if(*src > 0xF4) return(false);

  return(true);
}


//
// Unicode string class
//
void CUString::AsciiToUtf32(Utf32String &dst, const char *src, size_t slen)
{
  // Resize our destination buffer
  dst.resize(slen);

  // Prepare the source pointers
  const Utf8 *sn = reinterpret_cast<const Utf8*>(src);
  const Utf8 *se = sn + slen;

  // Prepare the destination pointers
  Utf32 *dn = &dst[0];

  // Convert all characters
  while(sn < se) *dn++ = *sn++;
}

void CUString::Utf8ToUtf32(Utf32String &dst, const Utf8 *src, size_t slen)
{
  // Index into the table below with the first byte of a UTF-8 sequence
  // to get the number of trailing bytes that are supposed to follow it.
  //
  // Note that legal UTF-8 values can't have 4 or 5 bytes. The table is
  // left as-is for anyone who may want to do such conversion, which was
  // allowed in earlier algorithms.
  static const char trailingBytesForUtf8[256] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
    2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2, 3,3,3,3,3,3,3,3,4,4,4,4,5,5,5,5
  };

  // Magic values subtracted from a buffer value during UTF-8 conversion.
  //
  // This table contains as many values as there might be trailing bytes
  // in a UTF-8 sequence.
  static const Utf32 offsetsFromUtf8[6] = {
    0x00000000UL, 0x00003080UL, 0x000E2080UL,
    0x03C82080UL, 0xFA082080UL, 0x82082080UL
  };

  // Some constants related to the UTF-8 -> UTF-32 conversion
  const Utf32 uniSurHighStart    = 0xD800;
  const Utf32 uniSurLowEnd       = 0xDFFF;
  const Utf32 uniMaxLegalUtf32   = 0x0010FFFF;
  const Utf32 uniReplacementChar = 0x0000FFFD;

  // Resize our destination buffer
  //
  // The maximum length of the resulting UTF-32 text will never be longer than
  // the length of the source UTF-8 text. The conversion buffer here is made 4
  // times longer just to prevent segmentation fault in case of errornous text
  // conversion.
  const size_t dlen = slen * 4;
  dst.resize(dlen);

  // Prepare the source pointers
  const Utf8 *sn = src;
  const Utf8 *se = sn + slen;

  // Prepare the destination pointers
  Utf32 *ds = &dst[0];
  Utf32 *dn = ds;
  Utf32 *de = ds + dlen;

  // Convert all characters
  while(sn < se) {
    // Break if the target is exhausted, should never happened
    if(dn >= de) break;

    // Get the number of extra bytes to be processed
    size_t extraBytesToRead = trailingBytesForUtf8[*sn];

    // Break if the source is exhausted,
    // should never happened on valid UTF-8 sequence
    if((sn + extraBytesToRead) >= se) break;

    // Check if the source is an illlegal UTF-8 sequence
    if(!_IsLegalUtf8_(sn, extraBytesToRead + 1)) break;

    // Generate one UTF-32 character from one or more UTF-8 characters
    Utf32 ch = 0;
    switch(extraBytesToRead) { // Everything falls through
      case 5: ch += *sn++; ch <<= 6;
      case 4: ch += *sn++; ch <<= 6;
      case 3: ch += *sn++; ch <<= 6;
      case 2: ch += *sn++; ch <<= 6;
      case 1: ch += *sn++; ch <<= 6;
      case 0: ch += *sn++;
    }

    ch -= offsetsFromUtf8[extraBytesToRead];

    // Check if the generated value exceed the maximum legal UTF-32 value
    if(ch > uniMaxLegalUtf32)
      *dn++ = uniReplacementChar;
    // UTF-16 surrogate values are illegal in UTF-32, and, anything over
    // plane 17 (> 0x10FFFF) is illegal too
    else if((ch >= uniSurHighStart) && (ch <= uniSurLowEnd))
      *dn++ = uniReplacementChar;
    // Got a valid UTF-32 character
    else
      *dn++ = ch;
  }

  // Make the destination buffer's size smaller if necessary
  size_t xlen = dn - ds;
  if(xlen < dlen) dst.resize(xlen);
}

void CUString::Utf32ToUtf8(string &dst, const Utf32 *src, size_t slen)
{
  // Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
  // into the first byte, depending on how many bytes follow. There are as
  // many entries in this table as there are UTF-8 sequence types (i.e: one
  // byte sequence, two bytes, etc.). Sequences for legal UTF-8 will be 4
  // or fewer bytes total.
  static const Utf8 firstByteMark[7] = { 0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC };

  // Some constants related to the UTF-32 -> UTF-8 conversion
  const Utf32 byteMask           = 0xBF;
  const Utf32 byteMark           = 0x80; 
  const Utf32 uniMaxLegalUtf32   = 0x0010FFFF;
  const Utf32 uniReplacementChar = 0x0000FFFD;

  // Resize our destination buffer
  //
  // The maximum length of the resulting UTF-8 text will never be longer than
  // 4 times the length of the source UTF-8 text (if the text is legal). The
  // conversion buffer here is made 6 times longer just to prevent segmentation
  // fault in case of errornous text conversion.
  const uint32_t dlen = slen * 6;
  dst.resize(dlen);

  // Prepare the source pointers
  const Utf32 *ss = src;
  const Utf32 *se = ss + slen;

  // Prepare the destination pointers
  Utf8 *ds = reinterpret_cast<Utf8*>(&dst[0]);
  Utf8 *dn = ds;
  Utf8 *de = ds + dlen;

  while(ss < se) {
    // Get the source character
    Utf32 ch = *ss++;

    // Figure out how many bytes the result will require.
    // Turn any illegally large UTF-32 things (> Plane 17)
    // into replacement chars.
    size_t bytesToWrite = 0;
         if(ch <  Utf32(0x00080)  ) bytesToWrite = 1;
    else if(ch <  Utf32(0x00800)  ) bytesToWrite = 2;
    else if(ch <  Utf32(0x10000)  ) bytesToWrite = 3;
    else if(ch <= uniMaxLegalUtf32) bytesToWrite = 4;
    else {
      bytesToWrite = 3;
      ch = uniReplacementChar;
    }

    // Break if the target is exhausted, should never happened
    dn += bytesToWrite;
    if(dn >= de) break;

    // Perform conversion
    switch (bytesToWrite) { // Everything falls through
      case 4: *--dn = Utf8((ch | byteMark) & byteMask); ch >>= 6;
      case 3: *--dn = Utf8((ch | byteMark) & byteMask); ch >>= 6;
      case 2: *--dn = Utf8((ch | byteMark) & byteMask); ch >>= 6;
      case 1: *--dn = Utf8( ch | firstByteMark[bytesToWrite]);
    }
    dn += bytesToWrite;
  }

  // Make the destination buffer's size smaller if necessary
  size_t xlen = dn - ds;
  if(xlen < dlen) dst.resize(xlen);
}

void CUString::CaseFolding(CUString &dst, const CUString &src, bool fullCaseFolding, bool specialCaseFolding)
{
  // A structure used for storing case-folding map entry
  struct _TCaseFoldingMap {
    CUString::Utf32 from;
    CUString::Utf32 to[3];
    uint32_t        toCount;
  };

  // NOTE: The case-folding maps below are generated using the utility
  //       application, "util/minor/gcs.cc" based on the data that was
  //       downloaded from :
  //           http://www.unicode.org/Public/UNIDATA/CaseFolding.txt

  // Common case-folding, common mappings shared by both simple and full mappings
  static const _TCaseFoldingMap _ccf[] = {
    {   65,{   97,    0,    0},1},  {   66,{   98,    0,    0},1},
    {   67,{   99,    0,    0},1},  {   68,{  100,    0,    0},1},
    {   69,{  101,    0,    0},1},  {   70,{  102,    0,    0},1},
    {   71,{  103,    0,    0},1},  {   72,{  104,    0,    0},1},
    {   73,{  105,    0,    0},1},  {   74,{  106,    0,    0},1},
    {   75,{  107,    0,    0},1},  {   76,{  108,    0,    0},1},
    {   77,{  109,    0,    0},1},  {   78,{  110,    0,    0},1},
    {   79,{  111,    0,    0},1},  {   80,{  112,    0,    0},1},
    {   81,{  113,    0,    0},1},  {   82,{  114,    0,    0},1},
    {   83,{  115,    0,    0},1},  {   84,{  116,    0,    0},1},
    {   85,{  117,    0,    0},1},  {   86,{  118,    0,    0},1},
    {   87,{  119,    0,    0},1},  {   88,{  120,    0,    0},1},
    {   89,{  121,    0,    0},1},  {   90,{  122,    0,    0},1},
    {  181,{  956,    0,    0},1},  {  192,{  224,    0,    0},1},
    {  193,{  225,    0,    0},1},  {  194,{  226,    0,    0},1},
    {  195,{  227,    0,    0},1},  {  196,{  228,    0,    0},1},
    {  197,{  229,    0,    0},1},  {  198,{  230,    0,    0},1},
    {  199,{  231,    0,    0},1},  {  200,{  232,    0,    0},1},
    {  201,{  233,    0,    0},1},  {  202,{  234,    0,    0},1},
    {  203,{  235,    0,    0},1},  {  204,{  236,    0,    0},1},
    {  205,{  237,    0,    0},1},  {  206,{  238,    0,    0},1},
    {  207,{  239,    0,    0},1},  {  208,{  240,    0,    0},1},
    {  209,{  241,    0,    0},1},  {  210,{  242,    0,    0},1},
    {  211,{  243,    0,    0},1},  {  212,{  244,    0,    0},1},
    {  213,{  245,    0,    0},1},  {  214,{  246,    0,    0},1},
    {  216,{  248,    0,    0},1},  {  217,{  249,    0,    0},1},
    {  218,{  250,    0,    0},1},  {  219,{  251,    0,    0},1},
    {  220,{  252,    0,    0},1},  {  221,{  253,    0,    0},1},
    {  222,{  254,    0,    0},1},  {  256,{  257,    0,    0},1},
    {  258,{  259,    0,    0},1},  {  260,{  261,    0,    0},1},
    {  262,{  263,    0,    0},1},  {  264,{  265,    0,    0},1},
    {  266,{  267,    0,    0},1},  {  268,{  269,    0,    0},1},
    {  270,{  271,    0,    0},1},  {  272,{  273,    0,    0},1},
    {  274,{  275,    0,    0},1},  {  276,{  277,    0,    0},1},
    {  278,{  279,    0,    0},1},  {  280,{  281,    0,    0},1},
    {  282,{  283,    0,    0},1},  {  284,{  285,    0,    0},1},
    {  286,{  287,    0,    0},1},  {  288,{  289,    0,    0},1},
    {  290,{  291,    0,    0},1},  {  292,{  293,    0,    0},1},
    {  294,{  295,    0,    0},1},  {  296,{  297,    0,    0},1},
    {  298,{  299,    0,    0},1},  {  300,{  301,    0,    0},1},
    {  302,{  303,    0,    0},1},  {  306,{  307,    0,    0},1},
    {  308,{  309,    0,    0},1},  {  310,{  311,    0,    0},1},
    {  313,{  314,    0,    0},1},  {  315,{  316,    0,    0},1},
    {  317,{  318,    0,    0},1},  {  319,{  320,    0,    0},1},
    {  321,{  322,    0,    0},1},  {  323,{  324,    0,    0},1},
    {  325,{  326,    0,    0},1},  {  327,{  328,    0,    0},1},
    {  330,{  331,    0,    0},1},  {  332,{  333,    0,    0},1},
    {  334,{  335,    0,    0},1},  {  336,{  337,    0,    0},1},
    {  338,{  339,    0,    0},1},  {  340,{  341,    0,    0},1},
    {  342,{  343,    0,    0},1},  {  344,{  345,    0,    0},1},
    {  346,{  347,    0,    0},1},  {  348,{  349,    0,    0},1},
    {  350,{  351,    0,    0},1},  {  352,{  353,    0,    0},1},
    {  354,{  355,    0,    0},1},  {  356,{  357,    0,    0},1},
    {  358,{  359,    0,    0},1},  {  360,{  361,    0,    0},1},
    {  362,{  363,    0,    0},1},  {  364,{  365,    0,    0},1},
    {  366,{  367,    0,    0},1},  {  368,{  369,    0,    0},1},
    {  370,{  371,    0,    0},1},  {  372,{  373,    0,    0},1},
    {  374,{  375,    0,    0},1},  {  376,{  255,    0,    0},1},
    {  377,{  378,    0,    0},1},  {  379,{  380,    0,    0},1},
    {  381,{  382,    0,    0},1},  {  383,{  115,    0,    0},1},
    {  385,{  595,    0,    0},1},  {  386,{  387,    0,    0},1},
    {  388,{  389,    0,    0},1},  {  390,{  596,    0,    0},1},
    {  391,{  392,    0,    0},1},  {  393,{  598,    0,    0},1},
    {  394,{  599,    0,    0},1},  {  395,{  396,    0,    0},1},
    {  398,{  477,    0,    0},1},  {  399,{  601,    0,    0},1},
    {  400,{  603,    0,    0},1},  {  401,{  402,    0,    0},1},
    {  403,{  608,    0,    0},1},  {  404,{  611,    0,    0},1},
    {  406,{  617,    0,    0},1},  {  407,{  616,    0,    0},1},
    {  408,{  409,    0,    0},1},  {  412,{  623,    0,    0},1},
    {  413,{  626,    0,    0},1},  {  415,{  629,    0,    0},1},
    {  416,{  417,    0,    0},1},  {  418,{  419,    0,    0},1},
    {  420,{  421,    0,    0},1},  {  422,{  640,    0,    0},1},
    {  423,{  424,    0,    0},1},  {  425,{  643,    0,    0},1},
    {  428,{  429,    0,    0},1},  {  430,{  648,    0,    0},1},
    {  431,{  432,    0,    0},1},  {  433,{  650,    0,    0},1},
    {  434,{  651,    0,    0},1},  {  435,{  436,    0,    0},1},
    {  437,{  438,    0,    0},1},  {  439,{  658,    0,    0},1},
    {  440,{  441,    0,    0},1},  {  444,{  445,    0,    0},1},
    {  452,{  454,    0,    0},1},  {  453,{  454,    0,    0},1},
    {  455,{  457,    0,    0},1},  {  456,{  457,    0,    0},1},
    {  458,{  460,    0,    0},1},  {  459,{  460,    0,    0},1},
    {  461,{  462,    0,    0},1},  {  463,{  464,    0,    0},1},
    {  465,{  466,    0,    0},1},  {  467,{  468,    0,    0},1},
    {  469,{  470,    0,    0},1},  {  471,{  472,    0,    0},1},
    {  473,{  474,    0,    0},1},  {  475,{  476,    0,    0},1},
    {  478,{  479,    0,    0},1},  {  480,{  481,    0,    0},1},
    {  482,{  483,    0,    0},1},  {  484,{  485,    0,    0},1},
    {  486,{  487,    0,    0},1},  {  488,{  489,    0,    0},1},
    {  490,{  491,    0,    0},1},  {  492,{  493,    0,    0},1},
    {  494,{  495,    0,    0},1},  {  497,{  499,    0,    0},1},
    {  498,{  499,    0,    0},1},  {  500,{  501,    0,    0},1},
    {  502,{  405,    0,    0},1},  {  503,{  447,    0,    0},1},
    {  504,{  505,    0,    0},1},  {  506,{  507,    0,    0},1},
    {  508,{  509,    0,    0},1},  {  510,{  511,    0,    0},1},
    {  512,{  513,    0,    0},1},  {  514,{  515,    0,    0},1},
    {  516,{  517,    0,    0},1},  {  518,{  519,    0,    0},1},
    {  520,{  521,    0,    0},1},  {  522,{  523,    0,    0},1},
    {  524,{  525,    0,    0},1},  {  526,{  527,    0,    0},1},
    {  528,{  529,    0,    0},1},  {  530,{  531,    0,    0},1},
    {  532,{  533,    0,    0},1},  {  534,{  535,    0,    0},1},
    {  536,{  537,    0,    0},1},  {  538,{  539,    0,    0},1},
    {  540,{  541,    0,    0},1},  {  542,{  543,    0,    0},1},
    {  544,{  414,    0,    0},1},  {  546,{  547,    0,    0},1},
    {  548,{  549,    0,    0},1},  {  550,{  551,    0,    0},1},
    {  552,{  553,    0,    0},1},  {  554,{  555,    0,    0},1},
    {  556,{  557,    0,    0},1},  {  558,{  559,    0,    0},1},
    {  560,{  561,    0,    0},1},  {  562,{  563,    0,    0},1},
    {  571,{  572,    0,    0},1},  {  573,{  410,    0,    0},1},
    {  577,{  660,    0,    0},1},  {  837,{  953,    0,    0},1},
    {  902,{  940,    0,    0},1},  {  904,{  941,    0,    0},1},
    {  905,{  942,    0,    0},1},  {  906,{  943,    0,    0},1},
    {  908,{  972,    0,    0},1},  {  910,{  973,    0,    0},1},
    {  911,{  974,    0,    0},1},  {  913,{  945,    0,    0},1},
    {  914,{  946,    0,    0},1},  {  915,{  947,    0,    0},1},
    {  916,{  948,    0,    0},1},  {  917,{  949,    0,    0},1},
    {  918,{  950,    0,    0},1},  {  919,{  951,    0,    0},1},
    {  920,{  952,    0,    0},1},  {  921,{  953,    0,    0},1},
    {  922,{  954,    0,    0},1},  {  923,{  955,    0,    0},1},
    {  924,{  956,    0,    0},1},  {  925,{  957,    0,    0},1},
    {  926,{  958,    0,    0},1},  {  927,{  959,    0,    0},1},
    {  928,{  960,    0,    0},1},  {  929,{  961,    0,    0},1},
    {  931,{  963,    0,    0},1},  {  932,{  964,    0,    0},1},
    {  933,{  965,    0,    0},1},  {  934,{  966,    0,    0},1},
    {  935,{  967,    0,    0},1},  {  936,{  968,    0,    0},1},
    {  937,{  969,    0,    0},1},  {  938,{  970,    0,    0},1},
    {  939,{  971,    0,    0},1},  {  962,{  963,    0,    0},1},
    {  976,{  946,    0,    0},1},  {  977,{  952,    0,    0},1},
    {  981,{  966,    0,    0},1},  {  982,{  960,    0,    0},1},
    {  984,{  985,    0,    0},1},  {  986,{  987,    0,    0},1},
    {  988,{  989,    0,    0},1},  {  990,{  991,    0,    0},1},
    {  992,{  993,    0,    0},1},  {  994,{  995,    0,    0},1},
    {  996,{  997,    0,    0},1},  {  998,{  999,    0,    0},1},
    { 1000,{ 1001,    0,    0},1},  { 1002,{ 1003,    0,    0},1},
    { 1004,{ 1005,    0,    0},1},  { 1006,{ 1007,    0,    0},1},
    { 1008,{  954,    0,    0},1},  { 1009,{  961,    0,    0},1},
    { 1012,{  952,    0,    0},1},  { 1013,{  949,    0,    0},1},
    { 1015,{ 1016,    0,    0},1},  { 1017,{ 1010,    0,    0},1},
    { 1018,{ 1019,    0,    0},1},  { 1024,{ 1104,    0,    0},1},
    { 1025,{ 1105,    0,    0},1},  { 1026,{ 1106,    0,    0},1},
    { 1027,{ 1107,    0,    0},1},  { 1028,{ 1108,    0,    0},1},
    { 1029,{ 1109,    0,    0},1},  { 1030,{ 1110,    0,    0},1},
    { 1031,{ 1111,    0,    0},1},  { 1032,{ 1112,    0,    0},1},
    { 1033,{ 1113,    0,    0},1},  { 1034,{ 1114,    0,    0},1},
    { 1035,{ 1115,    0,    0},1},  { 1036,{ 1116,    0,    0},1},
    { 1037,{ 1117,    0,    0},1},  { 1038,{ 1118,    0,    0},1},
    { 1039,{ 1119,    0,    0},1},  { 1040,{ 1072,    0,    0},1},
    { 1041,{ 1073,    0,    0},1},  { 1042,{ 1074,    0,    0},1},
    { 1043,{ 1075,    0,    0},1},  { 1044,{ 1076,    0,    0},1},
    { 1045,{ 1077,    0,    0},1},  { 1046,{ 1078,    0,    0},1},
    { 1047,{ 1079,    0,    0},1},  { 1048,{ 1080,    0,    0},1},
    { 1049,{ 1081,    0,    0},1},  { 1050,{ 1082,    0,    0},1},
    { 1051,{ 1083,    0,    0},1},  { 1052,{ 1084,    0,    0},1},
    { 1053,{ 1085,    0,    0},1},  { 1054,{ 1086,    0,    0},1},
    { 1055,{ 1087,    0,    0},1},  { 1056,{ 1088,    0,    0},1},
    { 1057,{ 1089,    0,    0},1},  { 1058,{ 1090,    0,    0},1},
    { 1059,{ 1091,    0,    0},1},  { 1060,{ 1092,    0,    0},1},
    { 1061,{ 1093,    0,    0},1},  { 1062,{ 1094,    0,    0},1},
    { 1063,{ 1095,    0,    0},1},  { 1064,{ 1096,    0,    0},1},
    { 1065,{ 1097,    0,    0},1},  { 1066,{ 1098,    0,    0},1},
    { 1067,{ 1099,    0,    0},1},  { 1068,{ 1100,    0,    0},1},
    { 1069,{ 1101,    0,    0},1},  { 1070,{ 1102,    0,    0},1},
    { 1071,{ 1103,    0,    0},1},  { 1120,{ 1121,    0,    0},1},
    { 1122,{ 1123,    0,    0},1},  { 1124,{ 1125,    0,    0},1},
    { 1126,{ 1127,    0,    0},1},  { 1128,{ 1129,    0,    0},1},
    { 1130,{ 1131,    0,    0},1},  { 1132,{ 1133,    0,    0},1},
    { 1134,{ 1135,    0,    0},1},  { 1136,{ 1137,    0,    0},1},
    { 1138,{ 1139,    0,    0},1},  { 1140,{ 1141,    0,    0},1},
    { 1142,{ 1143,    0,    0},1},  { 1144,{ 1145,    0,    0},1},
    { 1146,{ 1147,    0,    0},1},  { 1148,{ 1149,    0,    0},1},
    { 1150,{ 1151,    0,    0},1},  { 1152,{ 1153,    0,    0},1},
    { 1162,{ 1163,    0,    0},1},  { 1164,{ 1165,    0,    0},1},
    { 1166,{ 1167,    0,    0},1},  { 1168,{ 1169,    0,    0},1},
    { 1170,{ 1171,    0,    0},1},  { 1172,{ 1173,    0,    0},1},
    { 1174,{ 1175,    0,    0},1},  { 1176,{ 1177,    0,    0},1},
    { 1178,{ 1179,    0,    0},1},  { 1180,{ 1181,    0,    0},1},
    { 1182,{ 1183,    0,    0},1},  { 1184,{ 1185,    0,    0},1},
    { 1186,{ 1187,    0,    0},1},  { 1188,{ 1189,    0,    0},1},
    { 1190,{ 1191,    0,    0},1},  { 1192,{ 1193,    0,    0},1},
    { 1194,{ 1195,    0,    0},1},  { 1196,{ 1197,    0,    0},1},
    { 1198,{ 1199,    0,    0},1},  { 1200,{ 1201,    0,    0},1},
    { 1202,{ 1203,    0,    0},1},  { 1204,{ 1205,    0,    0},1},
    { 1206,{ 1207,    0,    0},1},  { 1208,{ 1209,    0,    0},1},
    { 1210,{ 1211,    0,    0},1},  { 1212,{ 1213,    0,    0},1},
    { 1214,{ 1215,    0,    0},1},  { 1217,{ 1218,    0,    0},1},
    { 1219,{ 1220,    0,    0},1},  { 1221,{ 1222,    0,    0},1},
    { 1223,{ 1224,    0,    0},1},  { 1225,{ 1226,    0,    0},1},
    { 1227,{ 1228,    0,    0},1},  { 1229,{ 1230,    0,    0},1},
    { 1232,{ 1233,    0,    0},1},  { 1234,{ 1235,    0,    0},1},
    { 1236,{ 1237,    0,    0},1},  { 1238,{ 1239,    0,    0},1},
    { 1240,{ 1241,    0,    0},1},  { 1242,{ 1243,    0,    0},1},
    { 1244,{ 1245,    0,    0},1},  { 1246,{ 1247,    0,    0},1},
    { 1248,{ 1249,    0,    0},1},  { 1250,{ 1251,    0,    0},1},
    { 1252,{ 1253,    0,    0},1},  { 1254,{ 1255,    0,    0},1},
    { 1256,{ 1257,    0,    0},1},  { 1258,{ 1259,    0,    0},1},
    { 1260,{ 1261,    0,    0},1},  { 1262,{ 1263,    0,    0},1},
    { 1264,{ 1265,    0,    0},1},  { 1266,{ 1267,    0,    0},1},
    { 1268,{ 1269,    0,    0},1},  { 1270,{ 1271,    0,    0},1},
    { 1272,{ 1273,    0,    0},1},  { 1280,{ 1281,    0,    0},1},
    { 1282,{ 1283,    0,    0},1},  { 1284,{ 1285,    0,    0},1},
    { 1286,{ 1287,    0,    0},1},  { 1288,{ 1289,    0,    0},1},
    { 1290,{ 1291,    0,    0},1},  { 1292,{ 1293,    0,    0},1},
    { 1294,{ 1295,    0,    0},1},  { 1329,{ 1377,    0,    0},1},
    { 1330,{ 1378,    0,    0},1},  { 1331,{ 1379,    0,    0},1},
    { 1332,{ 1380,    0,    0},1},  { 1333,{ 1381,    0,    0},1},
    { 1334,{ 1382,    0,    0},1},  { 1335,{ 1383,    0,    0},1},
    { 1336,{ 1384,    0,    0},1},  { 1337,{ 1385,    0,    0},1},
    { 1338,{ 1386,    0,    0},1},  { 1339,{ 1387,    0,    0},1},
    { 1340,{ 1388,    0,    0},1},  { 1341,{ 1389,    0,    0},1},
    { 1342,{ 1390,    0,    0},1},  { 1343,{ 1391,    0,    0},1},
    { 1344,{ 1392,    0,    0},1},  { 1345,{ 1393,    0,    0},1},
    { 1346,{ 1394,    0,    0},1},  { 1347,{ 1395,    0,    0},1},
    { 1348,{ 1396,    0,    0},1},  { 1349,{ 1397,    0,    0},1},
    { 1350,{ 1398,    0,    0},1},  { 1351,{ 1399,    0,    0},1},
    { 1352,{ 1400,    0,    0},1},  { 1353,{ 1401,    0,    0},1},
    { 1354,{ 1402,    0,    0},1},  { 1355,{ 1403,    0,    0},1},
    { 1356,{ 1404,    0,    0},1},  { 1357,{ 1405,    0,    0},1},
    { 1358,{ 1406,    0,    0},1},  { 1359,{ 1407,    0,    0},1},
    { 1360,{ 1408,    0,    0},1},  { 1361,{ 1409,    0,    0},1},
    { 1362,{ 1410,    0,    0},1},  { 1363,{ 1411,    0,    0},1},
    { 1364,{ 1412,    0,    0},1},  { 1365,{ 1413,    0,    0},1},
    { 1366,{ 1414,    0,    0},1},  { 4256,{11520,    0,    0},1},
    { 4257,{11521,    0,    0},1},  { 4258,{11522,    0,    0},1},
    { 4259,{11523,    0,    0},1},  { 4260,{11524,    0,    0},1},
    { 4261,{11525,    0,    0},1},  { 4262,{11526,    0,    0},1},
    { 4263,{11527,    0,    0},1},  { 4264,{11528,    0,    0},1},
    { 4265,{11529,    0,    0},1},  { 4266,{11530,    0,    0},1},
    { 4267,{11531,    0,    0},1},  { 4268,{11532,    0,    0},1},
    { 4269,{11533,    0,    0},1},  { 4270,{11534,    0,    0},1},
    { 4271,{11535,    0,    0},1},  { 4272,{11536,    0,    0},1},
    { 4273,{11537,    0,    0},1},  { 4274,{11538,    0,    0},1},
    { 4275,{11539,    0,    0},1},  { 4276,{11540,    0,    0},1},
    { 4277,{11541,    0,    0},1},  { 4278,{11542,    0,    0},1},
    { 4279,{11543,    0,    0},1},  { 4280,{11544,    0,    0},1},
    { 4281,{11545,    0,    0},1},  { 4282,{11546,    0,    0},1},
    { 4283,{11547,    0,    0},1},  { 4284,{11548,    0,    0},1},
    { 4285,{11549,    0,    0},1},  { 4286,{11550,    0,    0},1},
    { 4287,{11551,    0,    0},1},  { 4288,{11552,    0,    0},1},
    { 4289,{11553,    0,    0},1},  { 4290,{11554,    0,    0},1},
    { 4291,{11555,    0,    0},1},  { 4292,{11556,    0,    0},1},
    { 4293,{11557,    0,    0},1},  { 7680,{ 7681,    0,    0},1},
    { 7682,{ 7683,    0,    0},1},  { 7684,{ 7685,    0,    0},1},
    { 7686,{ 7687,    0,    0},1},  { 7688,{ 7689,    0,    0},1},
    { 7690,{ 7691,    0,    0},1},  { 7692,{ 7693,    0,    0},1},
    { 7694,{ 7695,    0,    0},1},  { 7696,{ 7697,    0,    0},1},
    { 7698,{ 7699,    0,    0},1},  { 7700,{ 7701,    0,    0},1},
    { 7702,{ 7703,    0,    0},1},  { 7704,{ 7705,    0,    0},1},
    { 7706,{ 7707,    0,    0},1},  { 7708,{ 7709,    0,    0},1},
    { 7710,{ 7711,    0,    0},1},  { 7712,{ 7713,    0,    0},1},
    { 7714,{ 7715,    0,    0},1},  { 7716,{ 7717,    0,    0},1},
    { 7718,{ 7719,    0,    0},1},  { 7720,{ 7721,    0,    0},1},
    { 7722,{ 7723,    0,    0},1},  { 7724,{ 7725,    0,    0},1},
    { 7726,{ 7727,    0,    0},1},  { 7728,{ 7729,    0,    0},1},
    { 7730,{ 7731,    0,    0},1},  { 7732,{ 7733,    0,    0},1},
    { 7734,{ 7735,    0,    0},1},  { 7736,{ 7737,    0,    0},1},
    { 7738,{ 7739,    0,    0},1},  { 7740,{ 7741,    0,    0},1},
    { 7742,{ 7743,    0,    0},1},  { 7744,{ 7745,    0,    0},1},
    { 7746,{ 7747,    0,    0},1},  { 7748,{ 7749,    0,    0},1},
    { 7750,{ 7751,    0,    0},1},  { 7752,{ 7753,    0,    0},1},
    { 7754,{ 7755,    0,    0},1},  { 7756,{ 7757,    0,    0},1},
    { 7758,{ 7759,    0,    0},1},  { 7760,{ 7761,    0,    0},1},
    { 7762,{ 7763,    0,    0},1},  { 7764,{ 7765,    0,    0},1},
    { 7766,{ 7767,    0,    0},1},  { 7768,{ 7769,    0,    0},1},
    { 7770,{ 7771,    0,    0},1},  { 7772,{ 7773,    0,    0},1},
    { 7774,{ 7775,    0,    0},1},  { 7776,{ 7777,    0,    0},1},
    { 7778,{ 7779,    0,    0},1},  { 7780,{ 7781,    0,    0},1},
    { 7782,{ 7783,    0,    0},1},  { 7784,{ 7785,    0,    0},1},
    { 7786,{ 7787,    0,    0},1},  { 7788,{ 7789,    0,    0},1},
    { 7790,{ 7791,    0,    0},1},  { 7792,{ 7793,    0,    0},1},
    { 7794,{ 7795,    0,    0},1},  { 7796,{ 7797,    0,    0},1},
    { 7798,{ 7799,    0,    0},1},  { 7800,{ 7801,    0,    0},1},
    { 7802,{ 7803,    0,    0},1},  { 7804,{ 7805,    0,    0},1},
    { 7806,{ 7807,    0,    0},1},  { 7808,{ 7809,    0,    0},1},
    { 7810,{ 7811,    0,    0},1},  { 7812,{ 7813,    0,    0},1},
    { 7814,{ 7815,    0,    0},1},  { 7816,{ 7817,    0,    0},1},
    { 7818,{ 7819,    0,    0},1},  { 7820,{ 7821,    0,    0},1},
    { 7822,{ 7823,    0,    0},1},  { 7824,{ 7825,    0,    0},1},
    { 7826,{ 7827,    0,    0},1},  { 7828,{ 7829,    0,    0},1},
    { 7835,{ 7777,    0,    0},1},  { 7840,{ 7841,    0,    0},1},
    { 7842,{ 7843,    0,    0},1},  { 7844,{ 7845,    0,    0},1},
    { 7846,{ 7847,    0,    0},1},  { 7848,{ 7849,    0,    0},1},
    { 7850,{ 7851,    0,    0},1},  { 7852,{ 7853,    0,    0},1},
    { 7854,{ 7855,    0,    0},1},  { 7856,{ 7857,    0,    0},1},
    { 7858,{ 7859,    0,    0},1},  { 7860,{ 7861,    0,    0},1},
    { 7862,{ 7863,    0,    0},1},  { 7864,{ 7865,    0,    0},1},
    { 7866,{ 7867,    0,    0},1},  { 7868,{ 7869,    0,    0},1},
    { 7870,{ 7871,    0,    0},1},  { 7872,{ 7873,    0,    0},1},
    { 7874,{ 7875,    0,    0},1},  { 7876,{ 7877,    0,    0},1},
    { 7878,{ 7879,    0,    0},1},  { 7880,{ 7881,    0,    0},1},
    { 7882,{ 7883,    0,    0},1},  { 7884,{ 7885,    0,    0},1},
    { 7886,{ 7887,    0,    0},1},  { 7888,{ 7889,    0,    0},1},
    { 7890,{ 7891,    0,    0},1},  { 7892,{ 7893,    0,    0},1},
    { 7894,{ 7895,    0,    0},1},  { 7896,{ 7897,    0,    0},1},
    { 7898,{ 7899,    0,    0},1},  { 7900,{ 7901,    0,    0},1},
    { 7902,{ 7903,    0,    0},1},  { 7904,{ 7905,    0,    0},1},
    { 7906,{ 7907,    0,    0},1},  { 7908,{ 7909,    0,    0},1},
    { 7910,{ 7911,    0,    0},1},  { 7912,{ 7913,    0,    0},1},
    { 7914,{ 7915,    0,    0},1},  { 7916,{ 7917,    0,    0},1},
    { 7918,{ 7919,    0,    0},1},  { 7920,{ 7921,    0,    0},1},
    { 7922,{ 7923,    0,    0},1},  { 7924,{ 7925,    0,    0},1},
    { 7926,{ 7927,    0,    0},1},  { 7928,{ 7929,    0,    0},1},
    { 7944,{ 7936,    0,    0},1},  { 7945,{ 7937,    0,    0},1},
    { 7946,{ 7938,    0,    0},1},  { 7947,{ 7939,    0,    0},1},
    { 7948,{ 7940,    0,    0},1},  { 7949,{ 7941,    0,    0},1},
    { 7950,{ 7942,    0,    0},1},  { 7951,{ 7943,    0,    0},1},
    { 7960,{ 7952,    0,    0},1},  { 7961,{ 7953,    0,    0},1},
    { 7962,{ 7954,    0,    0},1},  { 7963,{ 7955,    0,    0},1},
    { 7964,{ 7956,    0,    0},1},  { 7965,{ 7957,    0,    0},1},
    { 7976,{ 7968,    0,    0},1},  { 7977,{ 7969,    0,    0},1},
    { 7978,{ 7970,    0,    0},1},  { 7979,{ 7971,    0,    0},1},
    { 7980,{ 7972,    0,    0},1},  { 7981,{ 7973,    0,    0},1},
    { 7982,{ 7974,    0,    0},1},  { 7983,{ 7975,    0,    0},1},
    { 7992,{ 7984,    0,    0},1},  { 7993,{ 7985,    0,    0},1},
    { 7994,{ 7986,    0,    0},1},  { 7995,{ 7987,    0,    0},1},
    { 7996,{ 7988,    0,    0},1},  { 7997,{ 7989,    0,    0},1},
    { 7998,{ 7990,    0,    0},1},  { 7999,{ 7991,    0,    0},1},
    { 8008,{ 8000,    0,    0},1},  { 8009,{ 8001,    0,    0},1},
    { 8010,{ 8002,    0,    0},1},  { 8011,{ 8003,    0,    0},1},
    { 8012,{ 8004,    0,    0},1},  { 8013,{ 8005,    0,    0},1},
    { 8025,{ 8017,    0,    0},1},  { 8027,{ 8019,    0,    0},1},
    { 8029,{ 8021,    0,    0},1},  { 8031,{ 8023,    0,    0},1},
    { 8040,{ 8032,    0,    0},1},  { 8041,{ 8033,    0,    0},1},
    { 8042,{ 8034,    0,    0},1},  { 8043,{ 8035,    0,    0},1},
    { 8044,{ 8036,    0,    0},1},  { 8045,{ 8037,    0,    0},1},
    { 8046,{ 8038,    0,    0},1},  { 8047,{ 8039,    0,    0},1},
    { 8120,{ 8112,    0,    0},1},  { 8121,{ 8113,    0,    0},1},
    { 8122,{ 8048,    0,    0},1},  { 8123,{ 8049,    0,    0},1},
    { 8126,{  953,    0,    0},1},  { 8136,{ 8050,    0,    0},1},
    { 8137,{ 8051,    0,    0},1},  { 8138,{ 8052,    0,    0},1},
    { 8139,{ 8053,    0,    0},1},  { 8152,{ 8144,    0,    0},1},
    { 8153,{ 8145,    0,    0},1},  { 8154,{ 8054,    0,    0},1},
    { 8155,{ 8055,    0,    0},1},  { 8168,{ 8160,    0,    0},1},
    { 8169,{ 8161,    0,    0},1},  { 8170,{ 8058,    0,    0},1},
    { 8171,{ 8059,    0,    0},1},  { 8172,{ 8165,    0,    0},1},
    { 8184,{ 8056,    0,    0},1},  { 8185,{ 8057,    0,    0},1},
    { 8186,{ 8060,    0,    0},1},  { 8187,{ 8061,    0,    0},1},
    { 8486,{  969,    0,    0},1},  { 8490,{  107,    0,    0},1},
    { 8491,{  229,    0,    0},1},  { 8544,{ 8560,    0,    0},1},
    { 8545,{ 8561,    0,    0},1},  { 8546,{ 8562,    0,    0},1},
    { 8547,{ 8563,    0,    0},1},  { 8548,{ 8564,    0,    0},1},
    { 8549,{ 8565,    0,    0},1},  { 8550,{ 8566,    0,    0},1},
    { 8551,{ 8567,    0,    0},1},  { 8552,{ 8568,    0,    0},1},
    { 8553,{ 8569,    0,    0},1},  { 8554,{ 8570,    0,    0},1},
    { 8555,{ 8571,    0,    0},1},  { 8556,{ 8572,    0,    0},1},
    { 8557,{ 8573,    0,    0},1},  { 8558,{ 8574,    0,    0},1},
    { 8559,{ 8575,    0,    0},1},  { 9398,{ 9424,    0,    0},1},
    { 9399,{ 9425,    0,    0},1},  { 9400,{ 9426,    0,    0},1},
    { 9401,{ 9427,    0,    0},1},  { 9402,{ 9428,    0,    0},1},
    { 9403,{ 9429,    0,    0},1},  { 9404,{ 9430,    0,    0},1},
    { 9405,{ 9431,    0,    0},1},  { 9406,{ 9432,    0,    0},1},
    { 9407,{ 9433,    0,    0},1},  { 9408,{ 9434,    0,    0},1},
    { 9409,{ 9435,    0,    0},1},  { 9410,{ 9436,    0,    0},1},
    { 9411,{ 9437,    0,    0},1},  { 9412,{ 9438,    0,    0},1},
    { 9413,{ 9439,    0,    0},1},  { 9414,{ 9440,    0,    0},1},
    { 9415,{ 9441,    0,    0},1},  { 9416,{ 9442,    0,    0},1},
    { 9417,{ 9443,    0,    0},1},  { 9418,{ 9444,    0,    0},1},
    { 9419,{ 9445,    0,    0},1},  { 9420,{ 9446,    0,    0},1},
    { 9421,{ 9447,    0,    0},1},  { 9422,{ 9448,    0,    0},1},
    { 9423,{ 9449,    0,    0},1},  {11264,{11312,    0,    0},1},
    {11265,{11313,    0,    0},1},  {11266,{11314,    0,    0},1},
    {11267,{11315,    0,    0},1},  {11268,{11316,    0,    0},1},
    {11269,{11317,    0,    0},1},  {11270,{11318,    0,    0},1},
    {11271,{11319,    0,    0},1},  {11272,{11320,    0,    0},1},
    {11273,{11321,    0,    0},1},  {11274,{11322,    0,    0},1},
    {11275,{11323,    0,    0},1},  {11276,{11324,    0,    0},1},
    {11277,{11325,    0,    0},1},  {11278,{11326,    0,    0},1},
    {11279,{11327,    0,    0},1},  {11280,{11328,    0,    0},1},
    {11281,{11329,    0,    0},1},  {11282,{11330,    0,    0},1},
    {11283,{11331,    0,    0},1},  {11284,{11332,    0,    0},1},
    {11285,{11333,    0,    0},1},  {11286,{11334,    0,    0},1},
    {11287,{11335,    0,    0},1},  {11288,{11336,    0,    0},1},
    {11289,{11337,    0,    0},1},  {11290,{11338,    0,    0},1},
    {11291,{11339,    0,    0},1},  {11292,{11340,    0,    0},1},
    {11293,{11341,    0,    0},1},  {11294,{11342,    0,    0},1},
    {11295,{11343,    0,    0},1},  {11296,{11344,    0,    0},1},
    {11297,{11345,    0,    0},1},  {11298,{11346,    0,    0},1},
    {11299,{11347,    0,    0},1},  {11300,{11348,    0,    0},1},
    {11301,{11349,    0,    0},1},  {11302,{11350,    0,    0},1},
    {11303,{11351,    0,    0},1},  {11304,{11352,    0,    0},1},
    {11305,{11353,    0,    0},1},  {11306,{11354,    0,    0},1},
    {11307,{11355,    0,    0},1},  {11308,{11356,    0,    0},1},
    {11309,{11357,    0,    0},1},  {11310,{11358,    0,    0},1},
    {11392,{11393,    0,    0},1},  {11394,{11395,    0,    0},1},
    {11396,{11397,    0,    0},1},  {11398,{11399,    0,    0},1},
    {11400,{11401,    0,    0},1},  {11402,{11403,    0,    0},1},
    {11404,{11405,    0,    0},1},  {11406,{11407,    0,    0},1},
    {11408,{11409,    0,    0},1},  {11410,{11411,    0,    0},1},
    {11412,{11413,    0,    0},1},  {11414,{11415,    0,    0},1},
    {11416,{11417,    0,    0},1},  {11418,{11419,    0,    0},1},
    {11420,{11421,    0,    0},1},  {11422,{11423,    0,    0},1},
    {11424,{11425,    0,    0},1},  {11426,{11427,    0,    0},1},
    {11428,{11429,    0,    0},1},  {11430,{11431,    0,    0},1},
    {11432,{11433,    0,    0},1},  {11434,{11435,    0,    0},1},
    {11436,{11437,    0,    0},1},  {11438,{11439,    0,    0},1},
    {11440,{11441,    0,    0},1},  {11442,{11443,    0,    0},1},
    {11444,{11445,    0,    0},1},  {11446,{11447,    0,    0},1},
    {11448,{11449,    0,    0},1},  {11450,{11451,    0,    0},1},
    {11452,{11453,    0,    0},1},  {11454,{11455,    0,    0},1},
    {11456,{11457,    0,    0},1},  {11458,{11459,    0,    0},1},
    {11460,{11461,    0,    0},1},  {11462,{11463,    0,    0},1},
    {11464,{11465,    0,    0},1},  {11466,{11467,    0,    0},1},
    {11468,{11469,    0,    0},1},  {11470,{11471,    0,    0},1},
    {11472,{11473,    0,    0},1},  {11474,{11475,    0,    0},1},
    {11476,{11477,    0,    0},1},  {11478,{11479,    0,    0},1},
    {11480,{11481,    0,    0},1},  {11482,{11483,    0,    0},1},
    {11484,{11485,    0,    0},1},  {11486,{11487,    0,    0},1},
    {11488,{11489,    0,    0},1},  {11490,{11491,    0,    0},1},
    {65313,{65345,    0,    0},1},  {65314,{65346,    0,    0},1},
    {65315,{65347,    0,    0},1},  {65316,{65348,    0,    0},1},
    {65317,{65349,    0,    0},1},  {65318,{65350,    0,    0},1},
    {65319,{65351,    0,    0},1},  {65320,{65352,    0,    0},1},
    {65321,{65353,    0,    0},1},  {65322,{65354,    0,    0},1},
    {65323,{65355,    0,    0},1},  {65324,{65356,    0,    0},1},
    {65325,{65357,    0,    0},1},  {65326,{65358,    0,    0},1},
    {65327,{65359,    0,    0},1},  {65328,{65360,    0,    0},1},
    {65329,{65361,    0,    0},1},  {65330,{65362,    0,    0},1},
    {65331,{65363,    0,    0},1},  {65332,{65364,    0,    0},1},
    {65333,{65365,    0,    0},1},  {65334,{65366,    0,    0},1},
    {65335,{65367,    0,    0},1},  {65336,{65368,    0,    0},1},
    {65337,{65369,    0,    0},1},  {65338,{65370,    0,    0},1},
    {66560,{66600,    0,    0},1},  {66561,{66601,    0,    0},1},
    {66562,{66602,    0,    0},1},  {66563,{66603,    0,    0},1},
    {66564,{66604,    0,    0},1},  {66565,{66605,    0,    0},1},
    {66566,{66606,    0,    0},1},  {66567,{66607,    0,    0},1},
    {66568,{66608,    0,    0},1},  {66569,{66609,    0,    0},1},
    {66570,{66610,    0,    0},1},  {66571,{66611,    0,    0},1},
    {66572,{66612,    0,    0},1},  {66573,{66613,    0,    0},1},
    {66574,{66614,    0,    0},1},  {66575,{66615,    0,    0},1},
    {66576,{66616,    0,    0},1},  {66577,{66617,    0,    0},1},
    {66578,{66618,    0,    0},1},  {66579,{66619,    0,    0},1},
    {66580,{66620,    0,    0},1},  {66581,{66621,    0,    0},1},
    {66582,{66622,    0,    0},1},  {66583,{66623,    0,    0},1},
    {66584,{66624,    0,    0},1},  {66585,{66625,    0,    0},1},
    {66586,{66626,    0,    0},1},  {66587,{66627,    0,    0},1},
    {66588,{66628,    0,    0},1},  {66589,{66629,    0,    0},1},
    {66590,{66630,    0,    0},1},  {66591,{66631,    0,    0},1},
    {66592,{66632,    0,    0},1},  {66593,{66633,    0,    0},1},
    {66594,{66634,    0,    0},1},  {66595,{66635,    0,    0},1},
    {66596,{66636,    0,    0},1},  {66597,{66637,    0,    0},1},
    {66598,{66638,    0,    0},1},  {66599,{66639,    0,    0},1}
  };

  // Full case-folding, mappings that cause strings to grow in length
  static const _TCaseFoldingMap _fcf[] = {
    {  223,{  115,  115,    0},2},  {  304,{  105,  775,    0},2},
    {  329,{  700,  110,    0},2},  {  496,{  106,  780,    0},2},
    {  912,{  953,  776,  769},3},  {  944,{  965,  776,  769},3},
    { 1415,{ 1381, 1410,    0},2},  { 7830,{  104,  817,    0},2},
    { 7831,{  116,  776,    0},2},  { 7832,{  119,  778,    0},2},
    { 7833,{  121,  778,    0},2},  { 7834,{   97,  702,    0},2},
    { 8016,{  965,  787,    0},2},  { 8018,{  965,  787,  768},3},
    { 8020,{  965,  787,  769},3},  { 8022,{  965,  787,  834},3},
    { 8064,{ 7936,  953,    0},2},  { 8065,{ 7937,  953,    0},2},
    { 8066,{ 7938,  953,    0},2},  { 8067,{ 7939,  953,    0},2},
    { 8068,{ 7940,  953,    0},2},  { 8069,{ 7941,  953,    0},2},
    { 8070,{ 7942,  953,    0},2},  { 8071,{ 7943,  953,    0},2},
    { 8072,{ 7936,  953,    0},2},  { 8073,{ 7937,  953,    0},2},
    { 8074,{ 7938,  953,    0},2},  { 8075,{ 7939,  953,    0},2},
    { 8076,{ 7940,  953,    0},2},  { 8077,{ 7941,  953,    0},2},
    { 8078,{ 7942,  953,    0},2},  { 8079,{ 7943,  953,    0},2},
    { 8080,{ 7968,  953,    0},2},  { 8081,{ 7969,  953,    0},2},
    { 8082,{ 7970,  953,    0},2},  { 8083,{ 7971,  953,    0},2},
    { 8084,{ 7972,  953,    0},2},  { 8085,{ 7973,  953,    0},2},
    { 8086,{ 7974,  953,    0},2},  { 8087,{ 7975,  953,    0},2},
    { 8088,{ 7968,  953,    0},2},  { 8089,{ 7969,  953,    0},2},
    { 8090,{ 7970,  953,    0},2},  { 8091,{ 7971,  953,    0},2},
    { 8092,{ 7972,  953,    0},2},  { 8093,{ 7973,  953,    0},2},
    { 8094,{ 7974,  953,    0},2},  { 8095,{ 7975,  953,    0},2},
    { 8096,{ 8032,  953,    0},2},  { 8097,{ 8033,  953,    0},2},
    { 8098,{ 8034,  953,    0},2},  { 8099,{ 8035,  953,    0},2},
    { 8100,{ 8036,  953,    0},2},  { 8101,{ 8037,  953,    0},2},
    { 8102,{ 8038,  953,    0},2},  { 8103,{ 8039,  953,    0},2},
    { 8104,{ 8032,  953,    0},2},  { 8105,{ 8033,  953,    0},2},
    { 8106,{ 8034,  953,    0},2},  { 8107,{ 8035,  953,    0},2},
    { 8108,{ 8036,  953,    0},2},  { 8109,{ 8037,  953,    0},2},
    { 8110,{ 8038,  953,    0},2},  { 8111,{ 8039,  953,    0},2},
    { 8114,{ 8048,  953,    0},2},  { 8115,{  945,  953,    0},2},
    { 8116,{  940,  953,    0},2},  { 8118,{  945,  834,    0},2},
    { 8119,{  945,  834,  953},3},  { 8124,{  945,  953,    0},2},
    { 8130,{ 8052,  953,    0},2},  { 8131,{  951,  953,    0},2},
    { 8132,{  942,  953,    0},2},  { 8134,{  951,  834,    0},2},
    { 8135,{  951,  834,  953},3},  { 8140,{  951,  953,    0},2},
    { 8146,{  953,  776,  768},3},  { 8147,{  953,  776,  769},3},
    { 8150,{  953,  834,    0},2},  { 8151,{  953,  776,  834},3},
    { 8162,{  965,  776,  768},3},  { 8163,{  965,  776,  769},3},
    { 8164,{  961,  787,    0},2},  { 8166,{  965,  834,    0},2},
    { 8167,{  965,  776,  834},3},  { 8178,{ 8060,  953,    0},2},
    { 8179,{  969,  953,    0},2},  { 8180,{  974,  953,    0},2},
    { 8182,{  969,  834,    0},2},  { 8183,{  969,  834,  953},3},
    { 8188,{  969,  953,    0},2},  {64256,{  102,  102,    0},2},
    {64257,{  102,  105,    0},2},  {64258,{  102,  108,    0},2},
    {64259,{  102,  102,  105},3},  {64260,{  102,  102,  108},3},
    {64261,{  115,  116,    0},2},  {64262,{  115,  116,    0},2},
    {64275,{ 1396, 1398,    0},2},  {64276,{ 1396, 1381,    0},2},
    {64277,{ 1396, 1387,    0},2},  {64278,{ 1406, 1398,    0},2},
    {64279,{ 1396, 1389,    0},2}
  };

  // Simple case-folding, mappings to single characters (different from full case-folding)
  static const _TCaseFoldingMap _scf[] = {
    { 8072,{ 8064,    0,    0},1},  { 8073,{ 8065,    0,    0},1},
    { 8074,{ 8066,    0,    0},1},  { 8075,{ 8067,    0,    0},1},
    { 8076,{ 8068,    0,    0},1},  { 8077,{ 8069,    0,    0},1},
    { 8078,{ 8070,    0,    0},1},  { 8079,{ 8071,    0,    0},1},
    { 8088,{ 8080,    0,    0},1},  { 8089,{ 8081,    0,    0},1},
    { 8090,{ 8082,    0,    0},1},  { 8091,{ 8083,    0,    0},1},
    { 8092,{ 8084,    0,    0},1},  { 8093,{ 8085,    0,    0},1},
    { 8094,{ 8086,    0,    0},1},  { 8095,{ 8087,    0,    0},1},
    { 8104,{ 8096,    0,    0},1},  { 8105,{ 8097,    0,    0},1},
    { 8106,{ 8098,    0,    0},1},  { 8107,{ 8099,    0,    0},1},
    { 8108,{ 8100,    0,    0},1},  { 8109,{ 8101,    0,    0},1},
    { 8110,{ 8102,    0,    0},1},  { 8111,{ 8103,    0,    0},1},
    { 8124,{ 8115,    0,    0},1},  { 8140,{ 8131,    0,    0},1},
    { 8188,{ 8179,    0,    0},1}
  };

  // Special case-folding, for example: uppercase I and dotted uppercase I in Turkic languages (tr & az locales)
  static const _TCaseFoldingMap _tcf[] = {
    {   73,{  305,    0,    0},1},  {  304,{  105,    0,    0},1}
  };

  // Structure for containing the needed helper function
  struct _SH {
    static bool _mapChar(CUString &dst_, CUString::Utf32 src_, const _TCaseFoldingMap map[], uint32_t mapSize)
    {
      for(size_t j = 0; j < mapSize; ++j) {
        if(map[j].from == src_) {
          for(size_t k = 0; k < map[j].toCount; ++k) dst_.push_back(map[j].to[k]);
          return(true);
        }
      }
      return(false);
    }
  };

  // Clear the destination first
  dst.clear();

  // Map using special case-folding
  if(specialCaseFolding) {
    // Map using full case-folding
    if(fullCaseFolding) {
      for(size_t i = 0; i < src.length(); ++i) {
        if(_SH::_mapChar(dst, src[i], _tcf, elemsize(_tcf))) continue;
        if(_SH::_mapChar(dst, src[i], _ccf, elemsize(_ccf))) continue;
        if(_SH::_mapChar(dst, src[i], _fcf, elemsize(_fcf))) continue;
        dst.push_back(src[i]);
      }
    }
    // Map using simple case-folding
    else {
      for(size_t i = 0; i < src.length(); ++i) {
        if(_SH::_mapChar(dst, src[i], _tcf, elemsize(_tcf))) continue;
        if(_SH::_mapChar(dst, src[i], _ccf, elemsize(_ccf))) continue;
        if(_SH::_mapChar(dst, src[i], _scf, elemsize(_scf))) continue;
        dst.push_back(src[i]);
      }
    }
  }
  // Map without using special case-folding
  else {
    // Map using full case-folding
    if(fullCaseFolding) {
      for(size_t i = 0; i < src.length(); ++i) {
        if(_SH::_mapChar(dst, src[i], _ccf, elemsize(_ccf))) continue;
        if(_SH::_mapChar(dst, src[i], _fcf, elemsize(_fcf))) continue;
        dst.push_back(src[i]);
      }
    }
    // Map using simple case-folding
    else {
      for(size_t i = 0; i < src.length(); ++i) {
        if(_SH::_mapChar(dst, src[i], _ccf, elemsize(_ccf))) continue;
        if(_SH::_mapChar(dst, src[i], _scf, elemsize(_scf))) continue;
        dst.push_back(src[i]);
      }
    }
  }
}

} // namespace vge
