/** \file VGch.h */
// ############################################################################################
//  VGch.h
//  Definition of the most basic stuffs
//  This file is meant to be processed to become a precompiled header (gch)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___VGCH_H_
#define _BASE___VGCH_H_

#ifndef _DOXYGEN_SHOULD_IGNORE_THIS_
#  ifndef _XOPEN_SOURCE
#    define _XOPEN_SOURCE 600
#  endif
#endif

// Include the configuration header
#include "../VConfig.h"

// C++ headers
#include <exception>
#include <iosfwd>
#include <istream>
#include <new>
#include <list>
#include <map>
#include <ostream>
#include <set>
#include <string>
#include <vector>

// C++ enabled C headers
#include <cerrno>
#include <clocale>
#include <cmath>
#include <csignal>
#include <cstdarg>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>

// Unix standard headers
#include <dirent.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdint.h>
#include <termios.h>
#include <unistd.h>

// Unix system headers
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/shm.h>
#include <sys/syscall.h>
#include <sys/time.h>

// NOTE: If somehow VortexGE does not compile in your system,
//       try to uncomment these lines :
//#include <sys/select.h>
//#include <sys/stat.h>
//#include <sys/types.h>


//
// Versioning macros
//
//!
//! \addtogroup Versioning_Macros
//! @{

//!
//! \hideinitializer
//! Create a complete/combined version number (MJ.MN.RV) for VortexGE.
#define __VG_COMB_VER__(MJ, MN, RV) ((uint32_t(MJ) << 24) | (uint32_t(MN) << 16) | (uint32_t(RV) << 8))

//!
//! \hideinitializer
//! Define the current major version of VortexGE.
#define __VG_MAJOR__    0
//!
//! \hideinitializer
//! Define the current minor version of VortexGE.
#define __VG_MINOR__    6
//!
//! \hideinitializer
//! Define the current revision version of VortexGE.
#define __VG_REVISION__ 7
//!
//! \hideinitializer
//! Define the current combined version of VortexGE.
#define __VG_VERSION__  __VG_COMB_VER__(__VG_MAJOR__, __VG_MINOR__, __VG_REVISION__)

//!
//! \hideinitializer
//! Check for the minimum required version of VortexGE.
#define __VG_MIN_VER__(MJ, MN, RV) (__VG_VERSION__ >= __VG_COMB_VER__(MJ, MN, RV))

//!
//! \hideinitializer
//! Check for the minimum required version of GCC.
#define __GCC_MIN_VER__(V) ((__GNUC__ * 10000 + __GNUC_MINOR__ * 100 + __GNUC_PATCHLEVEL__) >= (V))

//! @}
//!

//
// Architecture dependent macros
//

// Optimization macros
#if defined(__GNUC__) && defined(__USE_ASM__)
#  if defined(__i386__)
#    define _ASM_A32_
#  endif
#  if defined(__x86_64__)
#    define _ASM_A64_
#  endif
#endif

#if defined(_ASM_A32_) || defined(_ASM_A64_)
#  define _ASM_ASM_
#  if defined(__USE_MMX__)
#    define _ASM_MMX_
#  endif
#  if defined(__USE_SSE__)
#    define _ASM_SSE_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_SSE2__)
#    define _ASM_SSE2_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_SSE3__)
#    define _ASM_SSE3_
#    define _ALIGN_128_BITS_
#  endif
#  if defined(__USE_3DNOW__)
#    define _ASM_3DNOW_
#  endif
#endif

// Alignment directive for buffer allocation
#if defined(__USE_ALIGNED_MEMOPS__) && defined(_ALIGN_128_BITS_)
#  define _ALIGN_BUFF_MEM_
#endif

// Alignment attribute, needed for declaring memory-aligned variables
//!
//! \ingroup Arch_Dep_Macros
//! \hideinitializer
//! Directive to make the variable's memory address is specially aligned for SSE support.\n\n
//! <b>Example:</b>
//! \code
//! uint32_t _sse var1[4];
//! uint32_t      var2[4];
//! \endcode
//! \n
//! In the above example:
//! - If VortexGE is built <b>with</b> SSE support, <code>var1</code> will be specially aligned to support SSE but <code>var2</code> will be aligned using the compiler's default alignment.
//! - If VortexGE is built <b>without</b> SSE support, both <code>var1</code> and <code>var2</code> will be aligned using the compiler's default alignment.
//!
//! <b>Note that this macro will be undefined using \#undef (thus become unusable
//! by one's applications), unless before including \ref VMain or VMain.h one has
//! already define the macro _KEEP_INTERNAL_MACROS_DEFINED.</b>
#if defined(_ASM_SSE_) || defined(_ASM_SSE2_) || defined(_ASM_SSE3_)
#  define _sse __attribute__((aligned(16)))
#else
#  define _sse
#endif


//
// These macros below are defined for shortcut & usage simplicity
//
//!
//! \addtogroup Shortcut_Macros
//! @{

// Convenience macros for calculating size
//!
//! \hideinitializer
//! Calculate the number of elements in an array.
#define elemsize(X) (sizeof(X) / sizeof(*X))
//!
//! \hideinitializer
//! Calculate the bit size of a type.
//! \n\n
//! This macro assumes that bit size of a 'char' is 8.
//! It is safe since VortexGE will not build if 'sizeof(char)' is not 1.
#define bitsize(X) (sizeof(X) * 8)

// Special C++ new operator that will not throw exception
//!
//! \hideinitializer
//! Special C++ new operator that will not throw exception.
#define _nothrow_new new(nothrow)

// Structure/union/data attributes
//!
//! \hideinitializer
//! Used by the library to pack data structures.
#define _attr_packed __attribute__((packed))
//!
//! \hideinitializer
//! Used by the library to declare SIMD (vector) variables/data types.
#define __attr_vector_size(X) __attribute__((vector_size(X)))

// Special function attributes
//!
//! \hideinitializer
//! Used by the library to mark functions that accept printf() style formatting characters.
//! This attribute will cause GCC to check the formatting characters for consistency.
#define _attr_format_printf(fsindex) __attribute__((format(printf,fsindex,fsindex+1)))
//!
//! \hideinitializer
//! Used by the library to tell the compiler about deprecated features.
#define _attr_deprecated __attribute__((deprecated))
//!
//! \hideinitializer
//! Used by the library to tell the compiler about unused features.
#define _attr_unused __attribute__((unused))
//!
//! \hideinitializer
//! Used by the library to mark functions which must be always inlined when compiled.
#define _attr_force_inline __attribute__((always_inline))

// Function optimization attributes
//!
//! \hideinitializer
//! GCC manual says:\n\n<I>
//! Many functions have no effects except the return value and their return value depends
//! only on the parameters and/or global variables. Such a function can be subject to
//! common subexpression elimination and loop optimization just as an arithmetic operator
//! would be. These functions should be declared with the attribute pure.</I>
#define _attr_pure __attribute__((pure))
//!
//! \hideinitializer
//! GCC manual says:\n\n<I>
//! Many functions do not examine any values except their parameters, and have no effects
//! except the return value. Basically this is just slightly more strict class than the
//! pure attribute above, since function is not allowed to read global memory.\n\n
//! Note that a function that has pointer parameters and examines the data pointed to must
//! not be declared const. Likewise, a function that calls a non-const function usually
//! must not be const. It does not make sense for a const function to return void.</I>
#define _attr_const __attribute__((const))
//!
//! \hideinitializer
//! GCC manual says:\n\n<I>
//! The malloc attribute is used to tell the compiler that a function may be treated as
//! if any non-NULL pointer it returns cannot alias any other pointer valid when the
//! function returns. This will often improve optimization. Standard functions with this
//! property include malloc and calloc. realloc-like functions have this property as long
//! as the old pointer is never referred to (including comparing it to the new pointer)
//! after the function returns a non-NULL value.
#define _attr_malloc __attribute__((malloc))
//!
//! \hideinitializer
//! Used by the library to mark functions that will never throw, just like C++'s throw().
#define _attr_nothrow __attribute__((nothrow))
//!
//! \hideinitializer
//! Used by the library to mark functions that will never return.
#define _attr_noreturn __attribute__((noreturn))
//!
//! \hideinitializer
//! Used by the library to mark functions so that GCC will optimize for faster calls.
#define _attr_fastcall __attribute__((fastcall))

// ELF visibility optimization attributes
//!
//! \hideinitializer
//! Used by the library to mark exported symbols.
//! Unmarked symbols are considered to be for internal use.
//! \n\n
//! <b>Note that this macro will be undefined using \#undef (thus become unusable
//!  by one's applications), unless before including \ref VMain or VMain.h one has
//! already define the macro _KEEP_INTERNAL_MACROS_DEFINED.</b>
#if __GCC_MIN_VER__(40000)
#  define VGE_EXPORT __attribute__((visibility("default")))
#else
#  define VGE_EXPORT
#endif

// Branch prediction optimization attributes
//!
//! \hideinitializer
//! Use this for manual branch prediction optimization.
#define _if_likely(X) if(__builtin_expect(!!(X), 1))
//!
//! \hideinitializer
//! Use this for manual branch prediction optimization.
#define _if_unlikely(X) if(__builtin_expect(!!(X), 0))

// Source file informations
//!
//! \hideinitializer
//! This macro will expand to a char* containing the current source file name.
#define _F_ __FILE__
//!
//! \hideinitializer
//! This macro will expand to an integer containing the current line number.
#define _L_ __LINE__
//!
//! \hideinitializer
//! This macro will expand to a char* containing the current function name.
#define _N_ __func__
//!
//! \hideinitializer
//! This macro will expand to the current source file name and the current line number,
//! separated by comma.
#define _FL_ __FILE__, __LINE__
//!
//! \hideinitializer
//! This macro will expand the current function name, the current source file name
//! and the current line number, separated by commas.
#define _NFL_ __func__, __FILE__, __LINE__

//! @}
//!

namespace vge {

  // Bring the std namespace into the scope of vge namespace
  using namespace std;


  //
  // Helper functions for clipping values
  //

  //!
  //! \ingroup SHF_ClippingValues
  //! Use this to clip the given value in the given minimum and maximum range.
  template <typename ValueT> inline
  const ValueT clip_value(const ValueT &min, const ValueT &val, const ValueT &max)
  { return( (val<min) ? min : ((val>max) ? max : val) ); }


  //
  // Helper functions for copying values
  //
  //!
  //! \addtogroup SHF_CopyingValues
  //! @{

  //!
  //! Use this to perform safe string copy with very little overhead.
  inline void safe_strcpy(char *dst, const char *src, size_t dstBuffSize)
  {
    while(*src && --dstBuffSize) *dst++ = *src++;
    *dst = 0;
  }

  //!
  //! Use this to copy elements from the range [src, src + n)
  //! to the range [dst, dst + n).\n
  //!
  //! Basicly, it performs the assignments *dst = *src, *(dst + 1) = *(src + 1),
  //! and so on in forward order (i.e. in order of increasing n).
  template<typename SrcIteratorT, typename SizeT, typename DstIteratorT> inline
  void copy_n(SrcIteratorT src, SizeT n, DstIteratorT dst)
  { for(; n; --n, ++dst, ++src) *dst = *src; }

  //!
  //! Use this to copy n bytes <b>(not elements)</b> from vector<SrcDataT> to DstDataT*.
  //! \warning Only use if SrcDataT and DstDataT 100% safe with bitwise copying (for example: if they are char/uint32_t/etc.).
  template<typename SrcDataT, typename SizeT, typename DstDataT> inline
  void copy_b(const vector<SrcDataT> &src, SizeT n, DstDataT *dst)
  { memcpy(dst, &src[0], n); }

  //!
  //! Use this to copy n bytes <b>(not elements)</b> from SrcDataT* to vector<DstDataT>.
  //! \warning Only use if SrcDataT and DstDataT 100% safe with bitwise copying (for example: if they are char/uint32_t/etc.).
  template<typename SrcDataT, typename SizeT, typename DstDataT> inline
  void copy_b(const SrcDataT *src, SizeT n, vector<DstDataT> &dst)
  { memcpy(&dst[0], src, n); }

  //!
  //! Use this to copy n bytes <b>(not elements)</b> from vector<SrcDataT> to vector<DstDataT>.
  //! \warning Only use if SrcDataT and DstDataT 100% safe with bitwise copying (for example: if they are char/uint32_t/etc.).
  template<typename SrcDataT, typename SizeT, typename DstDataT> inline
  void copy_b(const vector<SrcDataT> &src, SizeT n, vector<DstDataT> &dst)
  { memcpy(&dst[0], &src[0], n); }

  //!
  //! Use this to copy n bytes <b>(not elements)</b> from SrcDataT* to DstDataT*.
  //! \warning Only use if SrcDataT and DstDataT 100% safe with bitwise copying (for example: if they are char/uint32_t/etc.).
  template<typename SrcDataT, typename SizeT, typename DstDataT> inline
  void copy_b(const SrcDataT *src, SizeT n, DstDataT *dst)
  { memcpy(dst, src, n); }

  //! @}
  //!

  //
  // Unused parameter warning suppressor
  //
  //!
  //! \addtogroup SHF_UnParmSuppressor
  //! @{

  //!
  //! Use this to suppress "unused parameter" warning for 1 parameter.
  template <typename T> inline
  bool UNUSED_PARAM(const T &A)
  { return(&A == &A); }

  //!
  //! Use this to suppress "unused parameter" warning for 2 parameters.
  template <typename T1, typename T2> inline
  bool UNUSED_PARAMS_2(const T1 &A, const T2 &B)
  { return(UNUSED_PARAM(A) & UNUSED_PARAM(B)); }

  //!
  //! Use this to suppress "unused parameter" warning for 3 parameters.
  template <typename T1, typename T2, typename T3> inline
  bool UNUSED_PARAMS_3(const T1 &A, const T2 &B, const T3 &C)
  { return(UNUSED_PARAMS_2(A, B) & UNUSED_PARAM(C)); }

  //!
  //! Use this to suppress "unused parameter" warning for 4 parameters.
  template <typename T1, typename T2, typename T3, typename T4> inline
  bool UNUSED_PARAMS_4(const T1 &A, const T2 &B, const T3 &C, const T4 &D)
  { return(UNUSED_PARAMS_3(A, B, C) & UNUSED_PARAM(D)); }

  //!
  //! Use this to suppress "unused parameter" warning for 5 parameters.
  template <typename T1, typename T2, typename T3, typename T4, typename T5> inline
  bool UNUSED_PARAMS_5(const T1 &A, const T2 &B, const T3 &C, const T4 &D, const T5 &E)
  { return(UNUSED_PARAMS_4(A, B, C, D) & UNUSED_PARAM(E)); }

  //!
  //! Use this to suppress "unused parameter" warning for 6 parameters.
  template <typename T1, typename T2, typename T3, typename T4, typename T5, typename T6> inline
  bool UNUSED_PARAMS_6(const T1 &A, const T2 &B, const T3 &C, const T4 &D, const T5 &E, const T6 &F)
  { return(UNUSED_PARAMS_5(A, B, C, D, E) & UNUSED_PARAM(F)); }

  //!
  //! Use this to suppress "unused parameter" warning for 7 parameters.
  template <typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7> inline
  bool UNUSED_PARAMS_7(const T1 &A, const T2 &B, const T3 &C, const T4 &D, const T5 &E, const T6 &F, const T7 &G)
  { return(UNUSED_PARAMS_6(A, B, C, D, E, F) & UNUSED_PARAM(G)); }

  //!
  //! Use this to suppress "unused parameter" warning for 8 parameters.
  template <typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8> inline
  bool UNUSED_PARAMS_8(const T1 &A, const T2 &B, const T3 &C, const T4 &D, const T5 &E, const T6 &F, const T7 &G, const T8 &H)
  { return(UNUSED_PARAMS_7(A, B, C, D, E, F, G) & UNUSED_PARAM(H)); }

  //!
  //! Use this to suppress "unused parameter" warning for 9 parameters.
  template <typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9> inline
  bool UNUSED_PARAMS_9(const T1 &A, const T2 &B, const T3 &C, const T4 &D, const T5 &E, const T6 &F, const T7 &G, const T8 &H, const T9 &I)
  { return(UNUSED_PARAMS_8(A, B, C, D, E, F, G, H) & UNUSED_PARAM(I)); }

  //! @}
  //!
} // namespace vge
#endif
