/** \file BasicEntities.h */
// ############################################################################################
//  BasicEntities.h
//  Definition & implementation of various basic game entity classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _GAME_OBJECT___BASIC_ENTITIES_H_
#define _GAME_OBJECT___BASIC_ENTITIES_H_

// VortexGE headers
#include "../VortexLua/ExecutableByteCodes.h"
#include "../VortexLua/ExecutionSpace.h"
#include "MediaPool.tpp.h"


namespace vge {

  // Foward reference from "lib/2DManipulation/AnimatedImageController.h"
  class CAImgControl;


  //!
  //! \brief The base of all game entity classes
  //!
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBase : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new entity class with the given name.
      CGenttyBase(const char *entityName = NULL);
      //!
      //! Standard dtor
      virtual ~CGenttyBase() throw();

      // ID functions
      //!
      //! Returns the entity's name
      inline const string &GetEntityName() const
      { return(_name); }
      //!
      //! Set the entity's name
      inline void SetEntityName(const char *entityName)
      { _name = entityName ? entityName : ""; }

      // Access to the children
      //!
      //! Returns the entity's children
      inline list<CGenttyBase*> &ChildrenList()
      { return(_children); }
      //!
      //! Returns the entity's children
      inline const list<CGenttyBase*> &ChildrenList() const
      { return(_children); }

    protected:
      string             _name;     // Name of this entity
      list<CGenttyBase*> _children; // List of children
  };


  //!
  //! \brief The base of all game entity classes with script execution support
  //!
  //! All classes derived from this CGenttyBaseXSpace class must also initialize
  //! the virtually derived CGenttyBase class from within the derived class ctor.
  //! \n\n
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBaseXSpace : virtual public CGenttyBase {
    public:
      //!
      //! Construct a new entity class with the given parameters.
      //! \param[in] entityName      Name of the entity class instance, NULL means
      //!                            unnamed.
      //! \param[in] xspaceName      Name of the CLuaExecutionSpace instance in
      //!                            this entity, NULL means the name is the same
      //!                            with the entity name.
      //! \param[in] xspaceGroupName Group name of the CLuaExecutionSpace instance
      //!                            in this entity, NULL menas unnamed.
      //! \param[in] scriptManager   Script manager that will manahe the entity's
      //!                            CLuaExecutionSpace instance.
      CGenttyBaseXSpace(const char  *entityName      = NULL,
                        const char  *xspaceName      = NULL,
                        const char  *xspaceGroupName = NULL,
                        CLuaManager *scriptManager   = NULL);
      //!
      //! Standard dtor
      virtual ~CGenttyBaseXSpace() throw();

      // VortexLua related functions
      //!
      //! Direct access to the CLuaExecutionSpace instance of this entity.
      inline CLuaExecutionSpace &iXSpace()
      { return(_xSpace); }
      //!
      //! Direct access to the CLuaExecutionSpace instance of this entity.
      inline const CLuaExecutionSpace &iXSpace() const
      { return(_xSpace); }

    protected:
      CLuaExecutionSpace _xSpace; // CLuaExecutionSpace instance of this entity
  };


  //!
  //! \brief The base of all game entity classes with script executable support
  //!
  //! All classes derived from this CGenttyBaseEBC class must also initialize
  //! the virtually derived CGenttyBase class from within the derived class ctor.
  //! \n\n
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBaseEBC : virtual public CGenttyBase {
    public:
      //!
      //! Construct a new entity class with the given parameters.
      //! \param[in] entityName      Name of the entity class instance, NULL means
      //!                            unnamed.
      //! \param[in] exeName         Name of the CLuaExecutable instance in
      //!                            this entity, NULL means the name is the same
      //!                            with the entity name.
      //! \param[in] scriptManager   Script manager that will manahe the entity's
      //!                            CLuaExecutableinstance.
      CGenttyBaseEBC(const char  *entityName    = NULL,
                     const char  *exeName       = NULL,
                     CLuaManager *scriptManager = NULL);
      //!
      //! Standard dtor
      virtual ~CGenttyBaseEBC() throw();

      // VortexLua related functions
      //!
      //! Direct access to the CLuaExecutable instance of this entity.
      inline CLuaExecutable &iEBC()
      { return(_exe); }
      //!
      //! Direct access to the CLuaExecutable instance of this entity.
      inline const CLuaExecutable &iEBC() const
      { return(_exe); }
      //!
      //! Execute the script in the context of the given entity.
      virtual void ExecuteScript(CGenttyBaseXSpace &xspaceEntity);

    protected:
      CLuaExecutable _exe; // CLuaExecutionSpace instance of this entity
  };


  //!
  //! \brief The base of all game entity classes with script execution
  //!        and executable support
  //!
  //! All classes derived from this CGenttyBaseXSpaceEBC class must also initialize
  //! the virtually derived CGenttyBase, CGenttyBaseXSpace and CGenttyBaseEBC classes
  //! from within the derived class ctor.
  //! \n\n
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBaseXSpaceEBC : virtual public CGenttyBaseXSpace,
                                          virtual public CGenttyBaseEBC {
    public:
      //!
      //! Construct a new entity class with the given parameters.
      //! \param[in] entityName      Name of the entity class instance, NULL means
      //!                            unnamed.
      //! \param[in] xspaceName      Name of the CLuaExecutionSpace instance in
      //!                            this entity, NULL means the name is the same
      //!                            with the entity name.
      //! \param[in] xspaceGroupName Group name of the CLuaExecutionSpace instance
      //!                            in this entity, NULL menas unnamed.
      //! \param[in] exeName         Name of the CLuaExecutable instance in
      //!                            this entity, NULL means the name is the same
      //!                            with the entity name.
      //! \param[in] scriptManager   Script manager that will manahe the entity's
      //!                            CLuaExecutionSpace instance.
      CGenttyBaseXSpaceEBC(const char  *entityName      = NULL,
                           const char  *xspaceName      = NULL,
                           const char  *xspaceGroupName = NULL,
                           const char  *exeName         = NULL,
                           CLuaManager *scriptManager   = NULL);
      //!
      //! Standard dtor
      virtual ~CGenttyBaseXSpaceEBC() throw();

      // VortexLua related functions
      //!
      //! Execute the script in the context of the given entity.
      virtual void ExecuteScript(CGenttyBaseXSpace &xspaceEntity);
  };


  //!
  //! \brief The base of all game entity classes with image.
  //!
  //! The animated image is meant to be used for something like the
  //! entity's self portrait and not for Sprite2D.
  //! \n\n
  //! All classes derived from this CGenttyBaseImage class must also initialize
  //! the virtually derived CGenttyBase class from within the derived class ctor.
  //! \n\n
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBaseImage : virtual public CGenttyBase {
    public:
      //!
      //! Construct a new entity class with the given parameters.
      //! \param[in] entityName  Name of the entity class instance, NULL means
      //!                        unnamed.
      //! \param[in] image       Pooled CImage instance to be used. This entity
      //!                        class does not copy the image data so, if the
      //!                        original class instance is destroyed before this
      //!                        class, the result will be disastrous.
      CGenttyBaseImage(const char      *entityName = NULL,
                       const CImage_MP &image      = CImage_MP());
      //!
      //! Standard dtor
      virtual ~CGenttyBaseImage() throw();

      // Access to the CImage instance
      //!
      //! Direct access to the CImage instance of this entity.
      inline CImage *iImage()
      { return(_image.MediaInst()); }
      //!
      //! Direct access to the CImage instance of this entity.
      inline const CImage *iImage() const
      { return(_image.MediaInst()); }
      //!
      //! Sets a new CImage instance for this entity.
      inline void SetImage(const CImage_MP &image)
      { _image = image; }

    protected:
      CImage_MP _image; // Pooled CImage instance to be used when drawing this entity
  };


  //!
  //! \brief The base of all game entity classes with animated image.
  //!
  //! The animated image is meant to be used for something like the
  //! entity's self portrait and not for Sprite2D.
  //! \n\n
  //! All classes derived from this CGenttyBaseAImage class must also initialize
  //! the virtually derived CGenttyBase class from within the derived class ctor.
  //! \n\n
  //! See the page \ref Game_Entity_Macros for a more convenient way when
  //! initializing using \ref Game_Entity_IStruct.
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CGenttyBaseAImage : virtual public CGenttyBase {
    public:
      typedef vector<uint32_t>           CInit_AnimSetFrame;
      typedef vector<CInit_AnimSetFrame> CInit_AnimSet;

      static const CInit_AnimSet  _CEAS_;
      static const CInit_AnimSet *CreateEmptyAnimationSet;

    public:
      //!
      //! Construct a new entity class with the given parameters.
      //! \param[in] entityName Name of the entity class instance, NULL means
      //!                       unnamed.
      //! \param[in] aimage     Pooled CAnimatedImage instance to be used. This
      //!                       entity class does not copy the image data so, if
      //!                       the original class instance is destroyed before
      //!                       this class, the result will be disastrous.
      //! \param[in] asInit     Pointer to data set to be used to initialize the
      //!                       entity's CAImgControl class instance. Pass NULL
      //!                       if advanced frame control is not wanted. Pass
      //!                       CGenttyBaseAImage::CreateEmptyAnimationSet to create
      //!                       empty animation set.
      CGenttyBaseAImage(const char              *entityName = NULL,
                        const CAnimatedImage_MP &aimage     = CAnimatedImage_MP(),
                        const CInit_AnimSet     *asInit     = NULL);
      //!
      //! Standard dtor
      virtual ~CGenttyBaseAImage() throw();

      // Access to the CAnimatedImage instance
      //!
      //! Direct access to the CAnimatedImage instance of this entity.
      inline CAnimatedImage *iAImage()
      { return(_aimage.MediaInst()); }
      //!
      //! Direct access to the CAnimatedImage instance of this entity.
      inline const CAnimatedImage *iAImage() const
      { return(_aimage.MediaInst()); }
      //!
      //! Sets a new CAnimatedImage instance for this entity.
      inline void SetAImage(CAnimatedImage_MP &aimage)
      { _aimage = aimage; }

      // Access to the CAImgControl instance
      //!
      //! Direct access to the CAImgControl instance of this entity.
      inline CAImgControl *iAImgControl()
      { return(_aictl); }
      //!
      //! Direct access to the CAImgControl instance of this entity.
      inline const CAImgControl *iAImgControl() const
      { return(_aictl); }

    protected:
      CAnimatedImage_MP  _aimage; // Pooled CAnimatedImage instance to be used when drawing this entity
      CAImgControl      *_aictl;  // Advanced frame control for the animated image
  };

} // namespace vge
#endif
