// ############################################################################################
//  InterProcessCommunication.cc
//  Implementation of IPC classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "InterProcessCommunication.h"


namespace vge {

// MQ helper macros
#define _MQ_MATCH_(RT, SID)   ((RT.tMsgBuf.mtext[0] == 0) && (strcmp(RT.acSID, SID) == 0))
#define _MQ_SND_(QID, RT)     (msgsnd(QID, reinterpret_cast<msgbuf*>(&RT), sizeof(RT) - sizeof(intptr_t), 0) != -1)
#define _MQ_RCV_(QID, RT, TY) (msgrcv(QID, reinterpret_cast<msgbuf*>(&RT), sizeof(RT) - sizeof(intptr_t), TY, IPC_NOWAIT) != -1)


// Special constants for CID
static const uint32_t _CID_START_     = uint32_t(0x00100000);
static const uint32_t _CID_BANDWIDTH_ = uint32_t(0x00000080);

// Connection request
static const intptr_t _VIPC_CREQ_CONNECT_ = intptr_t(1);
static const char    *_VIPC_CSID_CONNECT_ = "VIPC CREQ CONNECT";
static const intptr_t _VIPC_SRSP_CONNECT_ = intptr_t(2);
static const char    *_VIPC_SSID_CONNECT_ = "VIPC SRSP CONNECT";

// Disconnection request
static const intptr_t _VIPC_CREQ_DISCONN_ = intptr_t(3);
static const char    *_VIPC_CSID_DISCONN_ = "VIPC CREQ DISCONN";

// Shut down notification
static const intptr_t _VIPC_SNTC_SHUTDWN_ = intptr_t(4);
static const char    *_VIPC_SSID_SHUTDWN_ = "VIPC SNTC SHUTDWN";

// Enumeration request
static const intptr_t _VIPC_CREQ_ENMBYID_ = intptr_t(5);
static const char    *_VIPC_CSID_ENMBYID_ = "VIPC CREQ ENMBYID";
static const intptr_t _VIPC_SRSP_ENMBYID_ = intptr_t(6);
static const char    *_VIPC_SSID_ENMBYID_ = "VIPC SRSP ENMBYID";

// Custom client to client transaction
static const char *_VIPC_CSID_CC2CTRN_ = "VIPC CSID CC2CTRN";


//
// The message queue server (manager) class
//
CMQServer::CMQServer(const char *pcConnName)
: _tKey(-1), _n32QID(-1)
{
  memset(_atCList, 0, sizeof(_TClientData) * _MAX_MQ_CLIENT_);
  if(pcConnName) Initialize(pcConnName);
}

CMQServer::~CMQServer() throw()
{
}

void CMQServer::Initialize(const char *pcConnName)
{
  // Create a temporary file to be used by ftok()
  snprintf(_acFileName, sizeof(_acFileName), "/tmp/%s", pcConnName);
  FILE *ptFile = fopen(_acFileName, "wb");
  if(ptFile == NULL)
    VError(VERR_OpenFile, _NFL_, "Couldn't create file \"%s\" to be used by ftok()\nMake sure that you have write permission to \"tmp\"", _acFileName);
  fclose(ptFile);
  VDebugPrint("Temporary file \"%s\" has been created\n", _acFileName);

  // Create a unique key
  _tKey = ftok(_acFileName, pcConnName[0]);
  if(_tKey == -1)
    VError(VERR_OpenFile, _NFL_, "Couldn't create a unique key for creating message queue");

  // Check if the same server already running
  _n32QID = msgget(_tKey, 0600);
  if(_n32QID != -1) {
    _n32QID = - 1;
    VError(VERR_MsgGet, _NFL_, "Another MQ server using the same 'pcConnName' already running");
  }

  // Create a new message queue
  _n32QID = msgget(_tKey, IPC_CREAT | 0600);
  if(_n32QID == -1)
    VError(VERR_MsgGet, _NFL_, "Couldn't create a new message queue");
  VDebugPrint("Server initialized\n");
}

void CMQServer::ShutDown(bool bForce)
{
  // Check for still connected client(s)
  if(bForce) {
    // Prepare notification structure
    _TSNtcSDwnData tSNSD;
    tSNSD.tMsgBuf.mtype    = _VIPC_SNTC_SHUTDWN_;
    tSNSD.tMsgBuf.mtext[0] = 0;
    strcpy(tSNSD.acSID, _VIPC_SSID_SHUTDWN_);
    // Notify client for shutdown
    bool bFound = false;
    for(size_t i = 0; i < _MAX_MQ_CLIENT_; ++i)
    if(_atCList[i].un32ClientID != 0) {
      bFound = true;
      tSNSD.un32ClientID = _atCList[i].un32ClientID;
      _MQ_SND_(_n32QID, tSNSD);
      VDebugPrint("Sending shutdown notification to client (%s, %08X)\n", _atCList[i].acUID, _atCList[i].un32ClientID);
      usleep(250000);
    }
    if(bFound) sleep(1);
  }
  else {
    size_t tot = 0;
    for(size_t i = 0; i < _MAX_MQ_CLIENT_; ++i)
      if(_atCList[i].un32ClientID != 0) ++tot;
    if(tot)
      VError(VERR_InUse, _NFL_, "Can't shutdown server now; %zd client(s) still connected", tot);
  }

  // Perform shutdown
  if(_n32QID != -1) {
    // Remove the message queue
    msqid_ds tDS;
    if(msgctl(_n32QID, IPC_RMID, &tDS) == -1)
      VError(VERR_MsgCtl, _NFL_, "Couldn't remove the message queue");
    _tKey   = -1;
    _n32QID = -1;
    memset(_atCList, 0, sizeof(_TClientData) * _MAX_MQ_CLIENT_);
    VDebugPrint("Server terminated\n");

    // Delete the temporary file
    unlink(_acFileName);
    VDebugPrint("Temporary file \"%s\" has been deleted\n", _acFileName);
  }
}

void CMQServer::Pool()
{
  // Check for connection request from client
  _TCReqConnData tCRCD;
  if(_MQ_RCV_(_n32QID, tCRCD, _VIPC_CREQ_CONNECT_)) {
    _TSRspConnData tSRCD;
    if(_MQ_MATCH_(tCRCD, _VIPC_CSID_CONNECT_)) {
      // Prepare response structure
      tSRCD.tMsgBuf.mtype    = _VIPC_SRSP_CONNECT_;
      tSRCD.tMsgBuf.mtext[0] = 0;
      tSRCD.un32RandomID     = tCRCD.un32RandomID;
      strcpy(tSRCD.acUID, tCRCD.acUID);
      strcpy(tSRCD.acSID, _VIPC_SSID_CONNECT_);
      // Search for an empty list
      size_t i;
      for(i = 0; i < _MAX_MQ_CLIENT_; ++i) {
        if(_atCList[i].un32ClientID == 0) {
          // Found, accept connection
          tSRCD.un32ClientID       = (i * _CID_BANDWIDTH_) + _CID_START_;
          _atCList[i].un32ClientID = tSRCD.un32ClientID;
          strcpy(_atCList[i].acUID, tCRCD.acUID);
          VDebugPrint("Connection request from new client (%s, %08X) accepted\n", tCRCD.acUID, tCRCD.un32RandomID);
          break;
        }
      }
      // No empty list, reject connection
      if(i >= _MAX_MQ_CLIENT_) {
        tSRCD.un32ClientID = 0;
        VDebugPrint("Connection request from new client (%s, %08X) rejected\n", tCRCD.acUID, tCRCD.un32RandomID);
      }
      // Send response to client
      if(!_MQ_SND_(_n32QID, tSRCD)) goto ERRRsp;
    }
  }

  // Check for disconnection request from client
  _TCReqDscnData tCRDD;
  if(_MQ_RCV_(_n32QID, tCRDD, _VIPC_CREQ_DISCONN_)) {
    if(_MQ_MATCH_(tCRDD, _VIPC_CSID_DISCONN_)) {
      // Search for the client
      size_t i;
      for(i = 0; i < _MAX_MQ_CLIENT_; ++i) {
        if((tCRDD.un32ClientID == _atCList[i].un32ClientID) &&
        (strcmp(tCRDD.acUID, _atCList[i].acUID) == 0)) {
          // Found, delete the client from list
          _atCList[i].un32ClientID = 0;
          _atCList[i].acUID[0]     = 0;
          VDebugPrint("Disconnection request from client (%s, %08X) approved\n", tCRDD.acUID, tCRDD.un32ClientID);
          break;
        }
      }
      // Not found
      if(i >= _MAX_MQ_CLIENT_)
        VDebugPrint("Diconnection request from unregistered client (%s, %08X) ignored\n", tCRDD.acUID, tCRDD.un32ClientID);
    }
  }

  // Check for enumeration request from client
  _TCReqEByUData tCRED;
  if(_MQ_RCV_(_n32QID, tCRED, _VIPC_CREQ_ENMBYID_)) {
    _TSRspEByUData tSRED;
    if(_MQ_MATCH_(tCRED, _VIPC_CSID_ENMBYID_)) {
      // Search for the client
      size_t i;
      for(i = 0; i < _MAX_MQ_CLIENT_; ++i) {
        if(tCRED.un32ClientID == _atCList[i].un32ClientID) {
          // Prepare response structure
          tSRED.tMsgBuf.mtype    = _VIPC_SRSP_ENMBYID_;
          tSRED.tMsgBuf.mtext[0] = 0;
          tSRED.un32ClientID     = tCRED.un32ClientID;
          strcpy(tSRED.acSID, _VIPC_SSID_ENMBYID_);
          // Enumerate client
          tSRED.un32EnumCount = _EnumerateClientByUID_(tSRED.aun32ClientID, _ENMBYID_ALL_, tCRED.un32ClientID);
          // Send message
          if(!_MQ_SND_(_n32QID, tSRED)) goto ERRRsp;
          VDebugPrint("Enumeration request from client (%s, %08X) accepted\n", tCRED.acUID, tCRED.un32ClientID);
          break;
        }
      }
      // Not found
      if(i >= _MAX_MQ_CLIENT_)
        VDebugPrint("Enumeration request from unregistered client (%s, %08X) ignored\n", tCRED.acUID, tCRED.un32ClientID);
    }
  }

  // OK !
  return;

  // Error messages
ERRRsp:
  VError(VERR_MsgSnd, _NFL_, "Couldn't send response to client");
}

uint32_t CMQServer::EnumerateClientByUID(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate, uint32_t un32CIDExcept) const
{
  // Check
  if(pcUID2Enumerate) {
    if(strlen(pcUID2Enumerate) >= _MAX_MQ_STRBUF_)
      VError(VERR_InvalidParameter, _NFL_, "The user ID to be enumerated must not longer than %d characters", _MAX_MQ_STRBUF_ - 1);
  }
  else
    pcUID2Enumerate = "[NULL]";

  // Do enumeration
  return(_EnumerateClientByUID_(aun32ClientID, pcUID2Enumerate, un32CIDExcept));
}

uint32_t CMQServer::_EnumerateClientByUID_(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate, uint32_t un32CIDExcept) const
{
  // Search for clients with specified UID and store their CID in list
  memset(aun32ClientID, 0, sizeof(aun32ClientID[0]) * _MAX_MQ_CLIENT_);
  size_t tot = 0;
  if(strcmp(pcUID2Enumerate, _ENMBYID_ALL_) == 0) {
    for(size_t i = 0; i < _MAX_MQ_CLIENT_; ++i) {
      if(_atCList[i].un32ClientID)
        if(_atCList[i].un32ClientID != un32CIDExcept)
          aun32ClientID[tot++] = _atCList[i].un32ClientID;
    }
  }
  else {
    for(size_t i = 0; i < _MAX_MQ_CLIENT_; ++i) {
      if(_atCList[i].un32ClientID)
        if(_atCList[i].un32ClientID != un32CIDExcept)
          if(strcmp(pcUID2Enumerate, _atCList[i].acUID) == 0)
            aun32ClientID[tot++] = _atCList[i].un32ClientID;
    }
  }
  return(tot);
}


//
// The message queue client class
//
CMQClient::CMQClient()
: _tKey(-1), _n32QID(-1), _un32CID(0)
{
  memset(_acUserID, 0, sizeof(char) * _MAX_MQ_STRBUF_);
}

CMQClient::~CMQClient() throw()
{
  try {
    Disconnect();
  }
  catch(...) {}
}

void CMQClient::Connect(const char *pcConnName, const char *pcUserID, uint8_t un8TimeOutInSeconds)
{
  // Temporary variables
  _TCReqConnData tCRCD;
  _TSRspConnData tSRCD;
  char           acFile[PATH_MAX];
  time_t         tST;

  // Check
  if(pcUserID) {
    if(strlen(pcUserID) >= _MAX_MQ_STRBUF_)
      VError(VERR_InvalidParameter, _NFL_, "The given user ID must not longer than %d characters", _MAX_MQ_STRBUF_ - 1);
  }
  else
    pcUserID = "[NULL]";
  strcpy(_acUserID, pcUserID);

  // Create a unique key
  snprintf(acFile, sizeof(acFile), "/tmp/%s", pcConnName);
  _tKey = ftok(acFile, pcConnName[0]);
  if(_tKey == -1)
    VError(VERR_OpenFile, _NFL_, "Couldn't get a unique key for accesing message queue\nMake sure the correct MQ server is already initialized (running)");

  // Connect to an existing message queue
  _n32QID = msgget(_tKey, 0600);
  if(_n32QID == -1)
    VError(VERR_MsgGet, _NFL_, "Couldn't connect to the server message queue\nMake sure the correct MQ server is already initialized (running)");

  // Generate a new random number
  uint32_t rnum;
  int      fd;
  ssize_t  rz;
  if((fd = open("/dev/urandom", O_RDONLY)) < 0)  VError(VERR_OpenDevice, _NFL_, "Couldn't open device \"/dev/urandom\"");
  rz = read(fd, &rnum, sizeof(rnum));
  close(fd);
  if(rz != sizeof(rnum)) VError(VERR_ReadDevice, _NFL_, "Couldn't read from \"/dev/urandom\"");

  // Send a connection request
  tCRCD.tMsgBuf.mtype    = _VIPC_CREQ_CONNECT_;
  tCRCD.tMsgBuf.mtext[0] = 0;
  tCRCD.un32RandomID     = rnum;
  strcpy(tCRCD.acUID, pcUserID);
  strcpy(tCRCD.acSID, _VIPC_CSID_CONNECT_);
  if(!_MQ_SND_(_n32QID, tCRCD))
    VError(VERR_MsgSnd, _NFL_, "Couldn't send connection request to server");
  VDebugPrint("Requesting connection (%s, %08X)\n", tCRCD.acUID, tCRCD.un32RandomID);

  // Waiting for server to respond
  tST = time(NULL);
  for(;;) {
    // Check for response
    if(_MQ_RCV_(_n32QID, tSRCD, _VIPC_SRSP_CONNECT_)) {
      if(_MQ_MATCH_(tSRCD, _VIPC_SSID_CONNECT_)) {
        // Not for me, send it back
        if((strcmp(tSRCD.acUID, tCRCD.acUID) != 0) || (tSRCD.un32RandomID != tCRCD.un32RandomID))
          _MQ_SND_(_n32QID, tSRCD);
        // For me
        else {
          if(tSRCD.un32ClientID == 0)
            VError(VERR_TooMany, _NFL_, "Connection is refused by the server\nThe server may already has too many clients)");
          _un32CID = tSRCD.un32ClientID;
          VDebugPrint("Connection request has been accepted by the server\nEffective client ID is set to %08X\n", tSRCD.un32ClientID);
          break;
        }
      }
    }
    // Check for timeout
    if((time(NULL) - tST) >= time_t(un8TimeOutInSeconds))
      VError(VERR_Timeout, _NFL_, "Request timeout");
  }
}

void CMQClient::Disconnect()
{
  // Temporary variables
  _TCReqDscnData tCRDD;

  // Check
  if((_un32CID == 0) || (_acUserID[0] == 0)) return;

  // Send a disconnection request
  tCRDD.tMsgBuf.mtype    = _VIPC_CREQ_DISCONN_;
  tCRDD.tMsgBuf.mtext[0] = 0;
  tCRDD.un32ClientID     = _un32CID;
  strcpy(tCRDD.acUID, _acUserID);
  strcpy(tCRDD.acSID, _VIPC_CSID_DISCONN_);
  if(!_MQ_SND_(_n32QID, tCRDD))
    VError(VERR_MsgSnd, _NFL_, "Couldn't send disconnection request to the server");
  VDebugPrint("Requesting disconnection (%s, %08X)\n", tCRDD.acUID, tCRDD.un32ClientID);

  // Set variables
  _tKey        = -1;
  _n32QID      = -1;
  _un32CID     = 0;
  _acUserID[0] = 0;
}

bool CMQClient::GotShutDownNotification()
{
  // Temporary variables
  _TSNtcSDwnData tSNSD;

  // Check for shutdown notification from server
  if(_MQ_RCV_(_n32QID, tSNSD, _VIPC_SNTC_SHUTDWN_)) {
    if(_MQ_MATCH_(tSNSD, _VIPC_SSID_SHUTDWN_)) {
      // Not for me, send it back
      if(tSNSD.un32ClientID != _un32CID)
        _MQ_SND_(_n32QID, tSNSD);
      // For me
      else {
        VDebugPrint("Received shutdown notification from server\n");
        _tKey        = -1;
        _n32QID      = -1;
        _un32CID     = 0;
        _acUserID[0] = 0;
        return(true);
      }
    }
  }

  // Hasn't got any shutdown notification
  return(false);
}

uint32_t CMQClient::EnumerateByUID(uint32_t aun32ClientID[_MAX_MQ_CLIENT_], const char *pcUID2Enumerate, uint8_t un8TimeOutInSeconds) const
{
  // Temporary variables
  _TCReqEByUData tCRED;
  _TSRspEByUData tSRED;
  time_t               tST;

  // Check
  if(pcUID2Enumerate) {
    if(strlen(pcUID2Enumerate) >= _MAX_MQ_STRBUF_)
      VError(VERR_InvalidParameter, _NFL_, "The user ID to be enumerated must not longer than %d characters", _MAX_MQ_STRBUF_ - 1);
  }
  else
    pcUID2Enumerate = "[NULL]";

  // Send an enumeration request
  tCRED.tMsgBuf.mtype    = _VIPC_CREQ_ENMBYID_;
  tCRED.tMsgBuf.mtext[0] = 0;
  tCRED.un32ClientID     = _un32CID;
  strcpy(tCRED.acUID, _acUserID);
  strcpy(tCRED.acUID2Enum, pcUID2Enumerate);
  strcpy(tCRED.acSID, _VIPC_CSID_ENMBYID_);
  if(!_MQ_SND_(_n32QID, tCRED)) {
    VError(VERR_MsgSnd, _NFL_, "Couldn't send enumeration request to the server");
    return(0);
  }
  VDebugPrint("Requesting enumeration by UID (%s, %08X)\n", _acUserID, _un32CID);

  // Waiting for server to respond
  tST = time(NULL);
  for(;;) {
    // Check for response
    if(_MQ_RCV_(_n32QID, tSRED, _VIPC_SRSP_ENMBYID_)) {
      if(_MQ_MATCH_(tSRED, _VIPC_SSID_ENMBYID_)) {
        // Not for me, send it back
        if(tSRED.un32ClientID != _un32CID)
          _MQ_SND_(_n32QID, tSRED);
        // For me
        else {
          memcpy(aun32ClientID, tSRED.aun32ClientID, sizeof(aun32ClientID[0]) * _MAX_MQ_CLIENT_);
          VDebugPrint("Enumeration request approved\n");
          break;
        }
      }
    }
    // Check for timeout
    if((time(NULL) - tST) >= time_t(un8TimeOutInSeconds)) {
      VError(VERR_Timeout, _NFL_, "Request timeout");
      return(0);
    }
  }

  // OK !
  return(tSRED.un32EnumCount);
}

void CMQClient::CC2CSend(uint32_t un32TargetClientID, uint8_t un8CustomIdentifier, const void *pvCustomData, uint32_t un32CustomDataSize) const
{
  // Temporary variables
  _TCTrnCC2CData *ptCTCD;

  // Allocate memory
  ptCTCD = reinterpret_cast<_TCTrnCC2CData*>(malloc(sizeof(_TCTrnCC2CData) + un32CustomDataSize));
  if(ptCTCD == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for performing custom client to client transaction");

  // Prepare message structure
  ptCTCD->tMsgBuf.mtype      = un32TargetClientID + un8CustomIdentifier;
  ptCTCD->tMsgBuf.mtext[0]   = 0;
  ptCTCD->un32ClientID       = _un32CID;
  ptCTCD->un32CustomDataSize = un32CustomDataSize;
  strcpy(ptCTCD->acSID, _VIPC_CSID_CC2CTRN_);
  memcpy((reinterpret_cast<char*>(ptCTCD)) + sizeof(_TCTrnCC2CData), pvCustomData, un32CustomDataSize);

  // Send message
  if(msgsnd(_n32QID, reinterpret_cast<msgbuf*>(ptCTCD),
  sizeof(_TCTrnCC2CData) + un32CustomDataSize - sizeof(intptr_t), 0) == -1) {
    free(ptCTCD);
    VError(VERR_MsgSnd, _NFL_, "Couldn't send custom transaction");
  }
  VDebugPrint("Custom transaction (%08X.%02X, %d) has been sent\n", un32TargetClientID, un8CustomIdentifier, un32CustomDataSize);

  // OK !
  free(ptCTCD);
}

bool CMQClient::CC2CReceive(uint32_t *pun32SourceClientID, uint8_t un8CustomIdentifier, void *pvCustomData, uint32_t un32CustomDataSize) const
{
  // Temporary variables
  _TCTrnCC2CData *ptCTCD;

  // Allocate memory
  ptCTCD = reinterpret_cast<_TCTrnCC2CData*>(malloc(sizeof(_TCTrnCC2CData) + un32CustomDataSize));
  if(ptCTCD == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory for performing custom client to client transaction");

  // Get message
  if(msgrcv(_n32QID, reinterpret_cast<msgbuf*>(ptCTCD),
            sizeof(_TCTrnCC2CData) + un32CustomDataSize - sizeof(intptr_t),
            _un32CID + un8CustomIdentifier, MSG_NOERROR | IPC_NOWAIT) != -1) {
    if(_MQ_MATCH_((*ptCTCD), _VIPC_CSID_CC2CTRN_)) {
      VDebugPrint("Got custom transaction (%08X.%02X, %d)\n", ptCTCD->un32ClientID, un8CustomIdentifier, ptCTCD->un32CustomDataSize);
      if(ptCTCD->un32CustomDataSize != un32CustomDataSize)
        VDebugPrint("Custom client to client transaction from CID %08X rejected\n", ptCTCD->un32ClientID);
      else {
        *pun32SourceClientID = ptCTCD->un32ClientID;
        memcpy(pvCustomData, (reinterpret_cast<char*>(ptCTCD)) + sizeof(_TCTrnCC2CData), un32CustomDataSize);
        VDebugPrint("Custom client to client transaction from CID %08X accepted\n", ptCTCD->un32ClientID);
        free(ptCTCD);
        return(true);
      }
    }
    else
      VDebugPrint("Got invalid (unknown) custom client to client transaction\n");
  }

  // Exiting !
  free(ptCTCD);
  return(false);
}



//
// Exported functions implementation
//
bool VEnableGetChar(bool bState)
{
  // Static variables
  static termios tSaved;
  static bool  bSaved = false;

  // Temporary variables
  termios tA;

  // Save the original options
  if(!bSaved) {
    if(tcgetattr(STDIN_FILENO, &tSaved) != 0) return(false);
    bSaved = true;
    if(!bState) return(true);
  }

  // Restore the original options
  if(!bState) return(tcsetattr(STDIN_FILENO, TCSANOW, &tSaved) == 0);

  // Get current options
  if(tcgetattr(STDIN_FILENO, &tA) != 0) return(false);

  // Change options
  tA.c_lflag &= (~ECHO);
  tA.c_lflag &= (~ICANON);
  tA.c_iflag &= (~IGNBRK);
  tA.c_iflag &= (~BRKINT);
  tA.c_iflag &= (~PARMRK);
  tA.c_cc[VINTR] = 0;
  tA.c_cc[VMIN]  = 0;

  // Set new options
  return(tcsetattr(STDIN_FILENO, TCSANOW, &tA) == 0);
}

char VGetChar()
{
  static char acBuff[2];
  acBuff[0] = acBuff[1] = 0;
  if(read(STDIN_FILENO, &acBuff, 1) < 0) return(0);
  return(acBuff[0]);
}

} // namespace vge
