// ############################################################################################
//  StartupFunctions.cc
//  Implementation of VortexGE startup functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "FatalSignalHandler.h"

// GCC and Linux specific headers
#include <cxxabi.h>
#include <exception_defines.h>


namespace vge {

// From X11 video implementation file,
// "lib/VideoEnvironment/VID_X11/VX11Environment.h"
#ifdef __USE_X11_VIDEO__
extern void _XCrash_RestoreState_();
#endif

namespace xprv {

  // Print the exception information
  void VGE_EXPORT _PrintExceptionInfo_(const char *msg) _attr_noreturn;
  void VGE_EXPORT _PrintExceptionInfo_(const char *msg)
  {
    type_info  *t = abi::__cxa_current_exception_type();

    try {
      if(t) {
        const char *m = t->name();
        char        d[8192];
        size_t      z = sizeof(d);
        int32_t     s = -1;

        abi::__cxa_demangle(m, d, &z, &s);

        if(s == 0)
          VFatalError(VERR_Unknown, _NFL_, "%s called after throwing an instance of %s", msg, d);
        else
          VFatalError(VERR_Unknown, _NFL_, "%s called after throwing an instance of %s", msg, m);
      }
      else {
          VFatalError(VERR_Unknown, _NFL_, "%s called without an active exception", msg);
      }
    }
    catch(...) {}

    // Try to provide more information
    if(t) {
      try {
        __throw_exception_again;
      }
      catch (CException &e) {
        VPrint("  What()     = %s\n", e.What());
        VPrint("  Function() = %s\n", e.Location().Function());
        VPrint("  File()     = %s\n", e.Location().File());
        VPrint("  Line()     = %d\n", e.Location().Line());
        VPrint("  Code()     = %d\n", e.Code());
      }
      catch (exception &e) {
        VPrint("  What() = %s\n", e.what());
      }
      catch (...) { }
    }

    // Dump the stack and exit
    VDumpStack();
    exit(-SIGTERM);
  }

  // Unexpected handler so that unexpected throws won't cause
  // direct termination without any information
  void VGE_EXPORT _UnexpectedHandler_() _attr_noreturn;
  void VGE_EXPORT _UnexpectedHandler_()
  {
  // Try to restore the system state
#ifdef __USE_X11_VIDEO__
  _XCrash_RestoreState_();
#endif
    _PrintExceptionInfo_("_UnexpectedHandler_()");
  }

  // Terminate handler so that unhandled throws won't cause
  // direct termination without any information
  void VGE_EXPORT _TerminateHandler_() _attr_noreturn;
  void VGE_EXPORT _TerminateHandler_()
  {
  // Try to restore the system state
#ifdef __USE_X11_VIDEO__
  _XCrash_RestoreState_();
#endif
    _PrintExceptionInfo_("_TerminateHandler_()");
  }


  // Start up function
  void VGE_EXPORT _vgeStartupOTIF_() __attribute__((constructor));
  void VGE_EXPORT _vgeStartupOTIF_()
  {
    // Guard
    static bool bExecuted = false;
    if(bExecuted) return;
    bExecuted = true;

    // Install exception handlers
    set_terminate(_TerminateHandler_);
    set_unexpected(_UnexpectedHandler_);

    // Fatal signals handler
    VInstallFSH();

    // Enable VDebugPrint() ?
    const char *env = getenv("VGE_ENABLE_VDEBUGPRINT");
         if(VMatchEVar(env, "yes") == EMEV_Match) VEnableDebugPrint(true);
    else if(VMatchEVar(env,  "no") == EMEV_Match) VEnableDebugPrint(false);
    else if(env) VWarning(_NFL_, "Invalid value in environment variable \"VGE_VIDEO_CUSTOM_8BPP_PALETTE\"\nThe default system palette will be used");

    // Use "en_US.UTF-8" as the C locale
    setlocale(LC_ALL, "en_US.UTF-8");
  }

} // namespace xprv

} // namespace vge
