/** \file SystemUtilities.h */
// ############################################################################################
//  SystemUtilities.h
//  Definition of various system utilities
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _SYSTEM___SYSTEM_UTILITIES_H_
#define _SYSTEM___SYSTEM_UTILITIES_H_

// VortexGE headers
#include "../Base/VBasics.h"


namespace vge {
  //!
  //! \addtogroup SystemUtilities
  //! @{

  // CPU capabilities bit flags
  const uint32_t VCCAPS_FPU     = (uint32_t(1) <<  0);
  const uint32_t VCCAPS_CX8     = (uint32_t(1) <<  1);
  const uint32_t VCCAPS_CMOV    = (uint32_t(1) <<  2);
  const uint32_t VCCAPS_FXSR    = (uint32_t(1) <<  3);
  const uint32_t VCCAPS_SEP     = (uint32_t(1) <<  4);
  const uint32_t VCCAPS_CLFLUSH = (uint32_t(1) <<  5);
  const uint32_t VCCAPS_TSC     = (uint32_t(1) <<  6);
  const uint32_t VCCAPS_MSR     = (uint32_t(1) << 13);
  const uint32_t VCCAPS_MTRR    = (uint32_t(1) << 14);
  const uint32_t VCCAPS_PAT     = (uint32_t(1) << 15);
  const uint32_t VCCAPS_MMX     = (uint32_t(1) << 17);
  const uint32_t VCCAPS_EMMX    = (uint32_t(1) << 18);
  const uint32_t VCCAPS_SSE     = (uint32_t(1) << 19);
  const uint32_t VCCAPS_SSE2    = (uint32_t(1) << 20);
  const uint32_t VCCAPS_SSE3    = (uint32_t(1) << 21);
  const uint32_t VCCAPS_3DNOW   = (uint32_t(1) << 22);
  const uint32_t VCCAPS_E3DNOW  = (uint32_t(1) << 23);
  const uint32_t VCCAPS_EST     = (uint32_t(1) << 29);
  const uint32_t VCCAPS_HTT     = (uint32_t(1) << 30);
  const uint32_t VCCAPS_MONITOR = (uint32_t(1) << 31);

  // CPU information utilities
  //!
  //! Returns a uint32_t containing the CPU capabilities flag.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT uint32_t VGetCPUCaps();
  //!
  //! Print the CPU capabilities.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VPrintCPUCaps(bool bVerbose = false);

  // Check if the CPU has a particular ability
  inline bool CPU_HAS_FPU     (uint32_t X) { return((X & VCCAPS_FPU)     != 0); }
  inline bool CPU_HAS_CX8     (uint32_t X) { return((X & VCCAPS_CX8)     != 0); }
  inline bool CPU_HAS_CMOV    (uint32_t X) { return((X & VCCAPS_CMOV)    != 0); }
  inline bool CPU_HAS_FCMOV   (uint32_t X) { return(CPU_HAS_FPU(X) && CPU_HAS_CMOV(X)); }
  inline bool CPU_HAS_FXSR    (uint32_t X) { return((X & VCCAPS_FXSR)    != 0); }
  inline bool CPU_HAS_SEP     (uint32_t X) { return((X & VCCAPS_SEP)     != 0); }
  inline bool CPU_HAS_CLFLUSH (uint32_t X) { return((X & VCCAPS_CLFLUSH) != 0); }
  inline bool CPU_HAS_TSC     (uint32_t X) { return((X & VCCAPS_TSC)     != 0); }
  inline bool CPU_HAS_MSR     (uint32_t X) { return((X & VCCAPS_MSR)     != 0); }
  inline bool CPU_HAS_MTRR    (uint32_t X) { return((X & VCCAPS_MTRR)    != 0); }
  inline bool CPU_HAS_PAT     (uint32_t X) { return((X & VCCAPS_PAT)     != 0); }
  inline bool CPU_HAS_MMX     (uint32_t X) { return((X & VCCAPS_MMX)     != 0); }
  inline bool CPU_HAS_EMMX    (uint32_t X) { return((X & VCCAPS_EMMX)    != 0); }
  inline bool CPU_HAS_SSE     (uint32_t X) { return((X & VCCAPS_SSE)     != 0); }
  inline bool CPU_HAS_SSE2    (uint32_t X) { return((X & VCCAPS_SSE2)    != 0); }
  inline bool CPU_HAS_SSE3    (uint32_t X) { return((X & VCCAPS_SSE3)    != 0); }
  inline bool CPU_HAS_3DNOW   (uint32_t X) { return((X & VCCAPS_3DNOW)   != 0); }
  inline bool CPU_HAS_E3DNOW  (uint32_t X) { return((X & VCCAPS_E3DNOW)  != 0); }
  inline bool CPU_HAS_3DNOWPRO(uint32_t X) { return(CPU_HAS_3DNOW(X) && CPU_HAS_SSE(X)); }
  inline bool CPU_HAS_EST     (uint32_t X) { return((X & VCCAPS_EST)     != 0); }
  inline bool CPU_HAS_HTT     (uint32_t X) { return((X & VCCAPS_HTT)     != 0); }
  inline bool CPU_HAS_MONITOR (uint32_t X) { return((X & VCCAPS_MONITOR) != 0); }


  //!
  //! \brief File information structure
  //!
  //! \api pretty stable.
  struct TFileInfo {
    size_t Size;
    struct {
      u_char IsRegularFile  : 1;
      u_char IsDirectory    : 1;
      u_char IsCharDevice   : 1;
      u_char IsBlockDevice  : 1;
      u_char IsFIFO         : 1;
      u_char IsSocket       : 1;
      u_char IsPipe         : 1;
      u_char IsSymLink      : 1;
      u_char IsUnknown      : 1;
    } Type;
  };

  // File & directory related utilities
  //!
  //! Check if the given file name has the given extention.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT bool VFileHasExt(const string &fileName, const string &fileExt);
  //!
  //! Correct the file name's extention.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VCorrectFileExt(string &dstFileName, const string &srcFileName, const string &fileExt);
  //!
  //! Check if the given file is readable or not.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT bool VFileIsReadable(const string &fileName);
  //!
  //! Get the some information about the given file.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VGetFileInfo(TFileInfo &fi, const string &fileName);
  //!
  //! Delete the given file.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VRemoveFile(const string &fileName, bool ignoreIfNotFound = false);
  //!
  //! Rename the given file.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VRenameFile(const string &oldFileName, const string &newFileName);
  //!
  //! Create an empty file with the given name and permission (mode).
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VCreateEmptyFile(const string &fileName, mode_t mode = 0666);
  //!
  //! Returns the current executable name with absolute path.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT const char *VGetAbsoluteExecutableName();
  //!
  //! Returns the current executable's absolute path.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT const char *VGetAbsoluteExecutablePath();
  //!
  //! Make a temporary file descriptor.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT int VMakeTmpFD(const char *namePrefix);
  //!
  //! Make a FIFO (pipe).
  //!
  //! \api pretty stable.
  extern VGE_EXPORT bool VMakeFIFO(const char *pcPipeName);
  //!
  //! Get current working directory.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT const char *VGetCWD();
  //!
  //! Set current working directory.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VSetCWD(const string &path);
  //!
  //! Create a directory with the given name and permission (mode).
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VCreateDir(const string &path, mode_t mode = 0777);
  //!
  //! Delete the given directory.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VRemoveDir(const string &path, bool ignoreIfNotFound = false);
  //!
  //! List the contents of the given directory.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VListDirectory(vector<string> &list, const string &path);


  //!
  //! \brief Enumeration used by the VMatchEVar() function.
  //!
  //! \api pretty stable.
  enum EMatchEVar { EMEV_Empty, EMEV_Match, EMEV_Mismatch };

  // Environment variable related utilities
  //!
  //! Compare the contents of the given environment variable name with
  //! the given reference string.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT EMatchEVar VMatchEVar(const char *envVar, const char *refVal);


  // Other utilities
  //!
  //! The number of the system's ticks per second.
  const uint64_t TICKS_PER_SEC = uint64_t(1000000);

  //!
  //! Returns the current value of the system's ticks (the ellapsed time 
  //! in microseconds).
  inline uint64_t VGetSystemTicks()
  {
    struct timeval tv;
    gettimeofday(&tv, 0);
    return(uint64_t(tv.tv_sec) * TICKS_PER_SEC + tv.tv_usec);
  }

  //!
  //! Tokenize the given source string.
  //!
  //! \api pretty stable.
  extern VGE_EXPORT void VTokenize(vector<string> &tokens, const string &source, const string &delimiters = " ");

  //! @}
  //!
} // namespace vge
#endif
