/** \file CommonDialogPages.h */
// ############################################################################################
//  CommonDialogPages.h
//  Definition & implementation of most commonly used dialog page classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___COMMON_DIALOG_PAGES_H_
#define _VIDEO_ENVIRONMENT___COMMON_DIALOG_PAGES_H_

// VortexGE headers
#include "DialogPage.h"


namespace vge {

  //!
  //! \brief Message box
  //!
  //! \api experiment
  class CMessageBox : public CDialogPage {
    public:
      //!
      //! Construct an input box instance using the given parameters.
      CMessageBox(CVideoEnv          &iVideoEnv,
                  CEventHandlerBase  *eventHandler,
                  const string       &caption,
                  const string       &text,
                  EMsgBoxIcon         eIcon   = EMBI_Information,
                  EMsgBoxButtons      eButton = EMBB_OK,
                  EMsgBoxSize         eSize   = EMBS_Small);
      //!
      //! Standard dtor.
      virtual ~CMessageBox() throw();

      // Dialog functions
      //!
      //! Check if the dialog already completed (exit).
      inline bool Exit() const
      { return(_exit); }
      //!
      //! Get the result of the message box.
      inline const EMsgBoxRetVal GetResult() const
      { return(_resultCode); }

      // Derived from CDialogPage
      virtual void InitializeAllControls();
      virtual void HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str);

    protected:
      // Controls' event handlers
      void _onBtnOK(_CControl *ctrl);
      void _onBtnCancel(_CControl *ctrl);
      void _onBtnYes(_CControl *ctrl);
      void _onBtnNo(_CControl *ctrl);
      void _onBtnAbort(_CControl *ctrl);
      void _onBtnRetry(_CControl *ctrl);
      void _onBtnIgnore(_CControl *ctrl);

      // Derived from CDialogPage
      virtual void CustomDraw_OnRedrawAllControls ();
      virtual void CustomDraw_OnButtonPress(EBtnSymbol buttonSymbol);
      virtual void CustomDraw_OnButtonRelease(EBtnSymbol buttonSymbol);
      virtual void CustomDraw_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos);

      // Generic data
      string               _caption;
      string               _text;
      EMsgBoxIcon          _eIcon;
      EMsgBoxButtons       _eButton;
      EMsgBoxSize          _eSize;
      EMsgBoxRetVal        _resultCode;
      bool                 _exit;
      // Controls
      CPanelBar            _pbrIB;
      CHVLayoutCompound    _hvgIB;
      CSLTextBox           _txtIB;
      CButton              _btn1;
      CButton              _btn2;
      CButton              _btn3;
      TStyle               _style;
      uint32_t             _textYOffset;
  };


  //!
  //! \brief Single-line input box
  //!
  //! \api experiment
  class CInputBoxSL : public CDialogPage {
    public:
      //!
      //! Construct an input box instance using the given parameters.
      CInputBoxSL(CVideoEnv           &iVideoEnv,
                  CEventHandlerBase   *eventHandler,
                  const string        &caption,
                  const string        &defaultText,
                  EMsgBoxSize          eSize               = EMBS_Small,
                  ETBType              sltbType            = ETBT_TextNormal,
                  bool                 sltbEscCancel       = false,
                  PFNSLTextBoxCallback pfnSLTBUserCallback = NULL);
      //!
      //! Standard dtor.
      virtual ~CInputBoxSL() throw();

      // Dialog functions
      //!
      //! Check if the dialog already completed (exit).
      inline bool Exit() const
      { return(_exit); }
      //!
      //! Get the result of the input box.
      inline const string &GetResult() const
      { return(_resultText); }

      // Derived from CDialogPage
      virtual void InitializeAllControls();
      virtual void HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str);

    protected:
      // Controls' event handlers
      void _onBtnOK(_CControl *ctrl);
      void _onBtnCancel(_CControl *ctrl);

      // Generic data
      string               _caption;
      string               _defaultText;
      string               _resultText;
      EMsgBoxSize          _eSize;
      ETBType              _sltbType;
      bool                 _sltbEscCancel;
      PFNSLTextBoxCallback _pfnSLTBUC;
      bool                 _exit;
      // Controls
      CPanelBar            _pbrIB;
      CHVLayoutCompound    _hvgIB;
      CSLTextBox           _txtIB;
      CButton              _btnOK;
      CButton              _btnCancel;
  };


  //!
  //! \brief Multi-line input box
  //!
  //! \api experiment
  class CInputBoxML : public CDialogPage {
    public:
      //!
      //! Construct an input box instance using the given parameters.
      CInputBoxML(CVideoEnv           &iVideoEnv,
                  CEventHandlerBase   *eventHandler,
                  const string        &caption,
                  const string        &defaultText,
                  EMsgBoxSize          eSize               = EMBS_Small,
                  bool                 mltbEscCancel       = false,
                  PFNMLTextBoxCallback pfnMLTBUserCallback = NULL);
      //!
      //! Standard dtor.
      virtual ~CInputBoxML() throw();

      // Dialog functions
      //!
      //! Check if the dialog already completed (exit).
      inline bool Exit() const
      { return(_exit); }
      //!
      //! Get the result of the input box.
      inline const string &GetResult() const
      { return(_resultText); }

      // Derived from CDialogPage
      virtual void InitializeAllControls();
      virtual void HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str);

    protected:
      // Controls' event handlers
      void _onBtnOK(_CControl *ctrl);
      void _onBtnCancel(_CControl *ctrl);

      // Generic data
      string               _caption;
      string               _defaultText;
      string               _resultText;
      EMsgBoxSize          _eSize;
      bool                 _mltbEscCancel;
      PFNMLTextBoxCallback _pfnMLTBUC;
      bool                 _exit;
      // Controls
      CPanelBar            _pbrIB;
      CHVLayoutCompound    _hvgIB;
      CMLTextBox           _txtIB;
      CButton              _btnOK;
      CButton              _btnCancel;
  };


  // Some convenience functions
  //!
  //! Display a modal message box to indicate that a feature
  //! is not implemented yet.
  inline void VNotImplementedMsgBox(CVideoEnv &vidEnv, const string &caption)
  {
    vidEnv.MessageBox(caption, _M("Sorry,\nnot implemented yet!"),
                      EMBI_Information, EMBB_OK, EMBS_Tiny);
  }

} // namespace vge
#endif
