/** \file VX11Dialog2.h */
// ############################################################################################
//  VX11Dialog2.h
//  Definition & implementation of dialog classes used by CVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VID_X11/VX11Environment.h"


#ifndef _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG2_H_
#define _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG2_H_


namespace vge {

  //!
  //! \brief Command-button control
  //!
  //! \todo Make it use somekind of image list class so that some
  //!       buttons can share the same images ?
  //!
  //! \api beta
  class VGE_EXPORT CButton : public _CStringControl {
    public:
      CButton();
      virtual ~CButton() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] pcText               Text to be displayed on the button.
      //! \param[in] bInsideOutline       Flag to indicate whether the inner outline should be drawn or not.
      //! \param[in] bThickOutline        Flag to indicate whether to use thick outline or not.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       const char *pcText,
                       bool bInsideOutline = false, bool bThickOutline = false);
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] xpmDataBuffer        XPM image to be displayed on the button.
      //! \param[in] bInsideOutline       Flag to indicate whether the inner outline should be drawn or not.
      //! \param[in] bThickOutline        Flag to indicate whether to use thick outline or not.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       const char *xpmDataBuffer[],
                       bool bInsideOutline = false, bool bThickOutline = false);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *pcText, bool bInsideOutline = false,
                              bool bThickOutline = false, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, pcText, bInsideOutline, bThickOutline); UNUSED_PARAM(bDuplicateString); }
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *xpmDataBuffer[], bool bInsideOutline = false,
                              bool bThickOutline = false, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, xpmDataBuffer, bInsideOutline, bThickOutline); UNUSED_PARAM(bDuplicateString); }

    protected:
      // Make member functions of _CStringControl inaccessible by user
      _CSTRINGCONTROL_HIDE_MEMBERS()

      XSegment         _atSegTL[4];        // Top-left segments
      XSegment         _atSegBR[4];        // Bottom-right segments
      XRectangle       _tFace;             // Face rectangles
      XRectangle       _tIn;               // Inner outline
      XRectangle       _tOut;              // Outer outline
      XRectangle       _tOutThick;         // Outer thick outline
      int32_t          _n32Xt, _n32Yt;     // Caption (or image) position
      int32_t          _ulX1, _ulX2, _ulY; // Caption's underline position
      bool             _bInsideOutline;    // Flag for inner outline
      bool             _bThickOutline;     // Flag for thick outer outline
      Pixmap           _img;               // Button's image
      Pixmap           _imgMask;           // Button's image mask

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Check-box control
  //!
  //! \api beta
  class VGE_EXPORT CCheckBox : public _CStringControl {
    public:
      virtual ~CCheckBox() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] pcCaption            Text to be displayed for this control.
      //! \param[in] bOriginatedFromRight Flag to indicate whether the text origin should be on the right (instead of on the left).
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y, const char *pcCaption,
                       bool bOriginatedFromRight = false);
      
      // Control functions
      //!
      //! If the mouse pointer is in the control's area, toggle the checked state.
      bool ToggleValue(int32_t n32MouseX, int32_t n32MouseY);
      //!
      //! Set the checked state.
      inline void SetValue(bool bChecked)
      { _bChecked = bChecked; }
      //!
      //! Get the checked state.
      inline bool GetValue() const
      { return(_bChecked); }

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, const char *pcCaption,
                              bool bOriginatedFromRight = false, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, pcCaption, bOriginatedFromRight); UNUSED_PARAM(bDuplicateString); }

    protected:
      // Make member functions of _CStringControl inaccessible by user
      _CSTRINGCONTROL_HIDE_MEMBERS()

      XRectangle _tOut1, _tOut2;     // Border area
      XRectangle _tBox1, _tBox2;     // Fill area
      int32_t    _n32Xt, _n32Yt;     // Caption position
      int32_t    _ulX1, _ulX2, _ulY; // Caption's underline position
      bool       _bChecked;          // Value

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Radio button control
  //!
  //! \api beta
  class VGE_EXPORT CRadioButton : public _CStringControl {
    public:
      virtual ~CRadioButton() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] pcCaption            Text to be displayed for this control.
      //! \param[in] bOriginatedFromRight Flag to indicate whether the text origin should be on the right (instead of on the left).
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y, const char *pcCaption,
                       bool bOriginatedFromRight = false);

      // Control functions
      //!
      //! If the mouse pointer is in the control's area, toggle the checked state.
      bool ToggleValue(int32_t n32MouseX, int32_t n32MouseY, vector<CRadioButton*> &pcsRBInGroup);
      //!
      //! If the mouse pointer is in the control's area, toggle the checked state.
      bool ToggleValue(int32_t n32MouseX, int32_t n32MouseY, CRadioButton *csRBInGroup, ...);
      //!
      //! Set the checked state.
      inline void SetValue(bool bChecked)
      { _bChecked = bChecked; }
      //!
      //! Set the checked state.
      inline bool GetValue() const
      { return(_bChecked); }

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, const char *pcCaption,
                              bool bOriginatedFromRight = false, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, pcCaption, bOriginatedFromRight); UNUSED_PARAM(bDuplicateString); }

    protected:
      // Make member functions of _CStringControl inaccessible by user
      _CSTRINGCONTROL_HIDE_MEMBERS()

      XRectangle _tOut1, _tOut2;     // Border area
      XRectangle _tBox1, _tBox2;     // Fill area
      int32_t    _n32Xt, _n32Yt;     // Caption position
      int32_t    _ulX1, _ulX2, _ulY; // Caption's underline position
      bool       _bChecked;          // Value

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Spin-button control
  //!
  //! \api beta
  class VGE_EXPORT CSpinButton : public _CControl {
    public:
      virtual ~CSpinButton() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] eSpinButtonType      Style of the spin button.
      //! \param[in] eTextAligment        Alligment of the spin button's text.
      //! \param[in] eValueType           Type of the value.
      //! \param[in] bWrapValue           Flag to indicate whether the value should wrap if under/over its minimum/maximum setting.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       ESBStyle eSpinButtonType = ESBS_VLeft,
                       EHAlignment eTextAligment = EHTA_Left,
                       ESBValue eValueType = ESBV_Integer, bool bWrapValue = true);
      //!
      //! Set the minimum, maximum and step factor for spin buttons with
      //! integer-type values.
      inline void SetValueProp(int32_t n32Min, int32_t n32Max, int32_t n32Step);
      //!
      //! Set the minimum, maximum and step factor for spin buttons with
      //! floating-point-type values.
      inline void SetValuePropF(float fMin, float fMax, float fStep);

      // Control functions
      //!
      //! Update (redraw) the spin button's value
      void UpdateValue() const;
      //!
      //! If the mouse pointer is in the control's area, toggle the value up/down.
      bool ToggleValue(int32_t n32MouseX, int32_t n32MouseY);
      //!
      //! Set the current value for spin buttons with integer-type values.
      inline void SetValue(int32_t n32Value)
      { _uVP.tI.n32Cur = n32Value; _bDirty = true; }
      //!
      //! Set the current value for spin buttons with floating-point-type values.
      inline void SetValueF(float fValue)
      { _uVP.tF.fCur = fValue; _bDirty = true; }
      //!
      //! Get the current value for spin buttons with integer-type values.
      inline int32_t GetValue() const
      { return(_uVP.tI.n32Cur); }
      //!
      //! Get the current value for spin buttons with floating-point-type values.
      inline float GetValueF() const
      { return(_uVP.tF.fCur); }

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bUpPressed, bool bDnPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              ESBStyle eSpinButtonType = ESBS_VLeft, EHAlignment eTextAligment = EHTA_Left,
                              ESBValue eValueType = ESBV_Integer, bool bWrapValue = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, eSpinButtonType, eTextAligment, eValueType, bWrapValue); }

#define _vmcm mutable
    protected:
      ESBValue    _eVT;           // Value type
      EHAlignment _eTA;           // Text alignment mode
      XRectangle  _tOutline;      // Outline
      XSegment    _atText[4];     // Text box segments
      XSegment    _tUp, _tDn;     // Buttons' area
      XSegment    _atUp[4];       // Up button segments
      XSegment    _atDn[4];       // Down button segments
      XSegment    _atArrow[8];    // Arrows segments
      int32_t     _n32Xo;         // Caption X origin
      bool        _bWrap;         // Wrapping flag
_vmcm int32_t     _n32Xt;         // Text position
      int32_t     _n32Yt;         // ...
_vmcm bool        _bDirty;        // Should we update the _acValue[] below?
_vmcm uint8_t     _un8Len;        // String len of the value string
_vmcm char        _acValue[128];  // Value (in the form of string)
_vmcm union _UVP {                // Values (in the original form)
        struct _TVI {
          int32_t n32Min, n32Max, n32Step, n32Cur;
        } tI;
        struct _TVF {
          float fMin, fMax, fStep, fCur;
        } tF;
      } _uVP;
#undef _vmcm

      // Friend classes
      friend class CVideoEnv;
  };

  void CSpinButton::SetValueProp(int32_t n32Min, int32_t n32Max, int32_t n32Step)
  {
    _uVP.tI.n32Min  = n32Min;
    _uVP.tI.n32Max  = n32Max;
    _uVP.tI.n32Step = n32Step;
  }

  void CSpinButton::SetValuePropF(float fMin, float fMax, float fStep)
  {
    _uVP.tF.fMin  = fMin;
    _uVP.tF.fMax  = fMax;
    _uVP.tF.fStep = fStep;
  }


  //
  // Extended spin-button control
  //
  //       < NOT IMPLEMENTED YET !!!>


  //
  // Slider-bar control
  //
  //       < NOT IMPLEMENTED YET !!!>

} // namespace vge
#endif
